#ifndef AARECT_H_
#define AARECT_H_

#include "common.h"

namespace physics {
	class AARect
	{
		public:
		
		USE_2D();
		
		AARect () : m_min(), m_max() {}
		AARect (const Vector& p1, const Vector& p2);

		float width () const { return m_max.x - m_min.x; }
		float height () const { return m_max.y - m_min.y; }
		float depth () const { return m_max.z - m_min.z; }

		bool contains (const Vector &p) const;
		
		/**
		 * Do these axis-aligned rectangles intersect (or if dist is provided, are they at most dist apart)?
		 */
		bool intersects (const AARect& otherObject, float dist = 0) const;

		Vector getVertex (int i) const;
		
		Vector2D getExtremePoint (const Vector2D& dir, int& i) const;
		
		Vector getMin () const { return m_min; }
		Vector getMax () const { return m_max; }

		float getMaxX () const { return m_max.x; }
		float getMinX () const { return m_min.x; }
		float getMaxY () const { return m_max.y; }
		float getMinY () const { return m_min.y; }
		float getMaxZ () const { return m_max.z; }
		float getMinZ () const { return m_min.z; }

		private:

		Vector m_min;
		Vector m_max;
	};
};

inline physics::AARect::AARect (const Vector2D& p1, const Vector2D& p2) {
	if (p1.x < p2.x) {
		m_min.x = p1.x;
		m_max.x = p2.x;
	} else {
		m_min.x = p2.x;
		m_max.x = p1.x;
	}
	
	if (p1.y < p2.y) {
		m_min.y = p1.y;
		m_max.y = p2.y;
	} else {
		m_min.y = p2.y;
		m_max.y = p1.y;
	}
}

inline bool physics::AARect::contains (const Vector2D& p) const {
	return (p.x >= m_min.x && p.x <= m_max.x && p.y >= m_min.y && p.y <= m_max.y);
}

inline physics::Vector2D physics::AARect::getVertex (int i) const {
	switch (i % 4) {
		case 0: return Vector2D(m_min.x, m_min.y);
		case 1: return Vector2D(m_max.x, m_min.y);
		case 2: return Vector2D(m_max.x, m_max.y);
		case 3: return Vector2D(m_min.x, m_max.y);
	}

	return Vector2D();
}

inline physics::Vector2D physics::AARect::getExtremePoint (const physics::Vector2D& dir, int& i) const {
	if (dir.x < 0.0f) {
		i = (dir.y > 0.0f) ? 0 : 1;
	} else {
		i = (dir.y < 0.0f) ? 2 : 3;
	}
	
	return getVertex(i);
}

// Test another AABB against this one. If it intersects return 1 else 0
inline bool physics::AARect::intersects (const AARect& other, float dist) const {
	// Exit with no intersection if separated along an axis
	if (getMaxX() + dist < other.getMinX() || getMinX() - dist > other.getMaxX() ) return 0;
	if (getMaxY() + dist < other.getMinY() || getMinY() - dist > other.getMaxY() ) return 0;
	// Overlapping on all axes means AABBs are intersecting
	return 1;
}


#endif /* AARECT_H_ */

