#ifndef _PHYSICS_AREA_
#define _PHYSICS_AREA_

#include "common.h"

#include "child_object.h"
#include "moi.h"

namespace physics {

	class AARect;
	class Area : public ChildObject<TwoDee, Body2D> {
		public:

		USE_2D();

		enum Type {
			CIRCLE,
			LINE,
			CAPSULE,
			RECT,
			POLYGON
		};

		virtual ~Area () {}
		
		virtual bool contains (const Vector2D &p) const = 0;
		virtual bool intersects (const AARect& rect) const = 0;
		//virtual bool intersects (const Shape& shape) const = 0;
		//virtual bool intersectRay (const Vector& start, const Vector& dir, Vector& p) const;
		
		virtual float getRadius () const = 0;
		float getRadiusSq () const;

		virtual Vector2D getNormal (int i) const = 0;
		virtual Vector2D getVertex (int i) const = 0;

		virtual int getVertexCount () const = 0;

		/**
		 * Find the extreme point of the shape in some direction.
		 * 
		 * @param direction Direction in which to look (must be normalised).
		 * @param i Set by reference to the ID of the vertex which is at the extreme point.
		 * @return World coordinates of the extreme point.
		 */
		virtual Vector2D getExtremePoint (const Vector2D& dir, int& i) const = 0;

		// dir MUST be normalised
		Vector2D getExtremePoint (const Vector2D& dir) const;

		// Get Min/Max X/Y values
		virtual float getMinX () const {return getExtremePoint(Vector(-1,0)).x;}
		virtual float getMaxX () const {return getExtremePoint(Vector(1,0)).x;}
		virtual float getMinY () const {return getExtremePoint(Vector(0,-1)).y;}
		virtual float getMaxY () const {return getExtremePoint(Vector(0,1)).y;}

		virtual Type type () const = 0;

		float getDensity () const { return m_density; }
		
		/**
		 * Sets the density of this shape.
		 * @param density New density.
		 * @param updateBody If true, the mass and moment of inertia of the parent body will be recalculated.
		 */
		void setDensity (float density, bool updateBody = true);


		float getRestitution () const { return m_restitution; }
		void setRestitution (float f) { m_restitution = f; }

		float getFriction () const { return m_friction; }
		void setFriction (float f) { m_friction = f; }

		virtual float getArea () const = 0;

		float getMass () const { return m_density * getArea(); }

		virtual float getMOI () const = 0; // Moment of Inertia

		Body* getBody () { return static_cast<Body*>(getParent()); }
		const Body* getBody () const { return static_cast<const Body*>(getParent()); }

		protected:

		Area (const Vector& p = Vector(), const Orientation& o = Orientation()) : 
				 ChildObject<TwoDee, Body>(p, o), 
				 m_density(Globals::DEFAULT_DENSITY),
				 m_restitution(Globals::DEFAULT_RESTITUTION),
				 m_friction(Globals::DEFAULT_FRICTION) {}

		float m_density;
		float m_restitution;
		float m_friction;
	};
};

inline float physics::Area::getRadiusSq () const {
	const float r = getRadius();
	return r * r;
}

inline physics::Vector2D physics::Area::getExtremePoint (const physics::Vector2D& dir) const {
	int i;
	return getExtremePoint(dir, i);
}

// Some methods in Object rather understandably need to be able to see the definition of the parent type
#include "body.h"

inline void physics::Area::setDensity (float f, bool updateBody) {
	m_density = f;
	
	if (updateBody && Area::getBody()) {
		getBody()->shapesChanged();
	}
}

#endif

