#include "body.h"
#include "shape.h"

#include <algorithm>
#include <boost/foreach.hpp>

#include "contact.h"


using namespace physics;

template <typename d>
physics::Body<d>::~Body () {
	BOOST_FOREACH(Shape* s, m_shapes) {
		delete s;
	}
}

template <typename d>
void physics::Body<d>::add (Shape* shape) {
	m_shapes.push_back(shape);
	shape->setParent(this);
	shapesChanged();
}

//TODO Really need to declare some object-ownership rules at some point
template <typename d>
void physics::Body<d>::remove (Shape* shape) {
	m_shapes.remove(shape);
	shapesChanged();
}

template <typename d>
void physics::Body<d>::removeShapes () {
	m_shapes.clear();
}

template <typename d>
void physics::Body<d>::add (CollisionListener<d>* listener) {
	m_listeners.push_back(listener);
}

//TODO Really need to declare some object-ownership rules at some point
template <typename d>
void physics::Body<d>::remove (CollisionListener<d>* listener) {
	m_listeners.remove(listener);
}

template <typename d>
void physics::Body<d>::shapesChanged () {
	// Get new mass
	m_inverseMass = calculateInverseMass();

	// Find new centre of mass
	Vector offset;

	BOOST_FOREACH(Shape* shape, m_shapes) {
		offset += shape->getLocalPosition() * shape->getMass();
	}
	offset *= m_inverseMass;

	// Move body so its position is at the COM
	m_position += offset;

	// Find offset in local coordinates
	offset = getOrientation().inverseTransform(offset);

	// Move shapes in the opposite direction so nothing appears to change
	BOOST_FOREACH(Shape* shape, m_shapes) {
		shape->localTranslateInPlace(-offset);
	}

	// Get new moment of inertia
	m_moi = MOI(this);
	
	// Calculate new Bounding Radius
	m_boundingRadius = calculateBoundingRadius();
}

template <typename d>
void physics::Body<d>::thisMoved () {
	BOOST_FOREACH(Shape* shape, m_shapes) {
		shape->parentMoved();
	}
}

template <typename d>
float physics::Body<d>::calculateInverseMass () const {
	float m = 0;

	BOOST_FOREACH(Shape* shape, m_shapes) {
		m += shape->getMass();
	}

	if (m != 0) return 1.0f / m;
	else return 0;
}

template <typename d>
void physics::Body<d>::setStatic (bool s) {
	if (s) {
		m_inverseMass = 0;
		m_moi = MOI();
		m_velocity = Vector(); 
		m_rotation = Rotation();
	} else {
		shapesChanged();
	}
}

template <typename d>
void physics::Body<d>::setRotatable (bool r) {
	shapesChanged();
	
	if (! r) {
		m_moi = MOI();
		m_rotation = Rotation();
	}
}

template <typename d>
void physics::Body<d>::setPinned (bool p) {
	shapesChanged();
	
	if (p) {
		m_inverseMass = 0;
		m_velocity = Vector();
	}
}

template <typename d>
void physics::Body<d>::setAwake (const bool awake) {
	//if (awake == m_isAwake) { return; }
	
	if (awake && Globals::useSleeping) {
		m_isAwake = true;
		
		//initialise the RWAs to a high enough value that the body won't immediately fall asleep again.
		m_linearRWA = Globals::linearVelocitySleepEpsilon * 2;
		m_rotationalRWA = Globals::angularVelocitySleepEpsilon * 2;
	} else if (m_canSleep && !awake) {
		m_isAwake = false;
		killVelocity();
	}
}

template <typename d>
bool physics::Body<d>::contains (const Vector& p) const {
	BOOST_FOREACH(Shape* shape, m_shapes) {
		if (shape->contains(p)) { return true; }
	}
	return false;
}

template <typename d>
bool physics::Body<d>::intersects (const AARect& r) const {
	BOOST_FOREACH(Shape* shape, m_shapes) {
		if (shape->intersects(r)) { return true; }
	}
	return false;
}

template <typename d>
typename d::Vector Body<d>::getVelocityAtPoint (const Vector& p) const {
	Vector velocity = getVelocity();

	// Rotation affects the velocity of the collision point
	velocity += getRotation().getVelocityAtPoint(p - getPosition());
	return velocity;
}

template <typename d>
typename d::Vector Body<d>::getBiasVelocityAtPoint (const Vector& p) const {
	Vector velocity = m_biasVelocity;

	// Rotation affects the velocity of the collision point
	velocity += m_biasRotation.getVelocityAtPoint(p - getPosition());
	return velocity;
}

template <typename d>
typename d::Vector Body<d>::getDrag () {
	/************************************************************************
	This function returns a vector corresponding to the force the body would
	experience as a result of air resistance.
	************************************************************************/
	
	if (Globals::airResistance) {
		if (Globals::linearDrag) {
			float c = Globals::linearDragCoefficient;
			//std::cout << "v: " << getVelocity() << ", drag: " << (-c * getVelocity()) << std::endl;
			return -c * getVelocity();
		} else { //quadratic drag
			float c = Globals::quadraticDragCoefficient;
			float A = (m_boundingRadius * 2) * (m_boundingRadius * 2); //crude attempt to get an orthogonal area
			return -A * getVelocity() * getVelocity().length() * c;
		}
	} else return Vector(0, 0);
}

template <typename d>
void physics::Body<d>::applyForces (float t, const Vector& a, float frameRWABias) {
	
	if (!m_isAwake) return; //if it's asleep, don't bother processing this stuff.
	
	if (Globals::useSleeping && m_canSleep) {
		//Update current motion average, and sleep the body if necessary.
		float currentV = m_velocity.length();
		float currentTheta = m_rotation.scalarSize();
		
		m_linearRWA = frameRWABias * m_linearRWA + (1 - frameRWABias) * currentV;
		m_rotationalRWA = frameRWABias * m_rotationalRWA + (1 - frameRWABias) * currentTheta;

		if (m_linearRWA > 10 * Globals::linearVelocitySleepEpsilon)
			m_linearRWA = 10 * Globals::linearVelocitySleepEpsilon;
		if (m_rotationalRWA > 10 * Globals::angularVelocitySleepEpsilon)
			m_rotationalRWA = 10 * Globals::angularVelocitySleepEpsilon;
		
		if ((m_linearRWA < Globals::linearVelocitySleepEpsilon) && (m_rotationalRWA < Globals::angularVelocitySleepEpsilon)) {
			setAwake(false);
			return;
		}
	}
	// end of sleep processing stuff
	
	//apply air resistance
	if (Globals::airResistance) {
		applyImpulse(getDrag(), getPosition(), false);
	}
	
	if (m_inverseMass != 0) {
		//m_position += m_velocity * t + a * t * t / 2;
		m_velocity += a * t;
	}
}

template <typename d>
void physics::Body<d>::move (float t) {
	if (!m_isAwake) return; //if it's asleep, don't bother processing this stuff.
	
	m_position += m_velocity * t;// - a * t * t / 2;
	
	m_orientation += m_rotation * t;
	m_orientation.normalise();
	m_moi.updateOrientation(m_orientation);
	
	thisMoved();
}

template <typename d>
void physics::Body<d>::moveBias (float t) {
	if (!m_isAwake) return; //if it's asleep, don't bother processing this stuff.
	
	m_position += m_biasVelocity;
	
	m_orientation += m_biasRotation;
	
	m_orientation.normalise();
	
	m_moi.updateOrientation(m_orientation);
	
	thisMoved();
	
	m_biasVelocity = Vector();
	m_biasRotation = Rotation();
}

template <typename d>
void physics::Body<d>::setDensity (float f, bool updateMass) {
	BOOST_FOREACH(Shape* shape, m_shapes) {
		shape->setDensity(f, false);
	}
	
	if (updateMass) {
		shapesChanged();
	}
}

template <typename d>
void physics::Body<d>::setRestitution (float f) {
	BOOST_FOREACH(Shape* shape, m_shapes) {
		shape->setRestitution(f);
	}
}

template <typename d>
void physics::Body<d>::setFriction (float f) {
	BOOST_FOREACH(Shape* shape, m_shapes) {
		shape->setFriction(f);
	}
}

template <typename d>
const AARect physics::Body<d>::createAABB () const {
	// assert m_shapes.size() > 0
	
	typename shape_container::const_iterator it = m_shapes.begin();
	
	Shape* shape = *it;
	
	float maxX = shape->getMaxX();
	float minX = shape->getMinX();
	float maxY = shape->getMaxY();
	float minY = shape->getMinY();
	
	it++;

	for ( ; it != m_shapes.end(); it++) {
		shape = *it;
		
		float v = shape->getMaxX();
		if (v > maxX) { maxX = v; }
		
		v = shape->getMinX();
		if (v < minX) { minX = v; }
		
		v = shape->getMaxY();
		if (v > maxY) { maxY = v; }
		
		v = shape->getMinY();
		if (v < minY) { minY = v; }
	}

	return AARect(Vector2D(minX, minY), Vector2D(maxX, maxY));
}

// Does all of the actual work
template <typename d>
float physics::Body<d>::calculateBoundingRadius() const {
	
	// Temporary Variable that is going to store the largest radius
	float maxBoundingRadius = -1;
	float currentBoundingRadius = -1;
	float currentRadius;
	
	// Iterate over all of the shapes that make up the current object
	BOOST_FOREACH(Shape* shape, m_shapes) {
		
		// Save the radius of the current shape locally
		currentRadius = shape->getRadius();
		// Iterate over all the vertexes of each shape
		for (int vi = 0; vi < shape->getVertexCount(); vi++) {
			// Find the total distance from centre of body to furthest point on shape
			currentBoundingRadius = shape->getVertex(vi).distance(getPosition())+currentRadius;
			// If its larger than any previous seen value, store it
			if (currentBoundingRadius > maxBoundingRadius) {
				maxBoundingRadius = currentBoundingRadius;
			}		
		}
	}
	
	return maxBoundingRadius;
}

// Because we're implementing methods of a template class in a cpp file
// (rather than in the header), the compiler can have trouble linking the
// declaration and the implementation.
// To fix this, we use the following magic line of code.
// See: http://www.parashift.com/c++-faq-lite/templates.html#faq-35.15
// Note that this _must_ be at the bottom of the file

template class Body<TwoDee>;
template class Body<ThreeDee>;

