#ifndef _PHYSICS_BODY_
#define _PHYSICS_BODY_

#include "common.h"

#include "vector.h"
#include "orientation.h"
#include "rotation.h"
#include "moi.h"
#include "aarect.h"

#include <list>

#include <limits>

namespace physics {
	
	template <typename d> class CollisionListener;
	/*! \class Body
	*	Class that specifies the representation of bodies ( a entity comprised of one or more shapes) within the simulation. 
	*	Provides functions for creating, adding to, querying and deleting bodies.
	*/
	template <typename d>
	class Body {
		public:

		BODY_USE_DIMENSION(d);
		
		typedef std::list<Shape*> shape_container;
		typedef std::list<Joint<d>*> joint_container;
		typedef typename joint_container::iterator joint_iterator;
		typedef typename joint_container::const_iterator joint_citerator;

		Body () : data(0), m_inverseMass(0), m_moi(), m_orientation(), m_rotation(), m_isAwake(true), m_linearRWA(10), m_rotationalRWA(10), m_canSleep(Globals::useSleeping) {}

		explicit Body (Shape* shape) : data(0), m_orientation(), m_rotation(), m_isAwake(true), m_linearRWA(10), m_rotationalRWA(10), m_canSleep(Globals::useSleeping) {
			add(shape);
		}
		
		/**
		 * This will delete all shapes owned by this body.
		 * If you want to keep references to a shape, you should remove it
		 * from the body before the destructor is called.
		 */
		~Body ();
		
		//! Creates a new body and returns a pointer to it
		static Body* create () {
			return new Body();
		}
		
		//! Creates a body with a given shape and returns a pointer to it
		/*! \param shape A pointer to a shape to be included in the body
			\return body A pointer to the new body created
		*/
		static Body* create (Shape* shape) {
			return new Body(shape);
		}

		//! Creates a body with a given shape, sets it to static, and returns a pointer to it
		/* 	\param shape A poitner to a shape to be included in the body
			\return body A pointer to the new body created
		*/
		static Body* createStatic (Shape* shape) {
			Body* b = new Body(shape);

			b->setStatic(true);
			return b;
		}

		Vector getPosition () const { return m_position; }
		void setPosition (const Vector& p) { m_position = p; thisMoved(); }

		Vector getVelocity () const { return m_velocity; }
		void setVelocity (const Vector& v, bool wakeUp = true) { if (wakeUp) setAwake(true); m_velocity = v; }

		Orientation getOrientation () const { return m_orientation; }
		void setOrientation (const Orientation& o) { m_orientation = o; thisMoved(); }

		Rotation getRotation () const { return m_rotation; }
		void setRotation (const Rotation& r, bool wakeUp = true) { if (wakeUp) setAwake(true); m_rotation = r; }

		float getMass () const { return m_inverseMass ? (1.0f / m_inverseMass) : std::numeric_limits<float>::infinity(); }
		float getInverseMass () const { return m_inverseMass; }
		
		//! Return the Moment of Inertia
		MOI getMOI () const { return m_moi; }

		Vector getDrag();
		
		//! Queries whether a body is sleeping or not
		bool isAwake() const { return m_isAwake && ! isStatic(); }
		
		//! Allows a body to be slept and woken
		/*!	\param awake Boolean value. True sets body to be awake, false puts its to sleep
		*/
		void setAwake (const bool awake);
		
		//! Sets whether a body is allowed to sleep or not 
		void setCanSleep (const bool sleepable) { m_canSleep = sleepable; }

		//! Checks if a position is contained within the space the body occupies
		/*! \param Position Vector in world co-ordinates
		*/
		bool contains (const Vector& p) const;
		
		//! Checks if the body intersects a given Bounding Box
		/*! \param AARect Bounding box
		*/
		bool intersects (const AARect& rect) const;
		//bool intersects (const Shape& shape) const;
		
		// TODO: Implement a Shape::intersectRay method so we can actually do something here
		bool intersectsRay (const Vector& start, const Vector& dir, Vector& p) const { return false; }
		
		//! Add a shape to the body
		void add (Shape* area);
		//! Remove a shape from the body
		void remove (Shape* area);
		
		void removeShapes ();
		
		//! Gets all the shapes contained within this body
		/*! \return m_shapes List of all shapes within the body
		*/
		const shape_container& getShapes () const { return m_shapes; }
		
		/**
		 * Applies an impulse to the body.
		 * 
		 * @param impulse The impulse to apply.
		 * @param point Position in world coordinates to apply impulse at.
		 * @param wakeUp If true (the default), will wake the body up if it is asleep.
		 */
		void applyImpulse (const Vector& impulse, const Vector& point, bool wakeUp = true);
		
		/**
		 * Applies an impulse to the body at the centre of mass.
		 * This will not cause any change in rotation.
		 * 
		 * @param impulse The impulse to apply.
		 * @param wakeUp If true (the default), will wake the body up if it is asleep.
		 */
		void applyImpulse (const Vector& impulse, bool wakeUp = true);
		
		/**
		 * Only CattoContactResolver should use this method.
		 */
		void applyBiasImpulse (const Vector& impulse, const Vector& point);
		
		/**
		 * Only World should use this method.
		 */
		void applyForces (float timestep, const Vector& acceleration, float frameRWABias);
		
		/**
		 * Only World should use this method.
		 */
		void move (float timestep);
		
		/**
		 * Only World should use this method.
		 */
		void moveBias (float timestep);

		void translate (const Vector& v);

		void rotate (Rotation r);

		void killVelocity () { m_velocity = Vector(); m_rotation = Rotation(); }
		
		/**
		 * Used for any data the application using the library wants to attach.
		 */
		void* data;
		
		/**
		 * Transform a position from world coordinates to local coordinates.
		 */
		Vector worldToLocal (const Vector& p) const;
		
		/**
		 * Transform a position from local coordinates to world coordinates.
		 */
		Vector localToWorld (const Vector& p) const;

		/**
		 * Transform a direction from world coordinates to local coordinates.
		 */
		Vector worldToLocalDir (const Vector& v) const;
		
		/**
		 * Transform a direction from local coordinates to world coordinates.
		 */
		Vector localToWorldDir (const Vector& v) const;

		/**
		 * Find the combined linear and angular velocity of a point on the body
		 * 
		 * @param position Point on the body in world coordinates
		 */
		Vector getVelocityAtPoint (const Vector& position) const;
		
		/**
		 * Only CattoContactResolver should use this method.
		 */
		Vector getBiasVelocityAtPoint (const Vector& p) const;

		void shapesChanged ();

		// TODO: make these settings persist when shapes are added/removed
		
		bool isStatic () const { return m_inverseMass == 0 && m_moi.isInfinite(); }
		void setStatic (bool b);
		
		bool isRotatable () const { return ! m_moi.isInfinite(); }
		void setRotatable (bool b);
		
		bool isPinned () const { return m_inverseMass == 0; }
		void setPinned (bool b);
		
		/**
		 * This is a convenience method to set the density of every shape currently attached.
		 * It will not affect shapes that are added later.
		 * 
		 * @param density New density for all attached shapes.
		 * @param updateMass If true, recalculate this body's mass and moment of inertia.
		 */
		void setDensity (float density, bool updateMass = true);

		/**
		 * This is a convenience method to set the coefficient of restitution of every shape currently attached.
		 * It will not affect shapes that are added later.
		 */
		void setRestitution (float f);

		/**
		 * This is a convenience method to set the coefficient of friction of every shape currently attached.
		 * It will not affect shapes that are added later.
		 */
		void setFriction (float f);

		const AARect createAABB () const;
		
		// We could cache the bounding box instead of creating it every time
		// but currently it will only be requested once per body per frame
		// and we would have to regenerate it at the start of the frame anyway.
		const AARect getAABB () const { return createAABB(); }
		
		// We have implemented the following method as an alternative
		// (the results of which _can_ be cached):
		float calculateBoundingRadius () const;
		
		float getBoundingRadius () const { return m_boundingRadius; }
		
		#if BODY_STORES_CONTACTS
			std::list<Contact<d>*> contacts;
		#endif
		
		joint_container joints;
		
		void add (CollisionListener<d>* listener);
		void remove (CollisionListener<d>* listener);
		
		std::list<CollisionListener<d>*> m_listeners;

		private:

		// Prevent Body objects from being copied
		Body (const Body& b);
		Body& operator = (const Body& b);

		void thisMoved ();

		float calculateInverseMass () const;

		float m_inverseMass;
		MOI m_moi;
		
		float m_boundingRadius;

		Vector m_position;
		Vector m_velocity;
		
		Vector m_biasVelocity;

		Orientation m_orientation;
		Rotation m_rotation;
		
		Rotation m_biasRotation;

		shape_container m_shapes;
		
		bool m_isAwake; //Whether the body is awake or not.
		float m_linearRWA; //A recency-weighted average of the linear velocity of the object, used to determine whether to sleep it or not.
		float m_rotationalRWA; //A recency-weighted average of the rotational velocity of the object, used to determine whether to sleep it or not.
		bool m_canSleep; //Whether the body can be put to sleep or not. Certain bodies, such as user-controlled ones, shouldn't be slept.
	};
};

template <typename d>
inline typename d::Vector physics::Body<d>::localToWorld (const Vector& p) const {
	return getPosition() + getOrientation().transform(p);
}

template <typename d>
inline typename d::Vector physics::Body<d>::worldToLocal (const Vector& p) const {
	return getOrientation().inverseTransform(p - getPosition());
}

template <typename d>
inline typename d::Vector physics::Body<d>::localToWorldDir (const Vector& p) const {
	return getOrientation().transform(p);
}

template <typename d>
inline typename d::Vector physics::Body<d>::worldToLocalDir (const Vector& p) const {
	return getOrientation().inverseTransform(p);
}

template <typename d>
inline void physics::Body<d>::translate (const Vector& v) {
	m_position += v;

	thisMoved();
}

template <typename d>
inline void physics::Body<d>::rotate (Rotation r) {
	m_orientation += r;

	thisMoved();
}

template <typename d>
inline void physics::Body<d>::applyImpulse (const Vector& impulse, const Vector& point, bool wakeUp) {
	if (wakeUp) setAwake(true);
	
	m_velocity += impulse * getInverseMass();
	
	Vector offset = point - getPosition();
	m_rotation += getMOI().getRotationChange(impulse, offset);
}

template <typename d>
inline void physics::Body<d>::applyImpulse (const Vector& impulse, bool wakeUp) {
	if (wakeUp) setAwake(true);
	
	m_velocity += impulse * getInverseMass();
}

template <typename d>
inline void physics::Body<d>::applyBiasImpulse (const Vector& impulse, const Vector& point) {
	m_biasVelocity += impulse * getInverseMass();
	
	Vector offset = point - getPosition();
	m_biasRotation += getMOI().getRotationChange(impulse, offset);
}

#endif

