#ifndef _PHYSICS_BOX_
#define _PHYSICS_BOX_

#include "common.h"

#include "volume.h"

namespace physics {
	/*! \class Box
	*	Provides the box shape in the 3D simulation
	*/
	class Box : public Volume
	{
		public:
		
		Box (const Vector& p1, const Vector& p2);
		Box (const Vector& p, float width, float height, float depth, const Orientation& o = Orientation());
		
		bool contains (const Vector &p) const;
		
		float getRadius () const { return 0; }
		
		Vector getNormal (int i) const;
		
		Vector getVertex (int i) const { return worldPoints[i % 8]; }

		int getVertexCount () const { return 8; }
		
		Vector getExtremePoint (const Vector& dir, int& i) const;
		
		void getExtremeEdge (const Vector& dir, Vector3D& p1, Vector3D& p2) const;
		
		float getVolume () const { return size.x * size.y * size.z; }
		
		//! Get the Moment of Inertia
		/*! \return Vector3D; Moment of Inertia
		*/
		Vector3D getMOI () const; 
		
		Vector3D getSize () const { return size; }
		
		Type type () const { return BOX; }
		
		private:
		Vector size;
		
		void init ();
		
		virtual void thisMoved ();
		
		Vector worldPoints[8];
	};
};

inline physics::Box::Box (const Vector& p1, const Vector& p2)
	: Volume((p1 + p2) / 2), size(p2 - p1) { init(); }

inline physics::Box::Box (const Vector& p, float w, float h, float d, const Orientation& o)
	: Volume(p, o), size(w, h, d) { init(); }

inline void physics::Box::init () {
	if (size.x < 0) { size.x *= -1; }
	if (size.y < 0) { size.y *= -1; }
	if (size.z < 0) { size.z *= -1; }
	
	thisMoved();
}

inline physics::Vector3D physics::Box::getNormal (int i) const {
	i = (i % 3);
	
	Vector normal;
	
	normal[i] = 1;
	
	return localToWorldDir(normal);
}

inline bool physics::Box::contains (const physics::Vector3D &p) const {
	Vector q = worldToLocal(p);
	
	q += size / 2;
	
	return (q.x >= 0 && q.x <= size.x && q.y >= 0 && q.y <= size.y && q.z >= 0 && q.z <= size.z);
}

inline physics::Vector3D physics::Box::getExtremePoint (const physics::Vector3D& direction, int& i) const {
	i = 0;
	
	Vector dir = worldToLocalDir(direction);
	
	if (dir.x > 0.0f) { i += 1; }
	if (dir.y > 0.0f) { i += 2; }
	if (dir.z > 0.0f) { i += 4; }
	
	return worldPoints[i];
}

inline void physics::Box::getExtremeEdge (const physics::Vector3D& dir, physics::Vector3D& p1, physics::Vector3D& p2) const {
	int i;
	
	p1 = getExtremePoint(dir, i);
	
	p2 = worldPoints[i^1];
	
	float d = dir.dot(p2);
	
	if (dir.dot(worldPoints[i^2]) > d) {
		p2 = worldPoints[i^2];
		d = dir.dot(p2);
	}
	
	if (dir.dot(worldPoints[i^4]) > d) {
		p2 = worldPoints[i^4];
		//d = dir.dot(p2);
	}
}

inline physics::Vector3D physics::Box::getMOI () const {
	return Vector3D(
		size.y * size.y + size.z * size.z,
		size.x * size.x + size.z * size.z,
		size.x * size.x + size.y * size.y
	) * getMass() / 12;
}

inline void physics::Box::thisMoved () {
	const Vector d = size / 2;
	
	worldPoints[0] = localToWorld(Vector(-d.x, -d.y, -d.z));
	worldPoints[1] = localToWorld(Vector(d.x,  -d.y, -d.z));
	worldPoints[2] = localToWorld(Vector(-d.x, d.y,  -d.z));
	worldPoints[3] = localToWorld(Vector(d.x,  d.y,  -d.z));
	worldPoints[4] = localToWorld(Vector(-d.x, -d.y, d.z));
	worldPoints[5] = localToWorld(Vector(d.x,  -d.y, d.z));
	worldPoints[6] = localToWorld(Vector(-d.x, d.y,  d.z));
	worldPoints[7] = localToWorld(Vector(d.x,  d.y,  d.z));
}

#endif
