#ifndef _PHYSICS_BROADPHASE_
#define _PHYSICS_BROADPHASE_

#include "common.h"
#include "world.h"
#include "joint.h"

#include <list>
#include <vector>

namespace physics {
	class AARect;
	/**
	 * Broad phase collision detection: detects pairs of bodies which are potentially colliding.
	 * Default naive implementation checks all pairs with n^2 performance.
	 */
	template <typename d>
	class BroadPhase {
		public:
		
		USE_DIMENSION(d);
		
		typedef std::pair<Body*, Body*> BodyPair;
		
		typedef typename World::body_pair_container output_container;
		
		BroadPhase () : m_world(0) {}
		virtual ~BroadPhase () {}
		
		/**
		 * Attach this BroadPhase object to a World.
		 * Subclasses that override this method should call this from their implementation.
		 */
		virtual void init (World* world) { m_world = world; }
		
		/**
		 * Find all pairs of bodies that might be colliding.
		 */
		virtual output_container findPairs ();
		
		/**
		 * Find all bodies located at this position (if any).
		 */
		virtual std::list<Body*> findBeneathAll (const Vector& position);
		
		/**
		 * Find all bodies touching this rect.
		 */
		virtual std::list<Body*> findBeneathAll (const AARect& rect);
		
		/**
		 * Find the first body located at this position (if any).
		 */
		virtual Body* findBeneath (const Vector& position);
		
		/**
		 * Find the first body touching this rect
		 */
		virtual Body* findBeneath (const AARect& rect);
		
		/**
		 * Intersect a ray with the world and return the first Body it hits
		 * 
		 * @param start Origin of the ray.
		 * @param dir Direction of the ray.
		 * @param p Set by reference to the intersection point found.
		 */
		virtual Body* intersectRay (const Vector& start, const Vector& dir, Vector& p);
		
		/**
		 * Notification that a body has been added to the world.
		 * Default implementation does nothing.
		 */
		virtual void bodyAdded (Body* body) {}
		
		/**
		 * Notification that a body has been removed from the world.
		 * Default implementation does nothing.
		 */
		virtual void bodyRemoved (Body* body) {}
		
		/**
		 * Convenience method for implementation classes.
		 * Returns true if we need to consider these two bodies for contact generation
		 * and they are potentially colliding.
		 */
		static inline bool test (const Body* a, const Body* b) {
			if (ignore(a, b)) {
				return false;
			}
			
			typename Body::joint_citerator it;
			
			for (it = a->joints.begin(); it != a->joints.end(); it++) {
				Joint<d>* j = *it;
				
				if (j->body1 == b || j->body2 == b) {
					if (j->ignoreContacts) {
						return false;
					}
				}
			}
			
			// TODO: make AABBs viable
			return testRadius(a, b);
		}
		
		/**
		 * Convenience method for implementation classes.
		 * Returns true if we don't need to consider these two bodies for contact generation.
		 */
		static inline bool ignore (const Body* a, const Body* b) {
			if (Globals::findSleepingContacts) {
				return (a->isStatic() && b->isStatic());
			} else {
				// If they are static they are also considered to be asleep.
				return (! a->isAwake() && ! b->isAwake());
			}
		}
		
		/**
		 * Convenience method for implementation classes.
		 * Returns true if the bounding radii for these bodies are overlapping.
		 */
		static inline bool testRadius (const Body* a, const Body* b, const float _near = Globals::_NEAR) {
			const float r = a->getBoundingRadius() + b->getBoundingRadius() + _near;
			
			return (a->getPosition().distanceSquared(b->getPosition()) <= r * r);
		}
		
		/**
		 * Convenience method for implementation classes.
		 * Returns true if the axis-aligned bounding boxes for these bodies are overlapping.
		 */
		static inline bool testAABB (const Body* a, const Body* b, const float _near = Globals::_NEAR) {
			return a->getAABB().intersects(b->getAABB(), _near);
		}
		
		World* getWorld () const { return m_world; }
		
		protected:
		
		World* m_world;
	};
	
	typedef BroadPhase<TwoDee> BroadPhase2D;
	typedef BroadPhase<ThreeDee> BroadPhase3D;
	typedef BroadPhase<TwoDee> NaiveBroadPhase2D;
	typedef BroadPhase<ThreeDee> NaiveBroadPhase3D;
	
	/** Sort and sweep broad-phase implementation. */
	template <typename d>
	class SortAndSweep : public BroadPhase<d> {
		public:
		
		USE_DIMENSION(d);
		
		typedef std::pair<Body*, Body*> BodyPair;
		
		typedef typename BroadPhase<d>::output_container output_container;
		
		SortAndSweep () {}
		
		virtual output_container findPairs ();
		
		virtual void bodyAdded (Body* body);
		virtual void bodyRemoved (Body* body);
				
		private:
		
			static bool comparator(const Body* body1, const Body* body2);

			std::vector<Body*> xlist;
			std::vector<Body*> ylist;
		
	};
};

#endif

