#include "broadphase_quadtree.h"
#include <boost/foreach.hpp>
#include "body.h"

using namespace physics;

// Keeping the quadtree up to date isn't completely implemented yet
// -- so we just throw it away and rebuild it every frame
#define REBUILD_EVERY_FRAME 1

/**
 * Set up an octree with a root node.
 */
void QuadtreeBroadPhase::init(World* world) {
	
	// Create a tree.
	tree = new Quadtree;
	
	// Create a root node.
	Node* root = new Node;
	root->key = 1;
	root->hasChild = 0;
	root->depth = 1;
	tree->nodes[1] = root;
	
	// TODO: make relevant/settable
	
	//root->centre = Vector(19*6, 50*6);
	//tree->width = 19 * 12;
	//tree->height = 50 * 12;
	//tree->levels = 6;
	
	
	root->centre = Vector(0, 50);
	tree->width = 200;
	tree->height = 100;
	tree->levels = 3;
	


	// Store the world.
	BroadPhase<TwoDee>::init(world);
}

/**
 * Find the pairs of bodies.
 */
BroadPhase<TwoDee>::output_container QuadtreeBroadPhase::findPairs () {
	
	#if REBUILD_EVERY_FRAME
	
	typedef std::pair<const int, Node*> pair;
	
	BOOST_FOREACH(pair pairing, tree->nodes) {
		delete pairing.second;
	}
	
	delete tree;
	
	init(getWorld());
	
	locations.clear();
	
	BOOST_FOREACH(Body* body, getWorld()->bodies) {
		bodyAdded(body);
	}
	
	#else
	
	// Iterate over all bodies in the world.
	BOOST_FOREACH(Body* body, getWorld()->bodies) {
	
		// Check if the body is in a different octant.
		if (body->isAwake()) {
		
			unsigned int oldKey = locations[body];
			unsigned int newKey = generateMortonKey(body->getPosition());
			int depth = tree->nodes[oldKey]->depth;
			oldKey = truncate(depth, oldKey);
			newKey = truncate(depth, newKey);
			
			// TODO: calculate newKey correctly (could now be straddling when it wasn't before)
		
			if (oldKey != newKey) {
				tree->nodes[oldKey]->bodies.remove(body);
				float r = body->getBoundingRadius();
				insert(tree->nodes[1], body, body->getPosition() - Vector3D(r, r, r), body->getPosition() + Vector3D(r, r, r));
			}
			
		}
		
	}
	
	#endif
	
	output_container bodies;
	
	// Recursve over all nodes in the tree, starting at the root.
	buildPairs(tree->nodes[1], bodies);
	
	return bodies;
	
}

/**
 * Recursively build the list.
 */
void QuadtreeBroadPhase::buildPairs (Node* node, output_container& list)
{
  // Keep track of all ancestor object lists in a stack.
  const int MAX_DEPTH = 20;
  static Node* ancestorStack[MAX_DEPTH];
  static int depth = 0;
  
  // Check collisions between all objects on this level and ancestor objects.
  ancestorStack[depth++] = node;
  for (int n = 0; n < depth; n++) {
		
		typedef std::list<Body*>::iterator iterator;
		
		for (iterator i = ancestorStack[n]->bodies.begin(); i != ancestorStack[n]->bodies.end(); i++) {
		
			Body* pA = *i;
			
			iterator j;
			
			if (ancestorStack[n] == node) {
				// If we're iterating over the same list twice, only need to consider the remaining bodies
				j = i;
				j++;
			} else {
				// Otherwise start at the beginning of the list as usual
				j = node->bodies.begin();
			}
		
			for (; j != node->bodies.end(); j++) {
				Body* pB = *j;
				
				if (test(pA, pB)) {
					list.push_back(BodyPair(pA, pB));
				}
			}
		}
  }
    
  // Recursively visit all existing children.
  for (int index = 0; index < 8; index++) {
  
  	// If the child node exists, recurse to it.
  	if ( node->hasChild & (1 << index) ) {
  		 unsigned int key = (node->key << 3) | index;
  		 buildPairs(tree->nodes[key], list);
  	}
  	
  }
    
	// Remove current node from ancestor stack before returning.
	depth--;
}

/**
 * Add a body into the Octree.
 */
void QuadtreeBroadPhase::bodyAdded (Body* body) {

	// Recurse over the tree, starting at the root.
	float r = body->getBoundingRadius();
	insert(tree->nodes[1], body, body->getPosition() - Vector3D(r, r, r), body->getPosition() + Vector3D(r, r, r));

}

/**
 * Recursively insert a body into the Octree.
 *
 * @param node Current node.
 * @param body Body to insert.
 * @param low Low point of bounding box.
 * @param high High point of bounding box.
 * @param depth Depth of current node.
 * @return Morton key?
 */
unsigned int QuadtreeBroadPhase::insert(Node* node, Body* body, const Vector& low, const Vector& high, int depth){
	
	int index = 0;
	bool straddle = false;
	unsigned int key = 0;
	
	// Calculate the Octant holding the body.
	if ( (low.x < node->centre.x && high.x >= node->centre.x && tree->width != 0) || 
		(low.y < node->centre.y && high.y >= node->centre.y && tree->height != 0) ) {
			
		straddle=true;
		
	} else {
		if (low.x >= node->centre.x && tree->width != 0) {
			index |= 1;
		}
		if (low.y >= node->centre.y && tree->height!= 0) {
			index |= 2;
		}
	}
	
	// If not straddling, then add to child.
	if( !straddle && (depth <= tree->levels) ){
	
		key = (node->key << 3) | index;
		
		// If child doesn't exist, create one.
		if( !(node->hasChild & (1 << index)) ){
			Node *child = new Node;
			child->key = key;
			child->hasChild = 0;
			child->depth = node->depth + 1;
			
			// Update current node.
			node->hasChild |= (1 << index);
			
			// Update the centre point of the child.
			float rx = tree->width  / (1 << (depth + 2));
			float ry = tree->height / (1 << (depth + 2));
			
			child->centre.x = node->centre.x + (index&1? rx: -rx);
			child->centre.y = node->centre.y + (index&2? ry: -ry);
			
			// Add the child node to the tree.
			tree->nodes[key] = child;
			
		}
		
		// Recursively add to the child.
		insert(tree->nodes[key], body, low, high, depth + 1);
		
	}
	
	// If straddling octants, or at bottom of tree, just add to this node.
	else {
		node->bodies.push_front(body);
		key = node->key;
		
		// Save the key in the locations map.
		locations[body] = key;
	}
	
	return key;	
}


/**
 * Remove a body from the Octree.
 */
void QuadtreeBroadPhase::bodyRemoved (Body* body) {
	
	remove(body);

}

void QuadtreeBroadPhase::remove (Body* body) {
	
	unsigned int key = locations[body];
	tree->nodes[key]->bodies.remove(body);
	locations.erase(body);
}

/**
 * Generate a hashcode for the vector supplied.
 */
unsigned int QuadtreeBroadPhase::generateMortonKey(const Vector& v){
	// Fractions, defined to save time later.
	float fx = 0.5, fy = 0.5, fz = 0.5;
	if (tree->width != 0) {
		fx = (v.x + tree->width/2) / tree->width;
	}
	if (tree->height != 0) {
		fy = (v.y + tree->height/2) / tree->height;
	}
	
	// Convert to fixed point number.
	unsigned int x = (unsigned int) (fx * 0x400);
	unsigned int y = (unsigned int) (fy * 0x400);
	unsigned int z = (unsigned int) (fz * 0x400);
	
	//Actually create morton key.
	x = (x | (x<<16)) & 0xff0000ff;
	x = (x | (x<< 8)) & 0x0300f00f;
	x = (x | (x<< 4)) & 0x030c30c3;
	x = (x | (x<< 2)) & 0x09249249;
	
	y = (y | (y<<16)) & 0xff0000ff;
	y = (y | (y<< 8)) & 0x0300f00f;
	y = (y | (y<< 4)) & 0x030c30c3;
	y = (y | (y<< 2)) & 0x09249249;
	
	z = (z | (z<<16)) & 0xff0000ff;
	z = (z | (z<< 8)) & 0x0300f00f;
	z = (z | (z<< 4)) & 0x030c30c3;
	z = (z | (z<< 2)) & 0x09249249;

	// Add sentinel bit and combine key.
	unsigned int key =(1<<30) + (z<<2) + (y<<1) + x;
		
	return key;
}

/**
 * Move the provided key to its new depth.
 */
unsigned int QuadtreeBroadPhase::truncate(int depth, unsigned int key){
	int cdepth = calcDepth(key);
	unsigned int k = key;
	
	// Because of the way Morton keys work, a bitshift moves the key to another level. 
	if(cdepth > depth) {
		k = (key >> ((cdepth - depth) * 3));
	}
	
	return k;
}

/**
 * Calculate the depth of a Morton key.
 */
int QuadtreeBroadPhase::calcDepth(unsigned int key){
	
	int depth = 1;
	unsigned int tmp = key;
	
	// Keep bitshifting by three whilst we can.
	while(tmp > 1){
		tmp = tmp >> 3;
		depth++;	
	}
	
	return depth;
	
}
