#ifndef _PHYSICS_CAPSULE_
#define _PHYSICS_CAPSULE_

#include "common.h"

#include "area.h"
#include "vector.h"
#include "aarect.h"

namespace physics {
	/*!
	*	Provides the capsule shape in the 2D simulation
	*/
	class Capsule2D : public Area
	{
		public:
		//! Constructor
		Capsule2D (Vector2D p1, Vector2D p2, float r) :
			Area((p1 + p2)/2),
			m_p1(p1), m_p2(p2),
			m_local_p1(p1 - getPosition()),
			m_local_p2(p2 - getPosition()),
			m_radius(r) {}
		
		bool contains (const Vector2D &p) const { return distanceSq(p) < getRadiusSq(); }
		bool intersects (const physics::AARect& rect) const;
		
		float getRadius () const { return m_radius; }
		
		//! Returns x coordinate of the first position
		float x1 () const { return m_p1.x; }
		//! Returns y coordinate of the first position
		float y1 () const { return m_p1.y; }
		//! Returns x coordinate of the second position
		float x2 () const { return m_p2.x; }
		//! Returns y coordinate of the second position
		float y2 () const { return m_p2.y; }
		
		//! Return the first position
		Vector2D getP1 () const { return m_p1; }
		//! Return the second position
		Vector2D getP2 () const { return m_p2; }
		
		//! Get method for the length of the straight segment
		/*! \return Float; Length of the straight segment
		*/
		float getLength () const { return m_local_p1.distance(m_local_p2); }
		//! Get method for the length of the straight segment squared
		/*! \return Float; Length of the straight segment squared
		*/
		float getLengthSq () const { return m_local_p1.distanceSquared(m_local_p2); }
		
		Vector2D getNormal () const;
		Vector2D getNormal (int i) const;
		Vector2D getVertex (int i) const { return (i % 2 == 0) ? m_p1 : m_p2; }
		
		int getVertexCount () const { return 2; }
		
		Vector2D getExtremePoint (const Vector2D& dir, int& i) const;
		
		float getArea () const { return m_radius * getLength() + Globals::PI * getRadiusSq(); }
		
		//! Get the Moment of Inertia
		/* \return Float; Moment of Inertia
		*/
		float getMOI () const { return getMass() * (getLengthSq() * 4 + getRadiusSq() * 7) / 12; }
		
		Type type () const { return CAPSULE; }
		
		private:
		
		// These act on the line segment and ignore the radius
		
		float distance (const Vector2D &p) const { return sqrt(distanceSq(p)); }
		float distanceSq (const Vector2D &p) const;
		
		virtual void thisMoved ();
		
		Vector2D m_p1;
		Vector2D m_p2;
		
		Vector2D m_local_p1;
		Vector2D m_local_p2;
		
		float m_radius;
		
	};
};

inline physics::Vector2D physics::Capsule2D::getNormal () const {
	Vector2D normal = m_p2 - m_p1;
	
	normal = Vector2D(-normal.y, normal.x);
	
	return normal.normalise();
}

inline physics::Vector2D physics::Capsule2D::getNormal (int i) const {
	Vector2D normal = getNormal();
	
	if (i % 2 != 0) { normal *= -1; }
	
	return normal;
}

inline physics::Vector2D physics::Capsule2D::getExtremePoint (const physics::Vector2D& dir, int& i) const {
	Vector2D p;
	
	if (dir.dot(m_p2 - m_p1) > 0) {
		i = 1;
		p = m_p2;
	} else {
		i = 0;
		p = m_p1;
	}
	
	p += m_radius * dir;
	
	return p;
}

inline void physics::Capsule2D::thisMoved () {
	m_p1 = localToWorld(m_local_p1);
	m_p2 = localToWorld(m_local_p2);
}

// Following algorithm from Real-Time Collision Detection by Christer Ericson

inline float physics::Capsule2D::distanceSq (const physics::Vector2D& c) const {
	Vector2D ab = m_p2 - m_p1;
	Vector2D ac = c - m_p1;
	Vector2D bc = c - m_p2;

	float e = ac.dot(ab);

	// Handle cases where c projects outside ab
	if (e <= 0.0f) return ac.dot(ac);
	float f = ab.dot(ab);
	if (e >= f) return bc.dot(bc);
	
	// Handle case where c projects onto ab
	return ac.dot(ac) - e * e / f;
}

inline bool physics::Capsule2D::intersects (const physics::AARect& rect) const {

	if (rect.contains(Capsule2D::getP1())) return 1;
	if (rect.contains(Capsule2D::getP2())) return 1;

	if (rect.contains(Capsule2D::getPosition())) return 1;

	return 0;
}

#endif

