#ifndef _PHYSICS_CHILD_OBJECT_
#define _PHYSICS_CHILD_OBJECT_

#include "common.h"

#include "vector.h"
#include "orientation.h"

namespace physics {
	
	/**
	 * ChildObject acts as a frame of reference: it has a position and orientation relative to its parent.
	 */
	template <typename d, typename Parent>
	class ChildObject {
		public:
		
		USE_DIMENSION(d);
		
		ChildObject (const Vector& p = Vector(), const Orientation& o = Orientation())
			: m_localPosition(p), m_worldPosition(p), m_localOrientation(o), m_worldOrientation(o), parent(0) {}
		
		virtual ~ChildObject () {}
		
		/** Get position in world coordinates. */
		Vector getPosition () const { return m_worldPosition; }
		
		/** Set position in world coordinates. */
		void setPosition (const Vector& p);
		
		/** Get position in local coordinates (relative to parent). */
		Vector getLocalPosition () const { return m_localPosition; }
		
		/** Set position in local coordinates (relative to parent). */
		void setLocalPosition (const Vector& p);
		
		/** Get orientation relative to world. */
		Orientation getOrientation () const { return m_worldOrientation; }
		
		/** Set orientation relative to world. */
		void setOrientation (const Orientation& o);
		
		/** Get orientation relative to parent. */
		Orientation getLocalOrientation () const { return m_localOrientation; }
		
		/** Set orientation relative to parent. */
		void setLocalOrientation (const Orientation& o);
		
		void translate (const Vector& v);
		void localTranslate (const Vector& v);
		
		/** Translate local position while leaving world position unchanged. */
		void localTranslateInPlace (const Vector& v) { m_localPosition += v; }
		
		void rotate (Rotation r);
		
		void move (const Vector& translation, Rotation r);
		
		Parent* getParent () const { return parent; }
		
		Vector worldToLocal (const Vector& v) const;
		Vector localToWorld (const Vector& v) const;
		
		Vector worldToLocalDir (const Vector& v) const;
		Vector localToWorldDir (const Vector& v) const;
		
		void parentMoved ();
		
		void setParent (Parent* p);
		
		protected:
		
		virtual void thisMoved () {}
		
		private:
		
		void updateWorldPosition ();
		void updateWorldOrientation ();
		
		Vector m_localPosition;
		Vector m_worldPosition;
		
		Orientation m_localOrientation;
		Orientation m_worldOrientation;
		
		Parent* parent;
	};
};

// Movement methods

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::setPosition (const Vector& p) {
	if (! parent) {
		m_localPosition = p;
	} else {
		m_localPosition = parent->worldToLocal(p);
	}
	
	m_worldPosition = p;
	
	thisMoved();
}

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::setLocalPosition (const Vector& p) {
	m_localPosition = p;
	
	updateWorldPosition();
	
	thisMoved();
}

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::translate (const Vector& v) {
	if (! parent) {
		m_localPosition += v;
		m_worldPosition = m_localPosition;
	} else {
		m_localPosition += parent->getOrientation().inverseTransform(v);
		m_worldPosition += v;
	}
	
	thisMoved();
}

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::localTranslate (const Vector& v) {
	m_localPosition += v;
	
	updateWorldPosition();
	
	thisMoved();
}

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::setOrientation (const Orientation& o) {
	if (! parent) {
		m_localOrientation = o;
	} else {
		m_localOrientation = o - parent->getOrientation();
	}
	
	m_worldOrientation = o;
	
	thisMoved();
}

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::setLocalOrientation (const Orientation& o) {
	m_localOrientation = o;
	
	updateWorldOrientation();
	
	thisMoved();
}

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::rotate (Rotation r) {
	m_localOrientation += r;
	
	updateWorldOrientation();
	
	thisMoved();
}

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::setParent (Parent* p) {
	parent = p;
	
	if (parent) {
		m_localPosition = parent->worldToLocal(m_worldPosition);
		m_localOrientation = m_worldOrientation - parent->getOrientation();
	} else {
		m_localPosition = m_worldPosition;
		m_localOrientation = m_worldOrientation;
	}
}

// Query methods

template <typename d, typename Parent>
inline typename d::Vector physics::ChildObject<d, Parent>::localToWorld (const Vector& p) const {
	return getPosition() + getOrientation().transform(p);
}

template <typename d, typename Parent>
inline typename d::Vector physics::ChildObject<d, Parent>::worldToLocal (const Vector& p) const {
	return getOrientation().inverseTransform(p - getPosition());
}

template <typename d, typename Parent>
inline typename d::Vector physics::ChildObject<d, Parent>::localToWorldDir (const Vector& p) const {
	return getOrientation().transform(p);
}

template <typename d, typename Parent>
inline typename d::Vector physics::ChildObject<d, Parent>::worldToLocalDir (const Vector& p) const {
	return getOrientation().inverseTransform(p);
}

// Update cached data

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::parentMoved () {
	updateWorldPosition();
	updateWorldOrientation();
	thisMoved();
}

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::updateWorldPosition () {
	if (parent) {
		m_worldPosition = parent->localToWorld(m_localPosition);
	} else {
		m_worldPosition = m_localPosition;
	}
}

template <typename d, typename Parent>
inline void physics::ChildObject<d, Parent>::updateWorldOrientation () {
	if (parent) {
		m_worldOrientation = parent->getOrientation() + m_localOrientation;
	} else {
		m_worldOrientation = m_localOrientation;
	}
}

#endif

