#ifndef _PHYSICS_CIRCLE_
#define _PHYSICS_CIRCLE_

#include "common.h"

#include "area.h"
#include "aarect.h"
#include "vector.h"

namespace physics {
	/*! \class Circle
	*	Provides the circle shape in the 2D simulation
	*/
	class Circle : public Area
	{
		public:
		
		Circle () : m_radius(0) {}
		explicit Circle (float r) : m_radius(r) {}
		Circle (const Vector2D& p, float r) : Area(p), m_radius(r) {}
		
		bool contains (const Vector2D &p) const;
		bool intersects (const AARect& rect) const;
		
		float getRadius () const { return m_radius; }
		void setRadius (float f) { m_radius = f; }
		
		Vector2D getNormal (int i) const { return Vector2D(); }
		Vector2D getVertex (int i) const { return getPosition(); }
		
		int getVertexCount () const { return 1; }
		
		Vector2D getExtremePoint (const Vector2D& dir, int& i) const;
		
		virtual float getMinX () const { return getPosition().x - m_radius; }
		virtual float getMaxX () const { return getPosition().x + m_radius; }
		virtual float getMinY () const { return getPosition().y - m_radius; }
		virtual float getMaxY () const { return getPosition().y + m_radius; }
		
		float getArea () const { return Globals::PI * getRadiusSq(); }
		
		//! Get the Moment of Intertia
		/*! \return Float; Moment of Inertia
		*/
		float getMOI () const { return getMass() * getRadiusSq() / 2; }
				
		Type type () const { return CIRCLE; }
		
		private:
		float m_radius;
	};
};

inline bool physics::Circle::contains (const physics::Vector2D &p) const {
	return (p.distanceSquared(getPosition()) <= getRadiusSq());
}

inline physics::Vector2D physics::Circle::getExtremePoint (const physics::Vector2D& dir, int& i) const {
	i = 0;
	
	return getPosition() + dir * m_radius;
}

//template <typename d>
//inline bool physics::Body<d>::intersects (const physics::AARect& rect) const {
inline bool physics::Circle::intersects (const physics::AARect& rect) const {

	//Graphics Gems. By Andrew S. Glassner
	//Pages 52 and 53
	
	//Translate co-ordinates placing centre of circle at origin
	float circlePosX = Circle::getPosition().x;
	float circlePosY = Circle::getPosition().y;
	float rMaxX = rect.getMaxX() - circlePosX;
	float rMaxY = rect.getMaxY() - circlePosY;
	float rMinX = rect.getMinX() - circlePosX;
	float rMinY = rect.getMinY() - circlePosY;
	
	float r = Circle::getRadius();
	
	if (rMaxX < 0) { //rectangle to left of circle centre
		if (rMaxY < 0) { //lower left corner
			return (rMaxX*rMaxX + rMaxY*rMaxY < r*r);
		} 
		else if (rMinY > 0) { //rectangle in upper left corner
			return (rMaxX*rMaxX + rMinY*rMinY < r*r);
		}
		else {
			return (fabs(rMaxX) < r);
		}
	}
	else if (rMinX > 0) {//rectangle to right of circle centre
		if (rMaxY < 0) { //rectangle in lower right corner
			return (rMinX*rMinX + rMaxY*rMaxY < r*r);
		}
		else if (rMinY > 0) { //rectangle in upper right
			return (rMinX*rMinX + rMinY*rMinY < r*r);
		}
		else { //rectangle due EAST of circle
			return (rMinX < r);
		}
	}
	else { //rectangle on circle vertical centre line
		if (rMaxY < 0) { //r due south
			return (fabs(rMaxY) < r);
		}
		else if (rMinY > 0) { //rectangle due north
			return (rMinY < r);
		}
		else { //rectangle contains circle centre point
			return 1;
		}
	}
	
	return 0;
}

#endif
