#include "contact.h"
#include "vector.h"
#include "body.h"
#include "shape.h"

#include <algorithm>

using namespace physics;

template <typename d>
void Contact<d>::init (Shape* shape1, Shape* shape2)
{
	shape[0] = shape1;
	shape[1] = shape2;
	
	body[0] = shape[0]->getBody();
	body[1] = shape[1]->getBody();
	
	offset[0] = position - body[0]->getPosition();
	offset[1] = position - body[1]->getPosition();
	
	contactToWorld = Matrix::createBasis(normal);
	
	calculateMass();
	calculateVelocities();
}

/** The idea is that rather than generating contacts fresh each frame, 
you keep them around and update them for the new positions/velocities
which should be faster to calculate */

// TODO: 3D-ness
template <typename d>
void Contact<d>::calculatePenetration ()
{
	if (invalid()) {
		// TODO: better way of marking it as invalid?
		setPenetration(-500);
		return;
	}
	
	Vector p[2];
	
	p[0] = shape[0]->getVertex(featureID[0]);
	p[1] = shape[1]->getVertex(featureID[1]);
	
	if (isEdge[0]) {
		normal = shape[0]->getNormal(featureID[0]);
	} else if (isEdge[1]) {
		normal = -shape[1]->getNormal(featureID[1]);
	} else {
		// circle-circle or vertex-circle
		normal = (p[1] - p[0]).normalise();
	}

	p[0] += normal * shape[0]->getRadius();
	p[1] -= normal * shape[1]->getRadius();
	
	setPenetration(normal.dot(p[0] - p[1]));
	
	if (isEdge[0]) {
		position = p[1] + normal * penetration / 2;
	} else if (isEdge[1]) {
		position = p[0] - normal * penetration / 2;
	} else {
		position = (p[0] + p[1]) / 2;
	}
	
	offset[0] = position - body[0]->getPosition();
	offset[1] = position - body[1]->getPosition();
	
	contactToWorld = Matrix::createBasis(normal);
}

template <typename d>
void Contact<d>::calculateMass () {
	
	inverseMass[0] = body[0]->getInverseMass();
	inverseMass[1] = body[1]->getInverseMass();
	
	// velocityPerImpulse is a matrix which given an impulse will give you
	// the change of velocity that impulse would cause at the contact point.
	velocityPerImpulse = Matrix();
	
	for (int i = 0; i < 2; i++) {
		{//if (body[i]->isAwake()) {
			velocityPerImpulse += Matrix::identity() * inverseMass[i] + body[i]->getMOI().getVelocityPerImpulse(offset[i]);
		}
	}
	
	// Convert velocityPerImpulse from world coordinates to contact coordinates
	velocityPerImpulse = contactToWorld.getTranspose() * velocityPerImpulse * contactToWorld;
	
	// This case should always be true
	if (velocityPerImpulse != Matrix()) {
		impulsePerVelocity = velocityPerImpulse.getInverse();
		
		massNormal = 1.0f / (velocityPerImpulse * Vector(1, 0)).x;
	} else {
	// But sometimes contactToWorld is inexplicably 0 and we need to take that into account.
		impulsePerVelocity = Matrix();
		
		massNormal = 0;
	}
}

// TODO: 3D-ness
template <typename d>
bool Contact<d>::invalid ()
{
	int e; // which shape contains the edge?
	int v; // which shape contains the vertex?
	
	if (isEdge[0]) {
		e = 0;
		v = 1;
	} else if (isEdge[1]) {
		e = 1;
		v = 0;
	} else {
		return false;
	}
	
	Vector vertex = shape[v]->getVertex(featureID[v]);
	
	Vector a = shape[e]->getVertex(featureID[e]);
	Vector b = shape[e]->getVertex(featureID[e] + 1);
	
	float t = (vertex - a).dot(b - a) / a.distanceSquared(b);
	
	return (t < 0 || t > 1);
}

template <typename d>
typename d::Vector Contact<d>::getRelativeContactVelocity ()
{
	Vector v1 = body[0]->getVelocityAtPoint(position);
	Vector v2 = body[1]->getVelocityAtPoint(position);
	
	return contactToWorld.getTranspose() * (v2 - v1);
}

template <typename d>
void Contact<d>::calculateVelocities ()
{
	if (penetration < 0) {
		approachSpeed = -10;
	} else {
		relativeVelocity = getRelativeContactVelocity();
		approachSpeed = -relativeVelocity.x; //relativeVelocity.dot(normal);
	}
}

template class Contact<TwoDee>;
template class Contact<ThreeDee>;

