#include "contact_generator.h"
#include "contact.h"


#include "vector.h"
#include "body.h"
#include "shape.h"

#include "sphere.h"
#include "box.h"

#include <boost/foreach.hpp>

using namespace physics;
using namespace std;

float min (const Vector2D& axis, Area* area, Vector2D& point, int& index)
{
	point = area->getExtremePoint(-axis, index);
	return axis.dot(point);
}

float max (const Vector2D& axis, Area* area, Vector2D& point, int& index)
{
	point = area->getExtremePoint(axis, index);
	return axis.dot(point);
}

template <typename d>
typename ContactGenerator<d>::output_container ContactGenerator<d>::findContacts (const input_container& pairs) {
	m_contacts.clear();
	
	BOOST_FOREACH(const BodyPair& pair, pairs) {
		check(pair.first, pair.second);
	}
	
	
	return m_contacts;
}

template <typename d>
void ContactGenerator<d>::check (Body* b1, Body* b2) {
	BOOST_FOREACH(Shape* shape1, b1->getShapes()) {
		BOOST_FOREACH(Shape* shape2, b2->getShapes()) {
			check(shape1, shape2);
		}
	}
}

template <>
bool ContactGenerator<TwoDee>::check (Area* area1, Area* area2)
{
	int vcount_area1 = area1->getVertexCount();
	int vcount_area2 = area2->getVertexCount();
	if (vcount_area1 == 0 || vcount_area2 == 0) { return false; }
	
	const bool findMultiple = Globals::findMultipleContacts;
	
	Contact contacts[2];
	
	int best = 0;
	int other = 1;
	
	contacts[0].penetration = 1000000;
	contacts[1].penetration = 1000000;
	
	// for each edge of body 1, check closest vertex of body 2
	for (int edgeID = 0; edgeID < vcount_area1; edgeID++)
	{
		Vector2D axis = area1->getNormal(edgeID);
		
		if (axis == Vector2D()) { continue; }
		
		float max1 = axis.dot(area1->getVertex(edgeID)) + area1->getRadius();
		
		int vertexID; //set to the vertex id of the vertex which is furthest out in whatever direction
		
		Vector2D point;
		
		float min2 = min(axis, area2, point, vertexID);
		
		float overlap = max1 - min2;
		
		if (overlap < -Globals::_NEAR) { return false; }
		
		int here;
		
		if (overlap < contacts[best].penetration)
		{
			if (findMultiple) {
				here = other;
				other = best;
				best = here;
			} else {
				here = best;
			}
		} else if (findMultiple && overlap < contacts[other].penetration) {
			here = other;
		} else {
			continue;
		}
		
		contacts[here].penetration = overlap;
		contacts[here].normal = axis;
		contacts[here].position = point;
		contacts[here].isEdge[0] = true;
		contacts[here].isEdge[1] = false;
		contacts[here].featureID[0] = edgeID;
		contacts[here].featureID[1] = vertexID;
	}
	
	// for each edge of body 2, check closest vertex of body 1
	for (int edgeID = 0; edgeID < vcount_area2; edgeID++)
	{
		Vector2D axis = -area2->getNormal(edgeID);
		
		if (axis == Vector2D()) { continue; }
		
		int vertexID;
		
		Vector2D point;
		
		float max1 = max(axis, area1, point, vertexID);
		float min2 = axis.dot(area2->getVertex(edgeID)) - area2->getRadius();
		
		float overlap = max1 - min2;
		
		if (overlap < -Globals::_NEAR) { return false; }
		
		int here;
		
		if (overlap < contacts[best].penetration)
		{
			if (findMultiple) {
				here = other;
				other = best;
				best = here;
			} else {
				here = best;
			}
		} else if (findMultiple && overlap < contacts[other].penetration) {
			here = other;
		} else {
			continue;
		}
		
		contacts[here].penetration = overlap;
		contacts[here].normal = axis;
		contacts[here].position = point;
		contacts[here].isEdge[0] = false;
		contacts[here].isEdge[1] = true;
		contacts[here].featureID[0] = vertexID;
		contacts[here].featureID[1] = edgeID;
	}
	
	float r = area1->getRadius() + area2->getRadius();
	
	// Check vertex-vertex collisions
	if (r > 0)
	{
		Vector2D axis = area2->getPosition() - area1->getPosition();
		
		axis = axis.normalise();
		
		// Find closest pair of vertices
		int vertex1, vertex2;
		
		area1->getExtremePoint(axis, vertex1);
		area2->getExtremePoint(-axis, vertex2);
		
		// Get normal separating these vertices
		axis = area2->getVertex(vertex2) - area1->getVertex(vertex1);
		
		axis = axis.normalise();
		
		Vector2D point;
		
		// Find min/max extent of shapes in this direction
		float max1 = max(axis, area1, point, vertex1);
		float min2 = min(axis, area2, point, vertex2);
		
		float overlap = max1 - min2;
		
		if (overlap < -Globals::_NEAR) { return false; }
		
		int here = -1;
		
		if (overlap < contacts[best].penetration)
		{
			if (findMultiple) {
				here = other;
				other = best;
				best = here;
			} else {
				here = best;
			}
		} else if (findMultiple && overlap < contacts[other].penetration) {
			here = other;
		}
		
		if (here != -1) {
			contacts[here].penetration = overlap;
			contacts[here].normal = axis;
			contacts[here].position = area1->getVertex(vertex1) + axis * (area1->getRadius() - overlap / 2);
			contacts[here].isEdge[0] = false;
			contacts[here].isEdge[1] = false;
			contacts[here].featureID[0] = vertex1;
			contacts[here].featureID[1] = vertex2;
		}
	}
	
	for (int i = 0; i < 2; i++) {
		Contact& contact = contacts[i];
		
		if (i != best) {
			if (! findMultiple) { continue; }
			
			if (contact.penetration > contacts[best].penetration + Globals::_NEAR) { continue; }
			
			if (contacts[best].normal.dot(contact.normal) < 0.9f) { continue; }
		}
		
		contact.init(area1, area2);
		
		if (! contact.invalid()) {
			Contact* newContact = new Contact(contact);
			
			m_contacts.push_back(newContact);
		}
		
		if (findMultiple) {
			// Look for adjacent features that might also be valid contact points
			if (contact.isEdge[0]) {
				int i = contact.featureID[1] - 1;
				if (i < 0) { i += vcount_area2; }
				
				checkAdjacent(contact, 1, i);
				
				i = contact.featureID[1] + 1;
				if (i >= vcount_area2) { i -= vcount_area2; }
				
				checkAdjacent(contact, 1, i);
			} else if (contact.isEdge[1]) {
				int i = contact.featureID[0] - 1;
				if (i < 0) { i += vcount_area1; }
				
				checkAdjacent(contact, 0, i);
				
				i = contact.featureID[0] + 1;
				if (i >= vcount_area1) { i -= vcount_area1; }
				
				checkAdjacent(contact, 0, i);
			}
		}
	}
	
	return true;
}

template <typename d>
void ContactGenerator<d>::checkAdjacent (Contact& c, int shape, int feature) {
	int id[2];
	id[0] = c.featureID[0];
	id[1] = c.featureID[1];
	id[shape] = feature;
	
	float overlap = c.normal.dot(c.shape0->getVertex(id[0]) - c.shape1->getVertex(id[1])) + c.shape0->getRadius() + c.shape1->getRadius();
	
	if (overlap < c.penetration + Globals::_NEAR && overlap > -Globals::_NEAR) {
		Contact* newContact = new Contact(c);
		
		newContact->featureID[shape] = id[shape];
		newContact->penetration = overlap;
		newContact->position = c.shape[shape]->getVertex(id[shape]);
		
		if (newContact->invalid()) {
			delete newContact;
		} else {
			m_contacts.push_back(newContact);
		}
	}
}

template <>
bool ContactGenerator<ThreeDee>::check (Volume* shape1, Volume* shape2) {
	switch (shape1->type()) {
		case Volume::SPHERE:
			switch (shape2->type()) {
				case Volume::SPHERE:
					return checkSphereSphere(static_cast<Sphere*>(shape1), static_cast<Sphere*>(shape2));
				case Volume::BOX:
					return checkSphereBox(static_cast<Sphere*>(shape1), static_cast<Box*>(shape2));
				default:
					return false;
			}
		case Volume::BOX:
			switch (shape2->type()) {
				case Volume::SPHERE:
					return checkBoxSphere(static_cast<Box*>(shape1), static_cast<Sphere*>(shape2));
				case Volume::BOX:
					return checkBoxBox(static_cast<Box*>(shape1), static_cast<Box*>(shape2));
				default:
					return false;
			}
		default:
			return false;
	}
	
	return false;
}

template <>
bool ContactGenerator<ThreeDee>::checkSphereSphere (Sphere* shape1, Sphere* shape2)
{
	float r = shape1->getRadius() + shape2->getRadius();
	
	Vector offset = shape2->getPosition() - shape1->getPosition();
	
	float overlap = r - offset.length();
	
	if (overlap < -Globals::_NEAR) { return false; }
	
	Contact contact;
	contact.penetration = overlap;
	contact.normal = offset.normalise();
	contact.position = shape1->getPosition() + contact.normal * (shape1->getRadius() - overlap / 2);
	contact.isEdge[0] = false;
	contact.isEdge[1] = false;
	contact.featureID[0] = 0;
	contact.featureID[1] = 0;
	
	contact.init(shape1, shape2);
	
	Contact* newContact = new Contact(contact);
	
	m_contacts.push_back(newContact);
	
	return true;
}

template <>
bool ContactGenerator<ThreeDee>::checkBoxSphere (Box* box, Sphere* sphere, bool flip)
{
	const float r = sphere->getRadius();
	
	const float rn = r + Globals::_NEAR;
	
	Vector p = box->worldToLocal(sphere->getPosition());
	
	Vector size = box->getSize();
	
	p += size / 2;
	
	if (p.x < -rn || p.x > size.x + rn || p.y < -rn || p.y > size.y + rn || p.z < -rn || p.z > size.z + rn) {
		return false;
	}
	
	Vector q = p;
	
	for (int i = 0; i < 3; i++) {
		if (p[i] < 0) {
			q[i] = 0;
		} else if (p[i] > size[i]) {
			q[i] = size[i];
		}
	}
	
	Vector normal;
	float overlap = 1000000;
	
	// Point is inside box
	if (p == q) {
		for (int i = 0; i < 3; i++) {
			if (p[i] < overlap) {
				overlap = p[i];
				normal = Vector();
				normal[i] = -1;
			}
			
			if (size[i] - p[i] < overlap) {
				overlap = size[i] - p[i];
				normal = Vector();
				normal[i] = 1;
			}
		}
		
		overlap += r;
	} else {
		overlap = r - p.distance(q);
		normal = (p - q).normalise();
	}
	
	if (overlap < -Globals::_NEAR) { return false; }
	
	normal = box->getOrientation().transform(normal);
	
	if (flip) {
		normal *= -1;
	}
	
	Contact contact;
	contact.penetration = overlap;
	contact.normal = normal;
	
	if (flip) {
		contact.position = sphere->getPosition() + normal * (r - overlap / 2);
	} else {
		contact.position = sphere->getPosition() - normal * (r - overlap / 2);
	}
	/*contact.isEdge[0] = false;
	contact.isEdge[1] = false;
	contact.featureID[0] = 0;
	contact.featureID[1] = 0;*/
	
	if (flip) {
		contact.init(sphere, box);
	} else {
		contact.init(box, sphere);
	}
	
	Contact* newContact = new Contact(contact);
	
	m_contacts.push_back(newContact);
	
	return true;
}

float testBoxBoxAxis (Box** shape, const Vector3D& dir, bool& flip) {
	int i;
	float max1 = dir.dot(shape[0]->getExtremePoint(dir, i));
	float min1 = dir.dot(shape[0]->getExtremePoint(-dir, i));
	float max2 = dir.dot(shape[1]->getExtremePoint(dir, i));
	float min2 = dir.dot(shape[1]->getExtremePoint(-dir, i));
	
	float overlap1 = max1 - min2;
	float overlap2 = max2 - min1;
	
	if (overlap2 < overlap1) {
		flip = true;
		return overlap2;
	} else {
		flip = false;
		return overlap1;
	}
}

inline float Dot (const Vector3D& a, const Vector3D& b) {
	return a.dot(b);
}

// Clamp n to lie within the range [min, max]
float Clamp(float n, float min, float max) {
    if (n < min) return min;
    if (n > max) return max;
    return n;
}

const float EPSILON = 0.000001f;

// Closest point between two segments from Christer Ericson's book

// Computes closest points C1 and C2 of S1(s)=P1+s*(Q1-P1) and
// S2(t)=P2+t*(Q2-P2), returning s and t. Function result is squared
// distance between between S1(s) and S2(t)
float ClosestPtSegmentSegment(const Vector3D& p1, const Vector3D& q1, const Vector3D& p2, const Vector3D& q2,
                              float &s, float &t, Vector3D &c1, Vector3D &c2)
{
	typedef physics::Vector3D Vector;
	
    Vector d1 = q1 - p1; // Direction vector of segment S1
    Vector d2 = q2 - p2; // Direction vector of segment S2
    Vector r = p1 - p2;
    float a = Dot(d1, d1); // Squared length of segment S1, always nonnegative
    float e = Dot(d2, d2); // Squared length of segment S2, always nonnegative
    float f = Dot(d2, r);

    // Check if either or both segments degenerate into points
    if (a <= EPSILON && e <= EPSILON) {
        // Both segments degenerate into points
        s = t = 0.0f;
        c1 = p1;
        c2 = p2;
        return Dot(c1 - c2, c1 - c2);
    }
    if (a <= EPSILON) {
        // First segment degenerates into a point
        s = 0.0f;
        t = f / e; // s = 0 => t = (b*s + f) / e = f / e
        t = Clamp(t, 0.0f, 1.0f);
    } else {
        float c = Dot(d1, r);
        if (e <= EPSILON) {
            // Second segment degenerates into a point
            t = 0.0f;
            s = Clamp(-c / a, 0.0f, 1.0f); // t = 0 => s = (b*t - c) / a = -c / a
        } else {
            // The general nondegenerate case starts here
            float b = Dot(d1, d2);
            float denom = a*e-b*b; // Always nonnegative

            // If segments not parallel, compute closest point on L1 to L2, and
            // clamp to segment S1. Else pick arbitrary s (here 0)
            if (denom != 0.0f) {
                s = Clamp((b*f - c*e) / denom, 0.0f, 1.0f);
            } else s = 0.0f;

            // Compute point on L2 closest to S1(s) using
            // t = Dot((P1+D1*s)-P2,D2) / Dot(D2,D2) = (b*s + f) / e
            t = (b*s + f) / e;

            // If t in [0,1] done. Else clamp t, recompute s for the new value
            // of t using s = Dot((P2+D2*t)-P1,D1) / Dot(D1,D1)= (t*b - c) / a
            // and clamp s to [0, 1]
            if (t < 0.0f) {
                t = 0.0f;
                s = Clamp(-c / a, 0.0f, 1.0f);
            } else if (t > 1.0f) {
                t = 1.0f;
                s = Clamp((b - c) / a, 0.0f, 1.0f);
            }
        }
    }

    c1 = p1 + d1 * s;
    c2 = p2 + d2 * t;
    return Dot(c1 - c2, c1 - c2);
}

template <>
bool ContactGenerator<ThreeDee>::checkBoxBox (Box* shape1, Box* shape2)
{
	Box* shape[2];
	
	shape[0] = shape1;
	shape[1] = shape2;
	
	#if 0
	
	for (int s = 0; s < 2; s++) {
		for (int i = 0; i < 8; i++) {
			Contact contact;
			
			contact.penetration = 1000000;
			
			Vector p = shape[!s]->getVertex(i);
			
			p = shape[s]->worldToLocal(p);
			
			Vector size = shape[s]->getSize();
			
			p += size / 2;
			
			for (int j = 0; j < 3; j++) {
				if (p[j] < contact.penetration) {
					contact.penetration = p[j];
					contact.normal = Vector();
					contact.normal[j] = -1;
				}
				
				if (size[j] - p[j] < contact.penetration) {
					contact.penetration = size[j] - p[j];
					contact.normal = Vector();
					contact.normal[j] = 1;
				}
			}
			
			if (contact.penetration >= 0) {
				contact.normal = shape[s]->getOrientation().transform(contact.normal);
				
				contact.position = shape[!s]->getVertex(i);
				
				if (s) { contact.normal *= -1; }
				
				contact.init(shape1, shape2);
				
				Contact* newContact = new Contact(contact);
				
				m_contacts.push_back(newContact);
			}
		}
	}
	
	#else
	
	Contact contact;
	
	contact.penetration = 1000000;
	
	bool flip;
	
	for (int s = 0; s < 2; s++) {
		for (int i = 0; i < 3; i++) {
			Vector3D dir = shape[s]->getNormal(i);
			
			float overlap = testBoxBoxAxis(shape, dir, flip);
			
			if (overlap < 0) {
				return false;
			}
			
			if (overlap < contact.penetration) {
				contact.penetration = overlap;
				contact.normal = flip ? -dir : dir;
				
				int j;
				contact.position = shape[!s]->getExtremePoint((s==0) ? -contact.normal : contact.normal, j);
				
				contact.featureID[s] = i + flip ? 3 : 0;
				contact.featureID[!s] = j;
				contact.isEdge[s] = 1; // Face
				contact.isEdge[!s] = 0; // Vertex
			}
		}
	}
	
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			// TODO: make this test robust (could be parallel)
			Vector3D dir = shape1->getNormal(i).cross(shape2->getNormal(j));
			
			if (dir == Vector3D()) {
				continue;
			}
			
			dir = dir.normalise();
			
			float overlap = testBoxBoxAxis(shape, dir, flip);
			
			if (overlap < 0) {
				return false;
			}
			
			if (overlap < contact.penetration) {
				contact.penetration = overlap;
				contact.normal = flip ? -dir : dir;
				
				Vector3D p1, p2, q1, q2;
				
				shape1->getExtremeEdge(contact.normal, p1, p2);
				shape2->getExtremeEdge(-contact.normal, q1, q2);
				
				Vector3D c1, c2;
				float s, t;
				
				ClosestPtSegmentSegment(p1, q1, p2, q2, s, t, c1, c2);
				
				contact.position = (c1 + c2) * 0.5f;
			}
		}
	}
	
	contact.init(shape1, shape2);
	
	Contact* newContact = new Contact(contact);
	
	m_contacts.push_back(newContact);
	
	#endif
	
	return true;
}

template <> bool ContactGenerator<TwoDee>::checkSphereSphere (Sphere* shape1, Sphere* shape2) { return false; }
template <> bool ContactGenerator<TwoDee>::checkBoxSphere (Box* shape1, Sphere* shape2, bool flip) { return false; }
template <> bool ContactGenerator<TwoDee>::checkBoxBox (Box* shape1, Box* shape2) { return false; }



template class ContactGenerator<TwoDee>;
template class ContactGenerator<ThreeDee>;

