#include "contact_list_base.h"

#include <algorithm>
#include <boost/foreach.hpp>

#include "body.h"

#include "collision_listener.h"

using namespace physics;
using namespace std;

// Don't enable this: it'll stop stuff from compiling and doesn't even work anyway
#define NAIVE_CONTACT_LIST_SORT 0

template <typename d, typename C>
BaseContactList<d, C>::~BaseContactList () {
	BOOST_FOREACH(Contact* contact, contacts) {
		delete contact;
	}
}

template <typename d, typename C>
Contact<d>* BaseContactList<d, C>::getGreatestVelocity () const {
#if NAIVE_CONTACT_LIST_SORT
	return contacts.front();
#else
	return *min_element(contacts.begin(), contacts.end(), CompareSpeed<Contact>());
#endif
}

template <typename d, typename C>
Contact<d>* BaseContactList<d, C>::getGreatestPenetration () const {
#if NAIVE_CONTACT_LIST_SORT
	return contacts.front();
#else
	return *min_element(contacts.begin(), contacts.end(), ComparePenetration<Contact>());
#endif
}

template <typename d, typename C>
Contact<d>* BaseContactList<d, C>::find (Contact* contact) {
	#if BODY_STORES_CONTACTS
	BOOST_FOREACH(Contact* c, contact->body0->contacts) {
		if (*c == *contact) {
			return c;
		}
	}
	BOOST_FOREACH(Contact* c, contact->body1->contacts) {
		if (*c == *contact) {
			return c;
		}
	}
	#else
	BOOST_FOREACH(Contact* c, contacts) {
		if (*c == *contact) {
			return c;
		}
	}
	#endif
	
	return 0;
}

// only used when cacheContacts is true
template <typename d, typename C>
void BaseContactList<d, C>::add (Contact* newContact) {
	Contact* oldContact = find(newContact);
	
	if (oldContact != 0) {
		// copy data, not pointer value
		//*oldContact = *newContact;
		delete newContact;
	} else {
		contacts.push_back(newContact);
		
		#if BODY_STORES_CONTACTS
			newContact->body0->contacts.push_back(newContact);
			newContact->body1->contacts.push_back(newContact);
		#endif
	}
}

template <typename d, typename C>
void BaseContactList<d, C>::addContacts (const input_container& newContacts)
{
	if (! cacheContacts) {
		
		// TODO: need to get some data from contacts from last frame
		
		BOOST_FOREACH(Contact* c, contacts) {
			#if BODY_STORES_CONTACTS
				c->body0->contacts.clear();
				c->body1->contacts.clear();
			#endif
			
			delete c;
		}
		
		// No need to call clear(): assign() does that for us.
		contacts.assign(newContacts.begin(), newContacts.end());
		
		BOOST_FOREACH(Contact* c, contacts) {
			// We shouldn't need to do this I think
			// (And for now it breaks in 3D)
			//c->calculatePenetration();
			//c->calculateMass();
			
			#if BODY_STORES_CONTACTS
				c->body0->contacts.push_back(c);
				c->body1->contacts.push_back(c);
			#endif
		}
	} else {
		BOOST_FOREACH(Contact* c, newContacts) {
			add(c);
		}
		
		for (typename container::iterator it = contacts.begin(); it != contacts.end(); /* Nothing */) {
			Contact* c = *it;
			
			c->calculatePenetration();
			c->calculateMass();
			c->calculateVelocities();
			
			if (c->penetration < -Globals::_NEAR) {
				#if BODY_STORES_CONTACTS
					c->body0->contacts.remove(c);
					c->body1->contacts.remove(c);
				#endif
				delete c;
				
				it = contacts.erase(it);
			} else {
				it++;
			}
		}
	}
}

template <typename d, typename C>
void BaseContactList<d, C>::removeContacts (Body<d>* body) {
	for (typename container::iterator it = contacts.begin(); it != contacts.end(); /* Nothing */) {
		Contact* c = *it;
		
		if (c->body0 == body || c->body1 == body) {
			#if BODY_STORES_CONTACTS
				c->body0->contacts.remove(c);
				c->body1->contacts.remove(c);
			#endif
			c->body0->setAwake(true);
			c->body1->setAwake(true);
			delete c;
			
			it = contacts.erase(it);
		} else {
			it++;
		}
	}
}

template <typename d, typename C>
void BaseContactList<d, C>::foreach (callback_type callback) {
	BOOST_FOREACH(Contact* c, contacts) {
		callback(c);
	}
}

template <typename d, typename C>
void BaseContactList<d, C>::notifyListeners () {
	for (typename C::iterator it = contacts.begin(); it != contacts.end(); /* Nothing */ ) {
		Contact* c = *it;
		
		bool remove = false;
		
		BOOST_FOREACH(CollisionListener<d>* listener, c->body0->m_listeners) {
			if (! listener->notify(c, 0)) {
				remove = true;
			}
		}
		
		BOOST_FOREACH(CollisionListener<d>* listener, c->body1->m_listeners) {
			if (! listener->notify(c, 1)) {
				remove = true;
			}
		}
		
		if (remove) {
			it = contacts.erase(it);
		} else {
			it++;
		}
	}
}

template <typename d, typename C>
void BaseContactList<d, C>::prepareVelocities () {
	BOOST_FOREACH(Contact* c, contacts) {
		c->calculateVelocities();
	}
	
	#if NAIVE_CONTACT_LIST_SORT
	//std::sort(contacts.begin(), contacts.end(), CompareSpeed<Contact>());
	contacts.sort(CompareSpeed<Contact>());
	#endif
}

template <typename d, typename C>
void BaseContactList<d, C>::preparePenetrations () {
	/*BOOST_FOREACH(Contact* c, contacts) {
		c->calculatePenetration();
	}*/

	#if NAIVE_CONTACT_LIST_SORT
	//std::sort(contacts.begin(), contacts.end(), ComparePenetration<Contact>());
	contacts.sort(ComparePenetration<Contact>());
	#endif
}

template <typename d, typename C>
void BaseContactList<d, C>::updateVelocities (Contact* c1)
{
	#if NAIVE_CONTACT_LIST_SORT
	container temp;
	#endif
	
	#if BODY_STORES_CONTACTS
	for (int i = 0; i < 2; i++) {
		Body<d>* body = c1->body[i];
		
		BOOST_FOREACH(Contact* c2, body->contacts) {
			c2->calculateVelocities();
			
			#if NAIVE_CONTACT_LIST_SORT
			contacts.remove(c2);
			temp.push_back(c2);
			#endif
		}
	}
	#else
	BOOST_FOREACH(Contact* c2, contacts) {
		if (c1->body0 == c2->body0 || c1->body0 == c2->body1 || c1->body1 == c2->body0 || c1->body1 == c2->body1)
		{
			c2->calculateVelocities();
			#if NAIVE_CONTACT_LIST_SORT
			it = contacts.erase(it);
			temp.push_back(c2);
			#endif
		}
		
	}
	#endif
	
	#if NAIVE_CONTACT_LIST_SORT
	temp.sort(CompareSpeed<Contact>());
	temp.unique();
	typename container::iterator it = contacts.begin();
	contacts.splice(it, temp);
	std::inplace_merge(contacts.begin(), it, contacts.end(), CompareSpeed<Contact>());
	#endif
}

template <typename d, typename C>
void BaseContactList<d, C>::updatePenetrations (Contact* c1, float* moveLinear)
{
	#if NAIVE_CONTACT_LIST_SORT
	container temp;
	#endif
	
	for (int i = 0; i < 2; i++) {
		if (moveLinear[i] == 0) { continue; }
		
		Body<d>* body = c1->body[i];
		
		#if BODY_STORES_CONTACTS
		BOOST_FOREACH(Contact* c2, body->contacts)
		#else
		BOOST_FOREACH(Contact* c2, contacts)
		#endif
		{
			if (body == c2->body[0])
			{
				c2->setPenetration( c2->penetration + moveLinear[i] * c1->normal.dot(c2->normal));
				#if NAIVE_CONTACT_LIST_SORT
				it = contacts.erase(it);
				temp.push_back(c2);
				#endif
			}
			else if (body == c2->body[1]) // technically unnecessary if BODY_STORES_CONTACTS is true
			{
				c2->setPenetration( c2->penetration - moveLinear[i] * c1->normal.dot(c2->normal));
				#if NAIVE_CONTACT_LIST_SORT
				it = contacts.erase(it);
				temp.push_back(c2);
				#endif
			}
			
			
		}
	}
	
	#if NAIVE_CONTACT_LIST_SORT
	temp.sort(ComparePenetration<Contact>());
	temp.unique();
	typename container::iterator it = contacts.begin();
	contacts.splice(it, temp);
	std::inplace_merge(contacts.begin(), it, contacts.end(), ComparePenetration<Contact>());
	#endif
}

template class BaseContactList<TwoDee, std::vector<Contact<TwoDee>*, std::allocator<Contact<TwoDee>*> > >;
template class BaseContactList<TwoDee, std::list<Contact<TwoDee>*, std::allocator<Contact<TwoDee>*> > >;

template class NaiveContactList<TwoDee, std::vector>;
template class NaiveContactList<TwoDee, std::list>;

template class BaseContactList<ThreeDee, std::vector<Contact<ThreeDee>*, std::allocator<Contact<ThreeDee>*> > >;
template class BaseContactList<ThreeDee, std::list<Contact<ThreeDee>*, std::allocator<Contact<ThreeDee>*> > >;

template class NaiveContactList<ThreeDee, std::vector>;
template class NaiveContactList<ThreeDee, std::list>;

