#ifndef _PHYSICS_CONTACT_RESOLVER_
#define _PHYSICS_CONTACT_RESOLVER_

#include "common.h"

#include "joint.h"

#include <list>
#include <vector>

namespace physics {
	
	template <typename d> class Contact;
	template <typename d> class ContactList;
	template <typename d> class Joint;
	

	template <typename d>
	/**
	 * Contact Resolver.
	 * Extended by the velocity resolver and position resolver algorithms.
	 */
	class ContactResolver {
		public:
		
		//! Constructor
		ContactResolver () {}
		//! Destructor
		virtual ~ContactResolver () {}
		
		virtual void init (World<d>* world) {}
		
		/*!
		*	Resolves contacts
		*	\param contacts Pointer to a list of contacts
		*	\param joints	Pointer to a list of joints
		*/
		virtual void resolve (ContactList<d>* contacts, std::vector<Joint<d>*>* joints) = 0;
	};
	
	/**
	 * Implementation of ContactResolver that does nothing.
	 */
	template <typename d>
	class NullContactResolver : public ContactResolver<d> {
		public:
		
		void resolve (ContactList<d>* contacts, std::vector<Joint<d>*>* joints) {}
	};
	
	template <typename d>
	/**
	* Simple implementation of a velocity contact resolver
	*/
	class StupidVelocityResolver : public ContactResolver<d> {
		public:
		
		USE_DIMENSION(d);
		
		typedef physics::Contact<d> Contact;
		typedef physics::Joint<d> Joint;

		StupidVelocityResolver (int iterations = -1, bool sortContacts = true);
		
		void resolve (ContactList<d>* contacts, std::vector<Joint*>* joints);
		
		private:
		
		bool sortContacts;
		int iterations;
		
		// So hacky...
		static ContactList<d>* contacts;
		
		// These modify the impulse parameter (which is in contact coordinates like everything else)
		static void applyFriction (Vector2D& impulse, const Matrix2& velocityPerImpulse, float friction, float desiredDeltaVelocity);
		static void applyFriction (Vector3D& impulse, const Matrix3& velocityPerImpulse, float friction, float desiredDeltaVelocity);
		
		static void resolve (Contact* contact);
		static void resolve (Contact& data);
	};
	
	template <typename d>
	/**
	* Simple implementation of a position contact resolver
	*/
	class StupidPositionResolver : public ContactResolver<d> {
		public:
		
		USE_DIMENSION(d);
		
		typedef physics::Contact<d> Contact;
		typedef physics::Joint<d> Joint;

		StupidPositionResolver (int iterations = -1, bool sortContacts = false);
		
		void resolve (ContactList<d>* contacts, std::vector<Joint*>* joints);
		
		private:
		
		bool sortContacts;
		int iterations;
		
		// So hacky...
		static ContactList<d>* contacts;
		
		static void resolve (Contact* contact);
		static void resolve (Contact& data, float* moveLinear);
	};
	
	template <typename d>
	/*!
	*	Velocity contact resolver based upon an algorithm by Erin Catto
	*/
	class CattoVelocityResolver : public ContactResolver<d> {
		public:
		
		USE_DIMENSION(d);
		
		typedef physics::Contact<d> Contact;
		typedef physics::Joint<d> Joint;
		
		CattoVelocityResolver (int iterationCount = 10) : iterations(iterationCount) {}
		
		void resolve (ContactList<d>* contacts, std::vector<Joint*>* joints);
		
		private:
		
		int iterations;
		
		static void applyFriction (physics::Contact<TwoDee>* contact);
		static void applyFriction (physics::Contact<ThreeDee>* contact);
		
		static void prepare (Contact* contact);
		static void resolve (Contact* contact);
	};
	
	template <typename d>
	/*! 
	*	Position contact resolver based upon an algorithm by Erin Catto
	*/
	class CattoPositionResolver : public ContactResolver<d> {
		public:
		
		USE_DIMENSION(d);
		
		typedef physics::Contact<d> Contact;
		typedef physics::Joint<d> Joint;
		
		CattoPositionResolver (int iterationCount = 10) : iterations(iterationCount) {}
		
		void resolve (ContactList<d>* contacts, std::vector<Joint*>* joints);
		
		private:
		
		int iterations;
		
		static void prepare (Contact* contact);
		static void resolve (Contact* contact);
	};
};

#endif

