#ifndef _PHYSICS_FORCES_
#define _PHYSICS_FORCES_

#include "common.h"

#include "vector.h"

namespace physics {
	/*!
	*	Parent class for all force generators. Allows for a force to be applied to a body within the world.
	*/
	class ForceGenerator {
		public:
		//! Constructor
		virtual ~ForceGenerator () {} 
		//! Apply a force
		/*! \param timestep Time step
		*/
		virtual void apply (float timestep) = 0;
	};
	
	template <typename d>
	class AnchoredSpring : public ForceGenerator {
		public:
		
		USE_DIMENSION(d);
		
		AnchoredSpring (float k, float restLength)
			: m_body(0), m_springConstant(k), m_restLength(restLength) {}
		
		AnchoredSpring (Vector anchor, Body* body, Vector bodyAnchor, float k, float restLength)
			: m_anchor(anchor), m_body(body), m_bodyAnchor(bodyAnchor), m_springConstant(k), m_restLength(restLength) {}
		
		virtual void apply (float timestep);
		
		Vector getAnchor () const { return m_anchor; }
		void setAnchor (const Vector& p) { m_anchor = p; }
		
		Body* getBody () const { return m_body; }
		Vector getBodyAnchor () const { return m_bodyAnchor; }

		void unsetBody () { m_body = 0; }
		void setBody (Body* b, const Vector& p) { m_body = b; m_bodyAnchor = p; }
		
		float getSpringConstant() { return m_springConstant; }
		float getRestLength() { return m_restLength; }

		protected:
		Vector m_anchor;
		Body* m_body;
		Vector m_bodyAnchor;
		float m_springConstant;
		float m_restLength;
	};
	
	typedef AnchoredSpring<TwoDee> AnchoredSpring2D;
	
	template <typename d>
	class CleverAnchoredSpring : public ForceGenerator {
		public:
		
		USE_DIMENSION(d);
		
		CleverAnchoredSpring (float k, float damping)
			: m_body(0), m_springConstant(k), m_damping(damping) {}
		
		CleverAnchoredSpring (Vector anchor, Body* body, Vector bodyAnchor, float k, float damping)
			: m_anchor(anchor), m_body(body), m_bodyAnchor(bodyAnchor), m_springConstant(k), m_damping(damping) {}
		
		virtual void apply (float timestep);
		
		Vector getAnchor () const { return m_anchor; }
		void setAnchor (const Vector& p) { m_anchor = p; }
		
		Vector getBodyAnchor() const { return m_bodyAnchor; }

		Body* getBody () const { return m_body; }
		void unsetBody () { m_body = 0; }
		void setBody (Body* b, const Vector& p) { m_body = b; m_bodyAnchor = p; }
		
		float getSpringConstant() { return m_springConstant; }
		float getDamping() { return m_damping; }

		protected:
		
		Vector m_anchor;
		Body* m_body;
		Vector m_bodyAnchor;
		float m_springConstant;
		float m_damping;
	};
	
	typedef CleverAnchoredSpring<TwoDee> CleverAnchoredSpring2D;
};

// This code originally from Game Physics Engine Development by Ian Millington

template <typename d>
void physics::AnchoredSpring<d>::apply (float timestep) {
	if (! m_body || m_body->isStatic()) { return; }

	Vector worldBodyAnchor = m_body->localToWorld(m_bodyAnchor);

	Vector offset = m_anchor - worldBodyAnchor;

	// Calculate the magnitude of the force
	float magnitude = offset.length();
	magnitude -= m_restLength;
	magnitude = fabs(magnitude);
	magnitude *= m_springConstant;

	// Calculate the final force and apply it
	m_body->applyImpulse(offset.normalise() * (m_body->getMass() * magnitude * timestep), worldBodyAnchor);
	
}

// This code originally from Game Physics Engine Development by Ian Millington

template <typename d>
void physics::CleverAnchoredSpring<d>::apply (float duration) {
	if (! m_body || m_body->isStatic()) { return; }

	Vector worldBodyAnchor = m_body->localToWorld(m_bodyAnchor);

	Vector position = worldBodyAnchor - m_anchor;
	
	Vector velocity = m_body->getVelocityAtPoint(worldBodyAnchor);

	// Calculate the constants and check they are in bounds.
	float gamma = 0.5f * sqrt(4 * m_springConstant - m_damping*m_damping);
	if (gamma == 0.0f) return;
	Vector c = position * (m_damping / (2.0f * gamma)) + velocity * (1.0f / gamma);

	// Calculate the target position
	Vector target = position * cos(gamma * duration) + c * sin(gamma * duration);
	target *= exp(-0.5f * duration * m_damping);

	// Calculate the resulting acceleration and therefore the force
	Vector accel = (target - position) * (1.0f / duration*duration) - velocity * duration;

	// Calculate the final force and apply it
	m_body->applyImpulse(accel * m_body->getMass(), worldBodyAnchor);
}

#endif

