#include "joint.h"

using namespace physics;

template <typename d>
void Joint<d>::set(Body* body1, Body* body2, const Vector& anchor1)
{
	this->body1 = body1;
	this->body2 = body2;
	
	localAnchor1 = body1->worldToLocal(anchor1);
	localAnchor2 = body2->worldToLocal(anchor1);
}

template <typename d>
void Joint<d>::prepare()
{
	float k_biasFactor = biasFactor * Globals::inverseTimeStep;

	// orientation will have changed - transform the local anchor position to match (so it stays fixed relative to local co-ordinates)
	r1 = body1->getOrientation().transform(localAnchor1);
	r2 = body2->getOrientation().transform(localAnchor2);

	velocityPerImpulse = Matrix();

	// build the 'K' matrix - see Erin's Slides
	velocityPerImpulse = Matrix::identity() * body1->getInverseMass() + Matrix::identity() * body2->getInverseMass();
	velocityPerImpulse += body1->getMOI().getVelocityPerImpulse(r1);
	velocityPerImpulse += body2->getMOI().getVelocityPerImpulse(r2);
	velocityPerImpulse += Matrix::identity() * softness;

	// M = inv(K)
	velocityPerImpulse.invert();

	Vector p1 = body1->getPosition() + r1;
	Vector p2 = body2->getPosition() + r2;
	Vector dp = p2 - p1;

	bias = -k_biasFactor * dp;
	
	body1->applyImpulse(-P, p1);
	body2->applyImpulse(P, p2);
}

template <typename d>
void Joint<d>::applyImpulses()
{
	Vector tmp1 = r1 + body1->getPosition();
	Vector tmp2 = r2 + body2->getPosition();
	
	Vector dv = body2->getVelocityAtPoint(tmp2) - body1->getVelocityAtPoint(tmp1);	
	
	Vector impulse = velocityPerImpulse * (bias - dv - softness * P);
	
	body1->applyImpulse(-impulse, tmp1);
	body2->applyImpulse(impulse, tmp2);


	P += impulse;
}


template <typename d>
bool Joint<d>::jointViolated()
{
	
	return false;
}


template class Joint<TwoDee>;
template class Joint<ThreeDee>;