#ifndef _PHYSICS_LINE_
#define _PHYSICS_LINE_

#include "common.h"

#include "area.h"
#include "vector.h"
#include "aarect.h"

namespace physics {
	/*! \class Line
	*	Provides the line shape in the 2D simulation
	*/
	class Line : public Area
	{
		public:
		Line () {}
		
		Line (float x1, float y1, float x2, float y2) :
			Area(Vector2D((x1 + x2)/2, (y1 + y2)/2)),
			m_p1(x1, y1), m_p2(x2, y2),
			m_local_p1(m_p1 - getPosition()),
			m_local_p2(m_p2 - getPosition()) {}
		
		Line (Vector2D p1, Vector2D p2) :
			Area((p1 + p2)/2),
			m_p1(p1), m_p2(p2),
			m_local_p1(p1 - getPosition()),
			m_local_p2(p2 - getPosition()) {}
		
		bool contains (const Vector2D &p) const { return distanceSq(p) < 0.0001f; }
		bool intersects (const physics::AARect& rect) const;
		
		float getRadius () const { return 0; }
		
		//! Returns x coordinate of the first position
		float x1 () const { return m_p1.x; }
		//! Returns y coordinate of the first position
		float y1 () const { return m_p1.y; }
		//! Returns x coordinate of the second position
		float x2 () const { return m_p2.x; }
		//! Returns y coordinate of the second position
		float y2 () const { return m_p2.y; }
		
		//! Return the first position
		Vector2D getP1 () const { return m_p1; }
		//! Return the second position
		Vector2D getP2 () const { return m_p2; }
		
		//! Get method for the length of the line
		/*! \return Float; Length of the line
		*/
		float getLength () const { return m_local_p1.distance(m_local_p2); }
		//! Get method for the length of the line squared
		/*! \return Float; Length of the line squared
		*/
		float getLengthSq () const { return m_local_p1.distanceSquared(m_local_p2); }
		
		float distance (const Vector2D &p) const { return sqrt(distanceSq(p)); }
		float distanceSq (const Vector2D &p) const;
		
		Vector2D getClosestPoint (const Vector2D& p, float& t) const;
		
		Vector2D getNormal () const;
		Vector2D getNormal (int i) const;
		Vector2D getVertex (int i) const { return (i % 2 == 0) ? m_p1 : m_p2; }
		
		int getVertexCount () const { return 2; }
		
		Vector2D getExtremePoint (const Vector2D& dir, int& i) const;
		
		float getArea () const { return 0.05f * getLength(); }
		
		//! Get the Moment of Inertia
		/*! \return Float; Moment of Inertia
		*/
		float getMOI () const { return getMass() * getLengthSq() / 12.0f; } // Moment of Inertia
		
		Type type () const { return LINE; }
		
		private:
		
		virtual void thisMoved ();
		
		Vector2D m_p1;
		Vector2D m_p2;
		
		Vector2D m_local_p1;
		Vector2D m_local_p2;
	};
	
	/**
	 * This is not a shape class yet: just used for calculating closest point
	 * between two line segments
	 */
	class Line3D {
		public:
		
		Line3D () {}
		Line3D (const Vector3D& p1, const Vector3D& p2) : p1(p1), p2(p2) {}
		
		Vector3D p1;
		Vector3D p2;
		
		/**
		 * Given segment ab and point c, computes closest point d on ab.
		 * Also returns t for the parametric position of d, d(t) = a + t*(b - a)
		 */
		static void closestPoint (const Vector3D& a, const Vector3D& b, const Vector3D& c, float& t, Vector3D& d);
		
		static void closestPoint (const Line3D& line, const Vector3D& p,
			float& t, Vector3D& d)
		{
			return Line3D::closestPoint(line.p1, line.p2, p, t, d);
		}
		
		/**
		 * Computes closest points C1 and C2 of S1(s)=P1+s*(Q1-P1) and
		 * S2(t)=P2+t*(Q2-P2), returning s and t. Function result is squared
		 * distance between between S1(s) and S2(t)
		 */
		static float closestPoint (const Vector3D& p1, const Vector3D& q1,
			const Vector3D& p2, const Vector3D& q2,
			float &s, float &t,
			Vector3D &c1, Vector3D &c2);
		
		static float closestPoint (const Line3D& line1, const Line3D& line2,
			float &s, float &t,
			Vector3D &c1, Vector3D &c2)
		{
			return Line3D::closestPoint(line1.p1, line2.p1, line1.p2, line2.p2, s, t, c1, c2);
		}
	};
};

inline physics::Vector2D physics::Line::getNormal () const {
	Vector2D normal = m_p2 - m_p1;
	
	normal = Vector2D(-normal.y, normal.x);
	
	return normal.normalise();
}

inline physics::Vector2D physics::Line::getNormal (int i) const {
	Vector2D normal = getNormal();
	
	if (i % 2 != 0) { normal *= -1; }
	
	return normal;
}

inline physics::Vector2D physics::Line::getExtremePoint (const physics::Vector2D& dir, int& i) const {
	if (dir.dot(m_p2 - m_p1) > 0) {
		i = 1;
		return m_p2;
	} else {
		i = 0;
		return m_p1;
	}
}

inline void physics::Line::thisMoved () {
	m_p1 = localToWorld(m_local_p1);
	m_p2 = localToWorld(m_local_p2);
}

// Following two algorithms from Real-Time Collision Detection by Christer Ericson

inline float physics::Line::distanceSq (const physics::Vector2D& c) const {
	Vector2D ab = m_p2 - m_p1;
	Vector2D ac = c - m_p1;
	Vector2D bc = c - m_p2;

	float e = ac.dot(ab);

	// Handle cases where c projects outside ab
	if (e <= 0.0f) return ac.dot(ac);
	float f = ab.dot(ab);
	if (e >= f) return bc.dot(bc);
	
	// Handle case where c projects onto ab
	return ac.dot(ac) - e * e / f;
}

inline physics::Vector2D physics::Line::getClosestPoint (const physics::Vector2D& p, float& t) const {
	Vector2D ab = m_p2 - m_p1;

	t = (p - m_p1).dot(ab) / ab.dot(ab);

	if (t < 0.0f) t = 0.0f;
	if (t > 1.0f) t = 1.0f;

	return m_p1 + t * ab;
}

inline bool physics::Line::intersects (const physics::AARect& rect) const {

	if (rect.contains(Line::getP1())) return 1;
	if (rect.contains(Line::getP2())) return 1;

	if (rect.contains(Line::getPosition())) return 1;

	return 0;
}

#endif

