#ifndef _PHYSICS_MATRIX_
#define _PHYSICS_MATRIX_

#include "common.h"

#include <math.h>
#include "vector.h"
#include "quaternion.h"

// TODO: make operator*= and friends return a reference
// TODO: tidy up generally so they both have same functionality

namespace physics {
	class Matrix2
	{
		public:

		// null matrix
		Matrix2 ()
		{
			data[0] = 0; data[1] = 0;
			data[2] = 0; data[3] = 0;
		}

		// create a matrix from explicit indices
		Matrix2 (float a, float b, float c, float d)
		{
			data[0] = a; data[1] = b;
			data[2] = c; data[3] = d;
		}

		// Identity matrix
		static Matrix2 identity () {
			return Matrix2(1, 0, 0, 1);
		}
		
		static Matrix2 columns (const Vector2D& a, const Vector2D& b) {
			return Matrix2(a.x, b.x, a.y, b.y);
		}

		static Matrix2 rows (const Vector2D& a, const Vector2D& b) {
			return Matrix2(a.x, a.y, b.x, b.y);
		}

		static Matrix2 createBasis (const Vector2D& v) {
			return Matrix2(v.x, -v.y, v.y, v.x);
		}
		
		Vector2D operator* (const Vector2D &v) const
		{
			return Vector2D(
				v.x * data[0] + v.y * data[1],
				v.x * data[2] + v.y * data[3]
			);
		}
		
		float getDeterminant () const
		{
			return data[0] * data[3] - data[1] * data[2];
		}

		Matrix2 getInverse () const
		{
			Matrix2 ret = *this;
			ret.invert();
			return ret;
		}

		// invert this
		void invert ()
		{
			float det = getDeterminant();
 			
			if (det == 0) {
				*this = Matrix2();
				return;
			}
			
			float invDet = 1.0f / det;
			
			float temp = data[0] * invDet;
			data[0] = data[3] * invDet;
			data[3] = temp;
			
			data[1] *= -invDet;
			data[2] *= -invDet;
		}
		
		Matrix2 getTranspose () const
		{
			Matrix2 ret = *this;
			ret.transpose();
			return ret;
		}
		
		void transpose ()
		{
			float temp = data[1];
			data[1] = data[2];
			data[2] = temp;
		}

		// matrix multiplication
		Matrix2 operator*(const Matrix2 &o) const
		{
			return Matrix2(
				data[0] * o.data[0] + data[1] * o.data[2], data[0] * o.data[1] + data[1] * o.data[3],
				data[2] * o.data[0] + data[3] * o.data[2], data[2] * o.data[1] + data[3] * o.data[3]);
		}

		// matrix addition
		Matrix2 operator+(const Matrix2 &o) const
		{
			return Matrix2(
			data[0] + o.data[0], data[1] + o.data[1],
			data[2] + o.data[2], data[3] + o.data[3]
			);
		}

		// matrix subtraction
		Matrix2 operator-(const Matrix2 &o) const
		{
			return Matrix2(
			data[0] - o.data[0], data[1] - o.data[1],
			data[2] - o.data[2], data[3] - o.data[3]
			);
		}


		// inplace matrix multiplication
		void operator*=(const Matrix2 &o)
		{
			*this = (*this * o);
		}

		// inplace scalar multiplication
		void operator*= (const float scalar)
		{
			data[0] *= scalar; data[1] *= scalar;
			data[2] *= scalar; data[3] *= scalar;
		}
		
		// scalar multiplication
		Matrix2 operator* (const float scalar) const
		{
			return Matrix2(
				data[0] * scalar, data[1] * scalar,
				data[2] * scalar, data[3] * scalar
			);
		}
		
		friend Matrix2 operator*(const float scalar, const Matrix2& m);
		
		// inplace scalar division
		void operator/= (const float scalar)
		{
			data[0] /= scalar; data[1] /= scalar;
			data[2] /= scalar; data[3] /= scalar;
		}
		
		// scalar division
		Matrix2 operator/ (const float scalar) const
		{
			return Matrix2(
				data[0] / scalar, data[1] / scalar,
				data[2] / scalar, data[3] / scalar
			);
		}
		
		// inplace addition
		void operator+=(const Matrix2 &o)
		{
			data[0] += o.data[0]; data[1] += o.data[1];
			data[2] += o.data[2]; data[3] += o.data[3];
		}
		
		// inplace subtraction
		void operator-=(const Matrix2 &o)
		{
			data[0] -= o.data[0]; data[1] -= o.data[1];
			data[2] -= o.data[2]; data[3] -= o.data[3];
		}
		
		bool operator==(const Matrix2 &o)
		{
			return (data[0] == o.data[0] && data[1] == o.data[1]
				&& data[2] == o.data[2] && data[3] == o.data[3]);
		}
		
		bool operator!=(const Matrix2 &o)
		{
			return ! (*this == o);
		}

		float data[4];
	};
	
	// general 3x3 matrix class - used mostly for holding the 3DMOI
	class Matrix3
	{
	public:
		// null matrix
		Matrix3()
		{
			data[0] = 0; data[1] = 0; data[2] = 0;
			data[3] = 0; data[4] = 0; data[5] = 0;
			data[6] = 0; data[7] = 0; data[8] = 0;
		}

		// create a matrix from explicit indices
		Matrix3(float a, float b, float c, float d, float e, float f, float g, float h, float i)
		{
			data[0] = a; data[1] = b; data[2] = c; 
			data[3] = d; data[4] = e; data[5] = f;
			data[6] = g; data[7] = h; data[8] = i;
		}

		// Identity matrix
		static Matrix3 identity () {
			return Matrix3(
				1, 0, 0,
				0, 1, 0,
				0, 0, 1);
		}

		// Diagonal matrix
		static Matrix3 diagonal (const Vector3D& v) {
			return Matrix3(
				v.x, 0, 0,
				0, v.y, 0,
				0, 0, v.z);
		}

		// Diagonal matrix
		static Matrix3 diagonal (float a, float b, float c) {
			return Matrix3(
				a, 0, 0,
				0, b, 0,
				0, 0, c);
		}

		static Matrix3 columns (const Vector3D& a, const Vector3D& b, const Vector3D& c) {
			return Matrix3(
				a.x, b.x, c.x,
				a.y, b.y, c.y,
				a.z, b.z, c.z);
		}

		static Matrix3 rows (const Vector3D& a, const Vector3D& b, const Vector3D& c) {
			return Matrix3(
				a.x, a.y, a.z,
				b.x, b.y, b.z,
				c.x, c.y, c.z);
		}

		// Skew-symmetric matrix
		static Matrix3 skewSymmetric (const Vector3D& v) {
			return Matrix3(
				0, -v.z, v.y,
				v.z, 0, -v.x,
				-v.y, v.x, 0);
		}

		// Outer product of two vectors
		static Matrix3 outerProduct (const Vector3D& a, const Vector3D& b) {
			return Matrix3(
				a.x * b.x, a.x * b.y, a.x * b.z,
				a.y * b.x, a.y * b.y, a.y * b.z,
				a.z * b.x, a.z * b.y, a.z * b.z);
		}
		
		static Matrix3 createBasis (const Vector3D& normal) {
			if (normal == Vector3D()) {
				return Matrix3::identity();
			}
			
			Vector3D tangent[2];
			
			// Check whether the Z axis is nearer to the X or Y axis.
			if(fabs(normal.x) > fabs(normal.y))
			{
				// Scaling factor to ensure the results are normalized.
				const float s = 1.0f / sqrt(normal.z*normal.z + normal.x*normal.x);
				
				// The new X axis is at right angles to the world Y axis.
				tangent[0].x = normal.z*s;
				tangent[0].y = 0;
				tangent[0].z = -normal.x*s;
				
				// The new Y axis is at right angles to the new X and Z axes.
				tangent[1].x = normal.y*tangent[0].x;
				tangent[1].y = normal.z*tangent[0].x - normal.x*tangent[0].z;
				tangent[1].z = -normal.y*tangent[0].x;
			}
			else
			{
				// Scaling factor to ensure the results are normalized.
				const float s = 1.0f / sqrt(normal.z*normal.z + normal.y*normal.y);
				
				// The new X axis is at right angles to the world X axis.
				tangent[0].x = 0;
				tangent[0].y = -normal.z*s;
				tangent[0].z = normal.y*s;
				
				// The new Y axis is at right angles to the new X and Z axes.
				tangent[1].x = normal.y*tangent[0].z -
				normal.z*tangent[0].y;
				tangent[1].y = -normal.x*tangent[0].z;
				tangent[1].z = normal.x*tangent[0].y;
			}
			
			return Matrix3::columns(normal, tangent[0], tangent[1]);
		}
		
		// creates a matrix representation of the orientation of the given quaternion
		explicit Matrix3 (const Quaternion &q)
        {
            data[0] = 1 - (2*q.y*q.y + 2*q.z*q.z);
            data[1] = 2*q.x*q.y + 2*q.z*q.w;
            data[2] = 2*q.x*q.z - 2*q.y*q.w;
            data[3] = 2*q.x*q.y - 2*q.z*q.w;
            data[4] = 1 - (2*q.x*q.x  + 2*q.z*q.z);
            data[5] = 2*q.y*q.z + 2*q.x*q.w;
            data[6] = 2*q.x*q.z + 2*q.y*q.w;
            data[7] = 2*q.y*q.z - 2*q.x*q.w;
            data[8] = 1 - (2*q.x*q.x  + 2*q.y*q.y);
        }

		Vector3D operator*(const Vector3D &vector) const
        {
            return Vector3D(
                vector.x * data[0] + vector.y * data[1] + vector.z * data[2],
                vector.x * data[3] + vector.y * data[4] + vector.z * data[5],
                vector.x * data[6] + vector.y * data[7] + vector.z * data[8]
            );
        }

		// inverse helper method - for static etc... 
		void setInverse(const Matrix3 &m)
        {
            float t4 = m.data[0]*m.data[4];
            float t6 = m.data[0]*m.data[5];
            float t8 = m.data[1]*m.data[3];
            float t10 = m.data[2]*m.data[3];
            float t12 = m.data[1]*m.data[6];
            float t14 = m.data[2]*m.data[6];

            // Calculate the determinant
            float t16 = (t4*m.data[8] - t6*m.data[7] - t8*m.data[8]+
                        t10*m.data[7] + t12*m.data[5] - t14*m.data[4]);

            // Make sure the determinant is non-zero.
            if (t16 == (float)0.0f) return;
            float t17 = 1/t16;

            data[0] = (m.data[4]*m.data[8]-m.data[5]*m.data[7])*t17;
            data[1] = -(m.data[1]*m.data[8]-m.data[2]*m.data[7])*t17;
            data[2] = (m.data[1]*m.data[5]-m.data[2]*m.data[4])*t17;
            data[3] = -(m.data[3]*m.data[8]-m.data[5]*m.data[6])*t17;
            data[4] = (m.data[0]*m.data[8]-t14)*t17;
            data[5] = -(t6-t10)*t17;
            data[6] = (m.data[3]*m.data[7]-m.data[4]*m.data[6])*t17;
            data[7] = -(m.data[0]*m.data[7]-t12)*t17;
            data[8] = (t4-t8)*t17;
        }

		// invert this
		void invert()
        {
        	*this = getInverse();
        }

		Matrix3 getInverse () const {
			Matrix3 ret;
			ret.setInverse(*this);
			return ret;
		}

		Matrix3 getTranspose () const
		{
			Matrix3 ret;
			
			ret.data[0] = data[0];
			ret.data[1] = data[3];
			ret.data[2] = data[6];
			ret.data[3] = data[1];
			ret.data[4] = data[4];
			ret.data[5] = data[7];
			ret.data[6] = data[2];
			ret.data[7] = data[5];
			ret.data[8] = data[8];
			
			return ret;
		}
		
		//void transpose ()

		// matrix multiplication
		Matrix3 operator*(const Matrix3 &o) const
        {
            return Matrix3(
                data[0]*o.data[0] + data[1]*o.data[3] + data[2]*o.data[6],
                data[0]*o.data[1] + data[1]*o.data[4] + data[2]*o.data[7],
                data[0]*o.data[2] + data[1]*o.data[5] + data[2]*o.data[8],

                data[3]*o.data[0] + data[4]*o.data[3] + data[5]*o.data[6],
                data[3]*o.data[1] + data[4]*o.data[4] + data[5]*o.data[7],
                data[3]*o.data[2] + data[4]*o.data[5] + data[5]*o.data[8],

                data[6]*o.data[0] + data[7]*o.data[3] + data[8]*o.data[6],
                data[6]*o.data[1] + data[7]*o.data[4] + data[8]*o.data[7],
                data[6]*o.data[2] + data[7]*o.data[5] + data[8]*o.data[8]
                );
        }


		// matrix addition
		Matrix3 operator+(const Matrix3 &o) const
		{
			return Matrix3(
			data[0] + o.data[0], data[1] + o.data[1], data[2] + o.data[2],
            data[3] + o.data[3], data[4] + o.data[4], data[5] + o.data[5],
            data[6] + o.data[6], data[7] + o.data[7], data[8] + o.data[8]
			);
		}

		// matrix subtraction
		Matrix3 operator-(const Matrix3 &o) const
		{
			return Matrix3(
			data[0] - o.data[0], data[1] - o.data[1], data[2] - o.data[2],
            data[3] - o.data[3], data[4] - o.data[4], data[5] - o.data[5],
            data[6] - o.data[6], data[7] - o.data[7], data[8] - o.data[8]
			);
		}


		// inplace matrix multiplication
		void operator*=(const Matrix3 &o)
        {
            float t1;
            float t2;
            float t3;

            t1 = data[0]*o.data[0] + data[1]*o.data[3] + data[2]*o.data[6];
            t2 = data[0]*o.data[1] + data[1]*o.data[4] + data[2]*o.data[7];
            t3 = data[0]*o.data[2] + data[1]*o.data[5] + data[2]*o.data[8];
            data[0] = t1;
            data[1] = t2;
            data[2] = t3;

            t1 = data[3]*o.data[0] + data[4]*o.data[3] + data[5]*o.data[6];
            t2 = data[3]*o.data[1] + data[4]*o.data[4] + data[5]*o.data[7];
            t3 = data[3]*o.data[2] + data[4]*o.data[5] + data[5]*o.data[8];
            data[3] = t1;
            data[4] = t2;
            data[5] = t3;

            t1 = data[6]*o.data[0] + data[7]*o.data[3] + data[8]*o.data[6];
            t2 = data[6]*o.data[1] + data[7]*o.data[4] + data[8]*o.data[7];
            t3 = data[6]*o.data[2] + data[7]*o.data[5] + data[8]*o.data[8];
            data[6] = t1;
            data[7] = t2;
            data[8] = t3;
        }

		// inplace scalar multiplication
		void operator*=(const float scalar)
        {
            data[0] *= scalar; data[1] *= scalar; data[2] *= scalar;
            data[3] *= scalar; data[4] *= scalar; data[5] *= scalar;
            data[6] *= scalar; data[7] *= scalar; data[8] *= scalar;
        }
		
		Matrix3 operator* (const float scalar) const
		{
			Matrix3 ret = *this;
			ret *= scalar;
			return ret;
		}
		
		friend Matrix3 operator*(const float scalar, const Matrix3& m);

		// inplace addition
		void operator+=(const Matrix3 &o)
        {
            data[0] += o.data[0]; data[1] += o.data[1]; data[2] += o.data[2];
            data[3] += o.data[3]; data[4] += o.data[4]; data[5] += o.data[5];
            data[6] += o.data[6]; data[7] += o.data[7]; data[8] += o.data[8];
        }

		// inplace subtraction
		void operator-=(const Matrix3 &o)
        {
            data[0] -= o.data[0]; data[1] -= o.data[1]; data[2] -= o.data[2];
            data[3] -= o.data[3]; data[4] -= o.data[4]; data[5] -= o.data[5];
            data[6] -= o.data[6]; data[7] -= o.data[7]; data[8] -= o.data[8];
        }

		bool operator==(const Matrix3 &o)
		{
			return (data[0] == o.data[0] && data[1] == o.data[1] && data[2] == o.data[2]
				&& data[3] == o.data[3] && data[4] == o.data[4] && data[5] == o.data[5]
				&& data[6] == o.data[6] && data[7] == o.data[7] && data[8] == o.data[8]);
		}
		
		bool operator!=(const Matrix3 &o)
		{
			return ! (*this == o);
		}


		float data[9];
		
	};


	class Matrix4

	{


	};

};

inline physics::Matrix2 physics::operator*(const float scalar, const physics::Matrix2& m) {
	return m * scalar;
}

inline physics::Matrix3 physics::operator*(const float scalar, const physics::Matrix3& m) {
	return m * scalar;
}

#endif

