#ifndef _PHYSICS_MOI_
#define _PHYSICS_MOI_

#include "common.h"

#include "matrix.h"
#include "orientation.h"
#include "rotation.h"

// TODO: Alan needs to document this class.

namespace physics {
	
	/** Moment of Inertia (2D) */
	class MOI2D {
		public:
		
		/** Creates a MOI object that prevents rotation. */
		MOI2D () : inverseMOI(0) {}
		
		/** Calculates the moment of inertia for a rigid body. */
		MOI2D (Body2D* body);
		
		/** Return true if this MOI prevents rotation. To be renamed? */
		bool isInfinite () const;
		
		void updateOrientation (const Orientation2D& o) {}
		
		/** Calculate the change of rotation if an impulse were to be applied at a given position. */
		Rotation2D getRotationChange (const Vector2D& impulse, const Vector2D& offset) const;
		
		Matrix2 getVelocityPerImpulse (const Vector2D& offset) const;
		
		private:
		
		float inverseMOI;
	};
	
	/** Moment of Inertia (3D) */
	class MOI3D {
		public:
		
		/** Creates a MOI object that prevents rotation. */
		MOI3D () : infinite(true) {}
		
		/** Calculates the moment of inertia for a rigid body. */
		MOI3D (Body3D* body);
		
		/** Return true if this MOI prevents rotation. To be renamed? */
		bool isInfinite () const;
		
		void updateOrientation (const Orientation3D& o);
		
		/** Calculate the change of rotation if an impulse were to be applied at a given position. */
		Rotation3D getRotationChange (const Vector3D& impulse, const Vector3D& offset) const;
		
		Matrix3 getVelocityPerImpulse (const Vector3D& offset) const;
		
		private:
		
		Matrix3 localInverseInertiaTensor;
		Matrix3 worldInverseInertiaTensor;
		bool infinite;
	};

// namespace physics continues...

inline bool MOI2D::isInfinite () const { return inverseMOI == 0; }

inline Rotation2D MOI2D::getRotationChange (const Vector2D& impulse, const Vector2D& offset) const {
	return (offset.x * impulse.y - offset.y * impulse.x) * inverseMOI;
}

inline Matrix2 MOI2D::getVelocityPerImpulse (const Vector2D& offset) const {
	return Matrix2(
		offset.y * offset.y, -offset.y * offset.x,
		-offset.x * offset.y, offset.x * offset.x
	) * inverseMOI;
}

inline bool MOI3D::isInfinite () const { return infinite; }

inline void MOI3D::updateOrientation (const Orientation3D& o) {
	worldInverseInertiaTensor = o.getInverseMatrix() * localInverseInertiaTensor * o.getMatrix();
}

inline Rotation3D MOI3D::getRotationChange (const Vector3D& impulse, const Vector3D& offset) const {
	if (infinite) {
		return Rotation3D();
	} else {
		return worldInverseInertiaTensor * offset.cross(impulse);
	}
}

inline Matrix3 MOI3D::getVelocityPerImpulse (const Vector3D& offset) const {
	if (infinite) {
		return Matrix3();
	} else {
		// Let's hope this is right... (now actually gives correct results, so that's encouraging)
		Matrix3 torquePerImpulse = Matrix3::skewSymmetric(offset);
		
		Matrix3 m = torquePerImpulse;
		m *= worldInverseInertiaTensor;
		m *= torquePerImpulse;
		m *= -1;
		
		return m;
	}
}

}; // End namespace physics

#endif

