#ifndef _PHYSICS_ORIENTATION_
#define _PHYSICS_ORIENTATION_

#include "common.h"

#include "quaternion.h"

// It's a shame we have to define this twice.
// Is there some better way of doing it?

namespace physics {
	class Vector2D;
	class Rotation2D;
	
	/** Stores an orientation in 2D. */
	class Orientation2D {
		public:
		
		Orientation2D () : value(0.0f) {}
		Orientation2D (float v) : value(v) {}
		
		/** Creates a random orientation. */
		static Orientation2D random ();
		
		/** Return this orientation as an angle of rotation about the z axis. */
		float getAngle () const { return value; }
		
		// TODO: limit value to [-pi,pi] or similar
		// (theoretically it could grow large and become imprecise)
		void normalise () {}
		
		/** Get the negation of this orientation (the same value in the opposite direction). */
		Orientation2D operator- () const;
		
		/** Combine two orientations. */
		Orientation2D operator+ (Orientation2D o) const;
		
		/** Get a relative orientation. */
		Orientation2D operator- (Orientation2D o) const;
		
		/** Rotate this orientation. A Rotation object in this context is just a change of orientation. */
		Orientation2D& operator+= (Rotation2D o);
		
		/** Rotate this orientation in the opposite direction. A Rotation object in this context is just a change of orientation. */
		Orientation2D& operator-= (Rotation2D o);
		
		/** Transform a point by this orientation. */
		Vector2D operator* (const Vector2D& v) const;
		
		/** Rotate a point using this orientation. */
		Vector2D transform (const Vector2D& v) const;
		
		/** Rotate a point in the opposite direction to this orientation. */
		Vector2D inverseTransform (const Vector2D& v) const;
		
		private:
		
		float value;
	};
	
	class Vector3D;
	class Rotation3D;
	class Matrix3;
	
	/** Stores an orientation in 3D. */
	class Orientation3D {
		public:
		
		Orientation3D () {}
		Orientation3D (const Quaternion& q) : quat(q) {}
		
		/** Creates a random orientation. */
		static Orientation3D random ();
		
		Quaternion getQuaternion () const { return quat; }
		
		Matrix3 getMatrix () const;
		
		Matrix3 getInverseMatrix () const;
		
		/** Get this orientation as a rotation about some axis. */
		void getAxisRotation (float& radians, float& x, float& y, float& z);
		
		void normalise () { quat = quat.normalise(); }
		
		/** Get the negation of this orientation (the same value in the opposite direction). */
		Orientation3D operator- () const;
		
		/** Combine two orientations. */
		Orientation3D operator+ (const Orientation3D& o) const;
		
		/** Get a relative orientation. */
		Orientation3D operator- (const Orientation3D& o) const;
		
		/** Rotate this orientation. A Rotation object in this context is just a change of orientation. */
		Orientation3D& operator+= (const Rotation3D& o);
		
		/** Rotate this orientation in the opposite direction. A Rotation object in this context is just a change of orientation. */
		Orientation3D& operator-= (const Rotation3D& o);
		
		/** Transform a point by this orientation. */
		Vector3D operator* (const Vector3D& v) const;
		
		/** Rotate a point using this orientation. */
		Vector3D transform (const Vector3D& v) const;
		
		/** Rotate a point in the opposite direction to this orientation. */
		Vector3D inverseTransform (const Vector3D& v) const;
		
		private:
		
		Quaternion quat;
	};
};

#include "vector.h"
#include "rotation.h"
#include "matrix.h"

#include <stdlib.h>

namespace physics {

//// 2D ////

inline Orientation2D Orientation2D::random () {
	return (rand() % 31415) / 10000.0f;
}

inline Orientation2D Orientation2D::operator- () const { return -value; }

inline Orientation2D Orientation2D::operator+ (Orientation2D o) const { return value + o.value; }

inline Orientation2D Orientation2D::operator- (Orientation2D o) const { return value - o.value; }

inline Orientation2D& Orientation2D::operator+= (Rotation2D o) { value += o.value; return *this; }

inline Orientation2D& Orientation2D::operator-= (Rotation2D o) { value -= o.value; return *this; }

inline Vector2D Orientation2D::operator* (const Vector2D& v) const { return transform(v); }

inline Vector2D Orientation2D::transform (const Vector2D& v) const {
	if (value == 0) { return v; }
	
	float _sin = sin(value);
	float _cos = cos(value);
	
	return Vector2D(_cos * v.x - _sin * v.y,
	                _sin * v.x + _cos * v.y);
}

inline Vector2D Orientation2D::inverseTransform (const Vector2D& v) const {
	if (value == 0) { return v; }
	
	float _sin = sin(-value);
	float _cos = cos(-value);
	
	return Vector2D(_cos * v.x - _sin * v.y,
	                _sin * v.x + _cos * v.y);
}

//// 3D ////

inline Orientation3D Orientation3D::random () {
	float angle = (rand() % 31415) / 10000.0f;
	Vector3D v;
	v.x = (rand() % 10000) / 10000.0f;
	v.y = (rand() % 10000) / 10000.0f;
	v.z = (rand() % 10000) / 10000.0f;
	v = v.normalise();
	return Quaternion::create(angle, v.x, v.y, v.z);
}

inline Matrix3 Orientation3D::getMatrix () const { return Matrix3(quat); }

inline Matrix3 Orientation3D::getInverseMatrix () const { return Matrix3(quat.conjugate()); }

inline Orientation3D Orientation3D::operator- () const { return quat.conjugate(); }

inline Orientation3D Orientation3D::operator+ (const Orientation3D& o) const { return quat * o.quat; }

inline Orientation3D Orientation3D::operator- (const Orientation3D& o) const { return quat * o.quat.conjugate(); }

inline Orientation3D& Orientation3D::operator+= (const Rotation3D& o) { quat.rotate(o.value); return *this; }

inline Orientation3D& Orientation3D::operator-= (const Rotation3D& o) { quat.rotate(-o.value); return *this; }

inline Vector3D Orientation3D::operator* (const Vector3D& v) const { return quat.transform(v); }

inline Vector3D Orientation3D::transform (const Vector3D& v) const { return quat.transform(v); }

inline Vector3D Orientation3D::inverseTransform (const Vector3D& v) const { return quat.inverseTransform(v); }

}; // End namespace physics

#endif

