#include "polygon.h"
#include "body.h"
#include "aarect.h"

using namespace physics;

Vector2D Polygon::getNormal (int i) const {
	Vector2D dir = getVertex(i + 1) - getVertex(i);
	dir = dir.normalise();
	return Vector2D(dir.y, -dir.x);
}

void Polygon::thisMoved () {
	for (int i = 0; i < count; i++) {
		worldPoints[i] = localToWorld(localPoints[i]);
	}
}

float Polygon::getArea () const {
	float area = 0;
	
	for (int i = 0; i < count; i++) {
		int j = i+1;
		if (j == count) { j = 0; }
		area += localPoints[i].x * localPoints[j].y;
		area -= localPoints[j].x * localPoints[i].y;
	}
	area /= 2;
	return area;
}

float Polygon::getMOI () const {
	float momentOfInertia = 0;
	float denom = 0;
	
	for (int i = 0; i < count; i++)	{
		Vector2D a = localPoints[(i+1)%count];
		Vector2D b = localPoints[i];
		
		float l = (Vector3D::crossProduct(a, b)).length();
		momentOfInertia += l * (a.dot(a) + a.dot(b) + b.dot(b));
		denom += l;
	}
	momentOfInertia *= getMass() / (6 * denom);
	return momentOfInertia;
}

Polygon* Polygon::create (Vector2D* points, const int count) {
	Vector2D* hullPoints = new Vector2D[count];
	
	int newCount;
	
	makeConvexHull(points, count, hullPoints, newCount);
	
	if (newCount < 3) {
		delete[] hullPoints;
		return 0;
	}
	
	Polygon* polygon = new Polygon(hullPoints, newCount);
	delete[] hullPoints;
	return polygon;
}

Polygon::Polygon (Vector2D* points, const int count) : count(count) {
	worldPoints.resize(count);
	localPoints.resize(count);
	
	// TODO: calculate centroid correctly: it's more complicated than just the average of all vertices
	Vector2D average;
	
	for (int i = 0; i < count; i++) {
		worldPoints[i] = points[i];
		average += points[i];
	}
	
	average /= count;
	
	for (int i = 0; i < count; i++) {
		localPoints[i] = points[i] - average;
	}
	
	setPosition(average);
}

Vector2D Polygon::getExtremePoint (const Vector2D& direction, int& index) const {
	index = 0;
	
	Vector2D point = worldPoints[0];
	
	float val = point.dot(direction);
	
	for (int i = 1; i < count; i++)	{
		float x = worldPoints[i].dot(direction);
		
		if (x > val) {
			val = x;
			index = i;
			point = worldPoints[i];
		}
	}
	
	return point;// + direction * getRadius();
}

bool Polygon::contains (const Vector2D& p) const {
	if (count < 3) { return false; }
	
	for (int i = 0; i < count; i++)	{
		if (getNormal(i).dot(p - getVertex(i)) > 0)	{
			return false;
		}
	}
	
	return true;
}

bool Polygon::intersects (const AARect &rect) const {
	
	//Does rect contains polygon centre
	if (rect.contains(Polygon::getPosition())) {
		return 1;
	}

	//Does rect contain any polygon vertex
	for (int i = 0; i < Polygon::getVertexCount(); i++) {
		if (rect.contains(Polygon::getVertex(i))) {
			return 1;
		}
	}

	//Does polygon contain rect centre
	//TODO: Uncomment when AARect becomes a member of object class
	//if (Polygon::contains(rect.getPosition())) {
	//	return 1;
	//}

	//Does polygon contain any rect vertex
	for (int i = 0; i < 4; i++) {
		if (Polygon::contains(rect.getVertex(i))) {
			return 1;
		}
	}

	return 0;
}

// returns twice the signed triangle area
// assuming origin is in the bottom left corner
// the result is < 0 if points are clockwise and > 0 if anticlockwise
// a result of 0 means they are colinear
inline float signedArea (const Vector2D& a, const Vector2D& b, const Vector2D& c) {
    return (a.x-b.x)*(b.y-c.y) - (b.x-c.x)*(a.y-b.y);
}

inline bool clockwise (const Vector2D& a, const Vector2D& b, const Vector2D& c) {
	return signedArea(a, b, c) < -0.00001f;
}

int convexHullNextPoint (const Vector2D *const points, const int count, const int last) {
	// candidate for the next point
	int next = 0;
	
	// check all points
	for (int i = 1; i < count; i++)	{
		if (i == last) { continue; }
		
		if (! clockwise(points[last], points[i], points[next])) {
			// next cannot be the next point, so replace it with i.
			next = i;
		}
	}
	
	return next;
}

void Polygon::makeConvexHull (Vector2D* oldPoints, const int count, Vector2D* newPoints, int& newCount) {
	// for fewer than 3 points, the output is the input
	if (count < 3) {
		if (count == 2 && oldPoints[0] == oldPoints[1]) {
			newCount = 1;
		} else {
			newCount = count;
		}
		
		for (int i = 0; i < count; i++) {
			newPoints[i] = oldPoints[i];
		}
		return;
	}
	
	// here is just the index of the point we're currently working on
	unsigned int here = 0;
	
	// find the point with smallest x
	for (int i = 1; i < count; i++) {
		if (oldPoints[i].x < oldPoints[here].x) {
			here = i;
		}
	}
	
	newPoints[0] = oldPoints[here];
	
	newCount = 0;
	
	// gift-wrapping algorithm
	while (true) {
		here = convexHullNextPoint(oldPoints, count, here);
		newCount++;
		if (oldPoints[here] == newPoints[0]) { break; }
		newPoints[newCount] = oldPoints[here];
	}
}

int Polygon::makeConvexHull (Vector2D* points, const int count) {
	Vector2D* newPoints = new Vector2D[count];
	
	int newCount = count;
	
	makeConvexHull(points, count, newPoints, newCount);
	
	for (int i = 0; i < newCount; i++) {
		points[i] = newPoints[i];
	}
	delete[] newPoints;	
	return newCount;
}
