#ifndef _PHYSICS_POLYGON_
#define _PHYSICS_POLYGON_

#include "common.h"

#include "area.h"
#include <vector>

namespace physics {
	/**	\class Polygon 
	*	Convex polygon class 
	*/
	class Polygon : public Area
	{
		public:
		//! Create a polygon. Performs convexity checking
		static Polygon* create (Vector2D* points, const int size);
		
		float getRadius () const { return 0; }
		
		bool contains (const Vector2D &p) const;
		bool intersects (const AARect& rect) const;
		
		Vector2D getNormal (int i) const;
		Vector2D getVertex (int i) const { return worldPoints[i % count]; }
		
		int getVertexCount () const { return worldPoints.size(); }
		
		Vector2D getExtremePoint (const Vector2D& dir, int& i) const;
		
		float getArea () const;
		
		//! Get the Moment of Inertia
		/*! \return Float; Moment of Inertia
		*/
		float getMOI () const;
		
		Type type () const { return POLYGON; }
		
		//! Scale the shape
		/*! \param scale Float specifying the value by which to scale
		*/
		void scale(float scale)
		{
			for (int i = 0; i < count; i++) {
				localPoints[i] = localPoints[i] * scale;
				worldPoints[i] = localToWorld(localPoints[i]);
			}
		} 
		
		/**
		 * Given an array of points, find the convex hull.
		 * 
		 * @return Number of points on the convex hull
		 * (values in the array past this length will be unchanged).
		 */
		static int makeConvexHull (Vector2D* points, const int count);
		
		static void makeConvexHull (Vector2D* oldPoints, const int count, Vector2D* newPoints, int& newCount);
		
		template <typename Container>
		static void makeConvexHull (Container& c);
		
		private:
		
		/*! Constructor
		* The constructor does no convexity checking: use Polygon::create() instead.
		*/
		Polygon (Vector2D* points, const int count);
		
		virtual void thisMoved ();
		
		int count;
		std::vector<Vector2D> localPoints;
		std::vector<Vector2D> worldPoints;

	};
};

template <typename Container>
void physics::Polygon::makeConvexHull (Container& c) {
	int count = c.size();
	physics::Vector2D* points = new physics::Vector2D[count];
	
	int i = 0;
	
	for (typename Container::iterator it = c.begin(); it != c.end(); it++) {
		points[i] = *it;
		i++;
	}
	
	count = makeConvexHull(points, count);
	
	c.assign(points, points + count);
	
	delete[] points;
}
		
#endif

