#include "contact_resolver.h"

#include "contact_list.h"

#include "vector.h"
#include "body.h"
#include "shape.h"

#include <boost/foreach.hpp>

using namespace physics;

inline float Min(float a, float b)
{
	return a < b ? a : b;
}

inline float Max(float a, float b)
{
	return a > b ? a : b;
}

inline float _Clamp(float a, float low, float high)
{
	return Max(low, Min(a, high));
}

template <typename d>
void CattoPositionResolver<d>::resolve (ContactList<d>* contacts, std::vector<Joint*>* joints)
{
	contacts->preparePenetrations();
	
	contacts->foreach(prepare);

	for (int i = 0; i < iterations; i++) {
		
		contacts->foreach(resolve);
		
	}
	
}

template <typename d>
void CattoPositionResolver<d>::prepare (Contact* c)
{
	c->bias = -0.8f * Min(0.0f, Globals::penetrationEpsilon - c->penetration);
	
	if (c->penetration > 0)
	{
		// Shouldn't we be applying the bias impulse here?
		// Catto doesn't seem to in Box2D_Lite and it seems to
		// break things, but I don't understand why.
	} else {
		c->Pnb = 0;
	}
}

template <typename d>
void CattoPositionResolver<d>::resolve (Contact* c)
{
	if (c->penetration < 0) { return; }

	Contact& data = *c;
	
	bool body0awake = data.body0->isAwake();
	bool body1awake = data.body1->isAwake();
	
	if (!body0awake && !body1awake) { return; }
	
	if (c->contactToWorld == Matrix()) {
		// How can this happen?
		//float foo = 1.0f / 0.0f;
		//std::cout << "Confusion\n";
		return;
	}
	
	Body* b1 = c->body0;
	Body* b2 = c->body1;
	
	// Relative velocity at contact
	Vector dv = b2->getBiasVelocityAtPoint(c->position) - b1->getBiasVelocityAtPoint(c->position);
	float vnb = dv.dot(c->normal);

	float dPnb = c->massNormal * (-vnb + c->bias);
	float Pnb0 = c->Pnb;
	c->Pnb = Max(Pnb0 + dPnb, 0.0f);
	dPnb = c->Pnb - Pnb0;

	Vector Pb = dPnb * c->normal;
	
	if (b1->isAwake()) b1->applyBiasImpulse(-Pb, c->position);
	if (b2->isAwake()) b2->applyBiasImpulse(Pb, c->position);
}

template class CattoPositionResolver<TwoDee>;
template class CattoPositionResolver<ThreeDee>;

