#include "contact_resolver.h"

#include "contact_list.h"
#include "contact_list_base.h"

#include "vector.h"
#include "body.h"
#include "shape.h"

#include <boost/foreach.hpp>

using namespace physics;

template <typename d>
ContactList<d>* StupidPositionResolver<d>::contacts = 0;

template <typename d>
StupidPositionResolver<d>::StupidPositionResolver (int iterationCount, bool sort) : sortContacts(sort), iterations(iterationCount) {
	if (iterations <= 0) {
		if (sortContacts) {
			iterations = 3;
		} else {
			iterations = 10;
		}
	}
}

// NB: Only called when sortContacts is false
template <typename d>
void StupidPositionResolver<d>::resolve (Contact* contact)
{
	if (contact->penetration < Globals::penetrationEpsilon) { return; }
	
	float moveLinear[2];
	
	resolve(*contact, moveLinear);
	contacts->updatePenetrations(contact, moveLinear);
}

template <typename d>
void StupidPositionResolver<d>::resolve (ContactList<d>* contacts, std::vector<Joint*>* joints)
{
	this->contacts = contacts;
	
	contacts->preparePenetrations();
	
	if (sortContacts) {
		int limit = (int) (contacts->size() * iterations);
		
		float moveLinear[2];
		
		for (int i = 0; i < limit; i++)
		{
			Contact* data = contacts->getGreatestPenetration();
			
			if (data->penetration < Globals::penetrationEpsilon) { break; }
			
			resolve(*data, moveLinear);
			contacts->updatePenetrations(data, moveLinear);
		}
	} else {
		for (int i = 0; i < iterations; i++) {
			contacts->foreach(resolve);
		}
	}
}

template <typename d>
void StupidPositionResolver<d>::resolve (Contact& data, float* moveLinear)
{
	Body* body1 = data.body[0];
	Body* body2 = data.body[1];
	
	float penetration = data.penetration;
	
	if (data.inverseMass[0] == 0) {
		moveLinear[0] = 0.0f;
		moveLinear[1] = penetration;
	} else if (data.inverseMass[1] == 0) {
		moveLinear[0] = -penetration;
		moveLinear[1] = 0.0f;
	} else {
		float totalMass = 1 / data.inverseMass[0] + 1 / data.inverseMass[1];
		moveLinear[0] = -penetration / data.inverseMass[1] / totalMass;
		moveLinear[1] = penetration / data.inverseMass[0] / totalMass;
	}
	
	// null body check
	if (body1) body1->translate(data.normal * moveLinear[0]);
	if (body2) body2->translate(data.normal * moveLinear[1]);
}

template class StupidPositionResolver<TwoDee>;
template class StupidPositionResolver<ThreeDee>;

