#include "profiler.h"

#include <stdlib.h>
#include <fstream>
#include <sstream>
#include <iostream>

using namespace physics;

physics::Profiler physics::Profiler::profiler;

Profiler::Profiler (){
	numWrites = 1;
	lastCompleteFrame = 0;
	
	// Default option values.
	enabled = false;
	
	outputFilename = "outputData";
	outputPath = "";
	fileOutput = false;
	graphOutput = false;
	stdOutput = false;
	
	framesToStore = 1;
}

void Profiler::startStep () //called when a new frame begins processing physics
{
	if (enabled) {
		ProfilerFrameData newFrame;  //declare a new frame data object
		frameData.push_back(newFrame);    //...and add it to the list
	
		if ((framesToStore != 0) && (frameData.size() > framesToStore)) frameData.pop_front(); //lose the oldest frame if frameData is getting too large.
	
		stepTimer.reset();
	}
}

void Profiler::endStep () //called when a new frame finishes processing physics
{
	if (enabled) {
		frameData.back().setFrameTime(stepTimer.getMicroTime());
		lastCompleteFrame = &(frameData.back());
	}
}

void Profiler::startMethod(std::string methodName) //called when some arbitrary function begins processing
{
	if (enabled) stepTimer.addMarker(methodName);
}

void Profiler::endMethod(std::string methodName) //called when some arbitrary function finishes processing
{
	if (enabled && !frameData.empty()) {
		frameData.back().setVar(methodName, stepTimer.getMicroTime(methodName));
		//if (frameData.size() < 2) 
			varsTracked.insert(methodName);
	}
	
}

void Profiler::setFrameVar(std::string varName, int value) {
	if (enabled && !frameData.empty()) {
		frameData.back().setVar(varName, value);
		//if (frameData.size() < 2) 
			varsTracked.insert(varName);
	}
}

void Profiler::end(){
	outputToFile();
	outputToStdOut();
	plotGraphs();
	frameData.clear();
	lastCompleteFrame = 0;
}


void Profiler::outputToFile()
{
	if (enabled && fileOutput) {
		std::ostringstream buffer;
		buffer << outputPath << outputFilename << "_" << numWrites << ".csv";
		
		outputFile.open(buffer.str().c_str(), std::ios::out);

		if ( !outputFile.is_open() ) {
			std::cerr << "Error: Attempt to log profiler data failed. There was a problem opening the file " << outputFilename << " for output." << std::endl;
		} else {
		
			//Add headers to the output file...
			outputFile << "#, 'Step Time'";
			for (std::set<std::string>::iterator varsTrackedIterator = varsTracked.begin(); varsTrackedIterator != varsTracked.end() ; varsTrackedIterator++)
			{
				outputFile << ",'" << *varsTrackedIterator << "'";
			}
			outputFile << std::endl;
		
			//Then write the data.
			int step = 1;
			for (std::list<ProfilerFrameData>::iterator frameDataIterator = frameData.begin(); frameDataIterator != frameData.end() ; frameDataIterator++)
			{
				outputFile << step << "," << frameDataIterator->getFrameTime();
				for (std::set<std::string>::iterator varsTrackedIterator = varsTracked.begin(); varsTrackedIterator != varsTracked.end() ; varsTrackedIterator++)
				{
					outputFile << "," << frameDataIterator->getVar(*varsTrackedIterator);
				}
				outputFile << std::endl;
				step++;
			}
			
			outputFile.close();
			numWrites++;
		}
	}
}

void Profiler::plotGraphs()
{
	if (enabled && graphOutput) {
		#ifdef WIN32
		int result = system("gnuplot gpfile"); 
		if (result != 0) {
			std::cerr << "Problem plotting graphs. Is gnuplot on the PATH? Is gpfile in the active directory?" << std::endl;
		}
		#else
		std::ostringstream buffer;
		buffer << "cat gpfile | sed \"s/outputData_1/outputData_" << numWrites-1 << "/g\" | sed \"s/run-profile_1/run-profile_" << numWrites-1 << "/g\" | sed \"s/extime-contacts-bodies_1/extime-contacts-bodies_" << numWrites-1 << "/g\" | gnuplot";
		int result = system(buffer.str().c_str()); 
		if (result != 0) {
			std::cerr << "Problem plotting graphs. Are gnuplot and sed on the PATH? Is gpfile in the active directory?" << std::endl;
		}
		#endif
	}
}

void Profiler::outputToStdOut()
{
	if (enabled && stdOutput) {
		std::cout << "PROFILER: ";
		std::cout << frameData.size() << " total steps logged. ";
		std::cout << meanStepTime() << " s/step average over all these steps." << std::endl;
	}
}

int Profiler::getLastFrameTime() { 
	if (enabled) {
		if (lastCompleteFrame != 0) return lastCompleteFrame->getFrameTime();
		else return 0;
	} else return -1;
}

int Profiler::getLastFrameVar(std::string varName) { 
	if (enabled) {
		if (lastCompleteFrame != 0) return lastCompleteFrame->getVar(varName);
		else return 0;
	} else return -1;
}


double Profiler::meanStepTime()
{
	if (enabled) {
		int sum = 0;
	
		for (std::list<ProfilerFrameData>::iterator frameDataIterator = frameData.begin(); frameDataIterator != frameData.end() ; frameDataIterator++)
		{
			sum += frameDataIterator->getFrameTime();
		}
		double divisor = double(frameData.size() * 1000000);

		return sum / (divisor == 0 ? 1 : divisor);
	}
	else return -1;
}

double Profiler::meanStepTime(int lastN)
{
	if (enabled) {	
		int sum = 0;
		int counter = 0;
	
		for (std::list<ProfilerFrameData>::reverse_iterator frameDataIterator = frameData.rbegin(); frameDataIterator != frameData.rend() && counter < lastN ; frameDataIterator++)
		{
			sum += frameDataIterator->getFrameTime();
			counter++;
		}
		double divisor = double(counter * 1000000);

		return sum / (divisor == 0 ? 1 : divisor);
	}
	else return -1;
}

