#ifndef _PHYSICS_PROFILER_
#define _PHYSICS_PROFILER_

#include "common.h"

#include <map>
#include <set>
#include <string>
#include <fstream>
#include <list>
#include "microtimer.h"

/*********************************************************************************\
This defines the Profiler class. The profiler is told when certain events occur
(such as the beginning and end of physics processing, and sub-tasks within that
processing), and uses a MicroTimer to to accurately work out how long they took.

These times are stored individually for each frame, in a list containing per-frame
data in objects called ProfilerFrameData.

\*********************************************************************************/

namespace physics {
	
	class ProfilerFrameData { //This class stores all data the profiler has on a single frame. At the moment, it only stores integers, but that can be altered if it turns out to be insufficient.
		public:
			
			ProfilerFrameData () {}
			~ProfilerFrameData () {}
			
			void setFrameTime(int usecs) {
				frameTime = usecs;
			}
			
			int getFrameTime() {
				return frameTime;
			}
			
			void setVar(std::string varName, int value) { //Store some arbitrary integer value, called varName.
				data[varName] = value;
			}
			
			int getVar(std::string varName) { //Get the previously stored arbitrary integer value called varName.
				if (data.find(varName) != data.end()) return data[varName]; //Return the value if we're actually storing it...
				else return 0;                                            //... or 0 if we aren't.
			}
					
		private:
			
			int frameTime;
			std::map<std::string, int> data;
	};
	
	class Profiler {
		public:
			
			static Profiler profiler; //Profiler instance accessible anywhere using the physics namespace...
			
			Profiler () ;
			~Profiler () {}
			
			void startStep (); //called when a step begins
			void endStep (); //called when a step ends
			
			void setFrameVar(std::string, int) ; //Store some arbitrary integer value - called varName - associated with the current frame.
			
			void startMethod(std::string); //called when an arbitrary method begins
			void endMethod(std::string); //called when an arbitrary method ends
			
			void outputToFile(); //outputs profiler data in csv format.
			void outputToStdOut(); //outputs human-readable stats to stdOut.
			void plotGraphs(); //outputs some nice graphs of profiler data.
			
			double meanStepTime (); //returns the mean time the entire current run has taken to execute a step, in seconds.
			double meanStepTime (int); //returns the mean time the last n steps have taken to execute a step, in seconds.
			
			int getLastFrameTime() ; //returns the physics time for the last frame in microseconds.
			int getLastFrameVar(std::string) ; //returns the value of a variable for the last frame.
			
			void end(); //Called when the world ends. Outputs logged data to a file, and wipes stored data.
			
			
			/* OPTIONS METHODS */
			void setFramesToStore(int newValue) { framesToStore = newValue; }
			void setOutputFilename(std::string newValue) { outputFilename = newValue; }
			
			void enable() { enabled = true; }
			void disable() { enabled = false; }
			bool isEnabled() { return enabled; }
			
			void enableFileOutput() { fileOutput = true; }
			void disableFileOutput() { fileOutput = false; graphOutput = false;}
			
			void enableStdOutput() { stdOutput = true; }
			void disableStdOutput() { stdOutput = false; }
			
			void enableGraphOutput() { if(fileOutput) graphOutput = true; }
			void disableGraphOutput() { graphOutput = false; }
			
		private:
			MicroTimer stepTimer;
			
			std::list<ProfilerFrameData> frameData;
			std::set<std::string> varsTracked;
			ProfilerFrameData * lastCompleteFrame;
			
			int numWrites; //How many times the profiler has written to a file this run.
			
			std::ofstream outputFile; //The output stream for csv-formatted profiler data
			
			/* PROFILER OPTIONS */
			bool enabled;                	//Whether the profiler is currently logging data or not.
			unsigned int framesToStore;           //How many frames the profiler should keep data on. Set to 0 for unlimited.
			std::string outputFilename;           //What output filename the profiler should write data to.
			std::string outputPath;           	//What path the profiler should write data to.
			bool fileOutput;
			bool stdOutput;
			bool graphOutput;
	};
};

#endif

