#ifndef _PHYSICS_QUATERNION_
#define _PHYSICS_QUATERNION_

#include "common.h"

#include <math.h>

// This wasn't as complete as I thought: some methods need to be optimised/checked/implemented.

namespace physics {
	
	/**
	 * Alternative for representing rotations. Advantages are that it does not suffer
	 * from gimble lock and produces much better interpolation.
	 */
	class Quaternion
	{
		public:
		
		/** Create a non-oriented quaternion */
		Quaternion ()
			: x(0.0), y(0.0), z(0.0), w(1.0) {}
		
		/** Create a quaternion from x/y/z/w components. */
		Quaternion (float x, float y, float z, float w)
			: x(x), y(y), z(z), w(w) {}
		
		~Quaternion() {}
		
		/**
		 * Create a quaternion from a rotation about an axis.
		 * e.g. Quaternion::create(2.0, 1.0, 0.0, 0.0) = 2 radians around X.
		 * Note: x/y/z _must_ be normalised.
		 */
		static Quaternion create (float radians, float x, float y, float z)
		{
			Quaternion q;
			
			float res = sin(radians / 2.0f);
			
			q.x = x*res;
			q.y = y*res;
			q.z = z*res;
			q.w = cos(radians / 2.0f);
			
			return q;
		}
		
		/** Get this quaternion as a rotation about some axis. */
		// TODO
		void getAxisRotation (float& radians, float& x, float& y, float& z);
		
		/** Scale the quaternion. */
		Quaternion operator* (float s) const {
			return Quaternion(x*s, y*s, z*s, w*s);
		}
	
		/** Scale the quaternion. */
		Quaternion& operator*= (float s) {
			x *= s;
			y *= s;
			z *= s;
			w *= s;
			
			return *this;
		}
	
		/** Add two quaternions together */
		Quaternion operator+ (const Quaternion &q) const {
			return Quaternion(x+q.x, y+q.y, z+q.z, w+q.w);
		}
	
		/** Add two quaternions together */
		Quaternion& operator+= (const Quaternion &q) {
			x += q.x;
			y += q.y;
			z += q.z;
			w += q.w;
			return *this;
		}
		
		/**
		 * Multiply two quaternions together. This combines the rotation as would
		 * combining two rotation matrices.
		 */
		Quaternion operator* (const Quaternion& q) const
		{
			return Quaternion(	w*q.x + x*q.w + y*q.z - z*q.y,
								w*q.y + y*q.w + z*q.x - x*q.z,
								w*q.z + z*q.w + x*q.y - y*q.x,
								w*q.w - x*q.x - y*q.y - z*q.z);
		}
		
		/**
		 * Multiply two quaternions together. This combines the rotation as would
		 * combining two rotation matrices.
		 */
		// TODO
		Quaternion& operator*= (const Quaternion& q);
		
		/**
		 * Interpolate between two quaternion rotations.
		 * @param a Start quaternion (t = 0).
		 * @param b End quaternion (t = 1).
		 * @param t Interpolation value between 0 and 1.
		 */
		static Quaternion slerp (const Quaternion& a, const Quaternion& b, float t) {
			float w1;
			float w2;
			float dot = a.x*b.x+a.y*b.y+a.z*b.z+a.w*b.w;
			float theta = acos(dot);
			float sintheta = sin(theta);
	
			if (sintheta > 0.001)
			{
				w1 = sin((1.0f-t)*theta) / sintheta;
				w2 = sin(t*theta) / sintheta;
			} else {
				w1 = 1.0f - t;
				w2 = t;
			}
	
			return a*w1+b*w2;
		}
	
		/** Conjugate, a kind of inverse. */
		Quaternion conjugate () const {
			return Quaternion(-x, -y, -z, w);
		}
		
		Quaternion normalise () const {
			float d = x*x + y*y + z*z + w*w;
			
			if (d == 0) {
				return Quaternion();
			}
			
			d = 1.0f / sqrt(d);
			
			return (*this) * d;
		}
		
		// TODO: check this is correct
		void rotate (const Vector3D& r) {
			*this += Quaternion(r.x, r.y, r.z, 0) * 0.5f * (*this);
		}
		
		/** Transform a point by the rotation represented by this quaternion. */
		// TODO: check this is correct
		// TODO: optimise
		Vector3D transform (const Vector3D& p) const {
			Quaternion q = (*this) * Quaternion(p.x, p.y, p.z, 0) * conjugate();
			
			return Vector3D(q.x, q.y, q.z);
		}
		
		/** Transform a point by the inverse of the rotation represented by this quaternion. */
		// TODO: check this is correct
		// TODO: optimise
		Vector3D inverseTransform (const Vector3D& p) const {
			Quaternion q = conjugate() * Quaternion(p.x, p.y, p.z, 0) * (*this);
			
			return Vector3D(q.x, q.y, q.z);
		}
		
		float x;
		float y;
		float z;
		float w;
	};
};

#endif
