#ifndef _PHYSICS_ROTATION_
#define _PHYSICS_ROTATION_

#include "common.h"

#include <cmath>

// It's a shame we have to define this twice.
// We could do it with templates, but would it be worth it?

namespace physics {
	
	/**
	 * Stores a rotation in 2D.
	 * 
	 * The term "rotation" almost always means "angular velocity" (rate of change of orientation).
	 * However, note that when integrated over time (i.e. multiplied by a duration), we instead get
	 * an absolute change of orientation, which we use to update an orientation object.
	 */
	class Rotation2D {
		public:
		
		Rotation2D () : value(0.0f) {}
		Rotation2D (float v) : value(v) {}
		
		float toFloat () const { return value; }
		
		float getValue () const { return value; }
		
		float scalarSize() { return std::abs(value); }
		
		Rotation2D& operator+= (Rotation2D r) { value += r.value; return *this; }
		Rotation2D operator+ (Rotation2D r) const { return value + r.value; }
		
		Rotation2D& operator-= (Rotation2D r) { value -= r.value; return *this; }
		Rotation2D operator- (Rotation2D r) const { return value - r.value; }
		
		Rotation2D operator- () const { return -value; }
		
		Rotation2D& operator*= (float t) { value *= t; return *this; }
		Rotation2D operator* (float t) const { return value * t; }
		
		friend Rotation2D operator* (float t, Rotation2D r) { return r.value * t; }
		
		/**
		 * In a rotating system, find the instantaneous velocity of a point.
		 */
		Vector2D getVelocityAtPoint (const Vector2D& p) const {
			return Vector2D(-value * p.y, value * p.x);
		}
		
		private:

		friend class Orientation2D;
		
		float value;
	};

	/**
	 * Stores a rotation in 3D.
	 * 
	 * The term "rotation" almost always means "angular velocity" (rate of change of orientation).
	 * However, note that when integrated over time (i.e. multiplied by a duration), we instead get
	 * an absolute change of orientation, which we use to update an orientation object.
	 */
	class Rotation3D {
		public:
		
		Rotation3D () : value() {}
		Rotation3D (const Vector3D& v) : value(v) {}
		
		Vector3D toVector () const { return value; }
		
		Vector3D getValue () const { return value; }
		
		float scalarSize() { return value.length(); }
		
		Rotation3D& operator+= (const Rotation3D& r) { value += r.value; return *this; }
		Rotation3D operator+ (const Rotation3D& r) const { return value + r.value; }
		
		Rotation3D& operator-= (const Rotation3D& r) { value -= r.value; return *this; }
		Rotation3D operator- (const Rotation3D& r) const { return value - r.value; }
		
		Rotation3D operator- () const { return -value; }
		
		Rotation3D& operator*= (float t) { value *= t; return *this; }
		Rotation3D operator* (float t) const { return value * t; }
		
		friend Rotation3D operator* (float t, const Rotation3D& r) { return r.value * t; }
		
		/**
		 * In a rotating system, find the instantaneous velocity of a point.
		 */
		Vector3D getVelocityAtPoint (const Vector3D& p) const {
			return value.cross(p);
		}
		
		private:
		
		friend class Orientation3D;
		
		Vector3D value;
	};
};

#endif

