#include "contact_resolver.h"

#include "contact_list.h"

#include "vector.h"
#include "body.h"
#include "shape.h"

#include <boost/foreach.hpp>

using namespace physics;

inline float Min(float a, float b)
{
	return a < b ? a : b;
}

inline float Max(float a, float b)
{
	return a > b ? a : b;
}

inline float _Clamp(float a, float low, float high)
{
	return Max(low, Min(a, high));
}

template <typename d>
void CattoVelocityResolver<d>::prepare (Contact* c)
{
	float k_biasFactor = 0 * 60; // 60 is the inverse timestep

	c->bias = -k_biasFactor * Min(0.0f, Globals::penetrationEpsilon - c->penetration);
	
	if (c->penetration > 0)
	{
		// Apply normal + friction impulse
		Vector P = c->Pn * c->normal + c->contactToWorld * c->frictionImpulse;
		
		if (c->body0->isAwake()) c->body0->applyImpulse(-P, c->position, false);
		if (c->body1->isAwake()) c->body1->applyImpulse(P, c->position, false);
	} else {
		c->Pn = 0;
		c->frictionImpulse = Vector();
		c->Pnb = 0;
	}
}

template <typename d>
void CattoVelocityResolver<d>::resolve (ContactList<d>* contacts, std::vector<Joint*>* joints)
{
	contacts->prepareVelocities();
	
	contacts->foreach(prepare);
	
	
	BOOST_FOREACH(Joint* joint, *joints)
	{
		joint->prepare();
	}

	for (int i = 0; i < iterations; i++) {
		
		contacts->foreach(resolve);

		
		BOOST_FOREACH(Joint* joint, *joints)
		{
			joint->applyImpulses();
		}

	}
	
}

template <typename d>
void CattoVelocityResolver<d>::applyFriction (physics::Contact<TwoDee>* c) {
	Body2D* b1 = c->body0;
	Body2D* b2 = c->body1;
	
	float massTangent = 1.0f / (c->velocityPerImpulse * Vector(0, 1)).y;
	
	float vt = c->getRelativeContactVelocity().y;
	float dPt = massTangent * (-vt);
	
	float friction = sqrt(c->shape0->getFriction() * c->shape1->getFriction());
	
	// Compute friction impulse
	float maxPt = friction * c->Pn;

	// Clamp friction
	float oldTangentImpulse = c->frictionImpulse.y;
	c->frictionImpulse.y = _Clamp(oldTangentImpulse + dPt, -maxPt, maxPt);
	dPt = c->frictionImpulse.y - oldTangentImpulse;

	// Apply contact impulse
	Vector Pt = c->contactToWorld * Vector2D(0, dPt);

	if (b1->isAwake()) b1->applyImpulse(-Pt, c->position, false);
	if (b2->isAwake()) b2->applyImpulse(Pt, c->position, false);
}

template <typename d>
void CattoVelocityResolver<d>::applyFriction (physics::Contact<ThreeDee>* c) {
	Body3D* b1 = c->body0;
	Body3D* b2 = c->body1;
	
	Vector3D v = c->getRelativeContactVelocity();
	
	float massTangent1 = 1.0f / (c->velocityPerImpulse * Vector3D(0, 1, 0)).y;
	float massTangent2 = 1.0f / (c->velocityPerImpulse * Vector3D(0, 0, 1)).z;
	
	// Shouldn't these be negated?
	Vector3D dPt;
	dPt.y = massTangent1 * (-v.y);
	dPt.z = massTangent2 * (-v.z);
	
	float friction = sqrt(c->shape0->getFriction() * c->shape1->getFriction());
	
	// Compute friction impulse
	float maxPt = Max(friction * c->Pn, 0);

	// Clamp friction
	Vector3D oldImpulse = c->frictionImpulse;
	
	Vector3D totalImpulse = oldImpulse + dPt;
	
	Vector3D tangent = totalImpulse.normalise();
	
	float size = totalImpulse.length();
	
	c->frictionImpulse = tangent * _Clamp(size, -maxPt, maxPt);
	dPt = c->frictionImpulse - oldImpulse;

	// Apply contact impulse
	Vector3D Pt = c->contactToWorld * dPt;

	if (b1->isAwake()) b1->applyImpulse(-Pt, c->position, false);
	if (b2->isAwake()) b2->applyImpulse(Pt, c->position, false);
}

template <typename d>
void CattoVelocityResolver<d>::resolve (Contact* c)
{
	if (c->penetration < 0) { return; }

	Contact& data = *c;
	
	bool body0awake = data.body0->isAwake();
	bool body1awake = data.body1->isAwake();
	
	if (!body0awake && !body1awake) { return; }
	
	if (c->contactToWorld == Matrix()) {
		// How can this happen?
		//float foo = 1.0f / 0.0f;
		//std::cout << "Confusion\n";
		return;
	}
	
	Body* b1 = c->body0;
	Body* b2 = c->body1;
	
	// Relative velocity at contact
	Vector dv = b2->getVelocityAtPoint(c->position) - b1->getVelocityAtPoint(c->position);

	// Compute normal impulse
	float vn = dv.dot(c->normal);
	
	if (Globals::useRestitution && c->approachSpeed > Globals::restitutionThreshold) {
		float restitution = 0.5f * (data.shape[0]->getRestitution() + data.shape[1]->getRestitution());
		
		float separating = restitution * c->approachSpeed;
		
		vn -= separating;
	}
	
	float dPn = c->massNormal * (-vn + c->bias);

	// Clamp the accumulated impulse
	float Pn0 = c->Pn;
	c->Pn = Max(Pn0 + dPn, 0.0f);
	dPn = c->Pn - Pn0;

	// TODO: remove magic number
	if (c->Pn > 2 * c->massNormal) {
		//Check if either body needs waking up.
		
		//Wake the sleeping one if only one is asleep and the other is not static.
		if ((body0awake) ^ (body1awake)) {
			if (body0awake) data.body1->setAwake(true);
			else data.body0->setAwake(true);
		}
	}
	
	// Apply contact impulse
	Vector Pn = dPn * c->normal;

	if (b1->isAwake()) b1->applyImpulse(-Pn, c->position, false);
	if (b2->isAwake()) b2->applyImpulse(Pn, c->position, false);

	applyFriction(c);
}

template class CattoVelocityResolver<TwoDee>;
template class CattoVelocityResolver<ThreeDee>;

