#include "contact_resolver.h"

#include "contact_list.h"
#include "contact_list_base.h"

#include "vector.h"
#include "body.h"
#include "shape.h"

#include <boost/foreach.hpp>

using namespace physics;

template <typename d>
ContactList<d>* StupidVelocityResolver<d>::contacts = 0;

template <typename d>
StupidVelocityResolver<d>::StupidVelocityResolver (int iterationCount, bool sort) : sortContacts(sort), iterations(iterationCount) {
	if (iterations <= 0) {
		if (sortContacts) {
			iterations = 3;
		} else {
			iterations = 10;
		}
	}
}

// NB: Only called when sortContacts is false
template <typename d>
void StupidVelocityResolver<d>::resolve (Contact* contact)
{
	if (contact->approachSpeed < Globals::penetrationEpsilon) { return; }
	
	resolve(*contact);
	contacts->updateVelocities(contact);
}

template <typename d>
void StupidVelocityResolver<d>::resolve (ContactList<d>* contacts, std::vector<Joint*>* joints)
{
	this->contacts = contacts;
	
	contacts->prepareVelocities();
	
	if (sortContacts) {
		int limit = (int) (contacts->size() * iterations);
		
		for (int i = 0; i < limit; i++)
		{
			Contact* data = contacts->getGreatestVelocity();
			
			if (data->approachSpeed < Globals::penetrationEpsilon) { break; }
			//if (data->penetration < 0) { break; }
			
			resolve(*data);
			contacts->updateVelocities(data);
		}
	} else {
		for (int i = 0; i < iterations; i++) {
			contacts->foreach(resolve);
		}
	}
}

template <typename d>
void StupidVelocityResolver<d>::applyFriction (Vector2D& impulseContact, const Matrix2& velocityPerImpulse, float friction, float desiredDeltaVelocity) {
	// Check for exceeding friction.
	float planarImpulse = fabs(impulseContact.y);
	
	if (planarImpulse > impulseContact.x * friction)
	{
		// We need to use dynamic friction.
		impulseContact.y /= planarImpulse;
		impulseContact.x = velocityPerImpulse.data[0] + velocityPerImpulse.data[1] * friction * impulseContact.y;
		impulseContact.x = desiredDeltaVelocity / impulseContact.x;
		impulseContact.y *= friction * impulseContact.x;
	}
}

template <typename d>
void StupidVelocityResolver<d>::applyFriction (Vector3D& impulseContact, const Matrix3& velocityPerImpulse, float friction, float desiredDeltaVelocity) {
	// Check for exceeding friction.
	float planarImpulse = sqrt(impulseContact.y*impulseContact.y +
									impulseContact.z*impulseContact.z);
	
	if (planarImpulse > impulseContact.x * friction)
	{
		// We need to use dynamic friction.
		impulseContact.y /= planarImpulse;
		impulseContact.z /= planarImpulse;
		impulseContact.x = velocityPerImpulse.data[0] +
			velocityPerImpulse.data[1] * friction * impulseContact.y +
			velocityPerImpulse.data[2] * friction * impulseContact.z;
		impulseContact.x = desiredDeltaVelocity / impulseContact.x;
		impulseContact.y *= friction * impulseContact.x;
		impulseContact.z *= friction * impulseContact.x;
	}
}

template <typename d>
void StupidVelocityResolver<d>::resolve (Contact& data)
{
	bool body0awake = data.body0->isAwake();
	bool body1awake = data.body1->isAwake();
	
	if (!body0awake && !body1awake) { return; }
	
	float approaching = data.approachSpeed;
	
	// TODO: remove magic number
	if (approaching > 0.25) {
		//Check if either body needs waking up.
		
		//Wake the sleeping one if only one is asleep and the other is not static.
		if ((body0awake) ^ (body1awake)) {
			if (body0awake) data.body1->setAwake(true);
			else data.body0->setAwake(true);
		}
	}
	
	float separating = 0;
	
	if (Globals::useRestitution && data.approachSpeed > Globals::restitutionThreshold) {
		float restitution = 0.5f * (data.shape[0]->getRestitution() + data.shape[1]->getRestitution());
		
		separating = restitution * approaching;
	}
	
	Vector velocityChange = -data.relativeVelocity;
	velocityChange.x = separating + approaching;
	
	Vector impulse = data.impulsePerVelocity * velocityChange;
	
	float friction = 0.5f * (data.shape[0]->getFriction() + data.shape[1]->getFriction());
	applyFriction(impulse, data.velocityPerImpulse, friction, separating + approaching);
	
	impulse = data.contactToWorld * impulse;
	
	if (data.body0->isAwake()) data.body0->applyImpulse(-impulse, data.position, false);
	if (data.body1->isAwake()) data.body1->applyImpulse(impulse, data.position, false);
}

template class StupidVelocityResolver<TwoDee>;
template class StupidVelocityResolver<ThreeDee>;

