#include "world.h"

#include "broadphase.h"
#include "broadphase_quadtree.h"
#include "broadphase_octree.h"
#include "contact_generator.h"
#include "contact_list_base.h"
#include "contact_resolver.h"
#include "body.h"
#include "forces.h"
#include "joint.h"

#include <math.h>
#include <boost/foreach.hpp>
#include <algorithm>

using namespace physics;

template <typename d>
World<d>::World (
	float timestep,
	BroadPhase<d>* broadphase,
	ContactGenerator<d>* contactGenerator,
	ContactList<d>* contactList,
	ContactResolver<d>* velocityResolver,
	ContactResolver<d>* positionResolver
)
: m_timestep(timestep),
	m_broadphase(broadphase),
	m_contactGenerator(contactGenerator),
	m_contactList(contactList),
	m_velocityResolver(velocityResolver),
	m_positionResolver(positionResolver),
	m_gravity(0, Globals::GRAVITY)
{ init(); }

template <typename d>
void World<d>::init () {
	if (! m_timestep) { m_timestep = Globals::DEFAULT_TIMESTEP; }
	
	// Default algorithms
	// TODO: make customisable
	
	typedef typename d::HackBroadPhase DefaultBroadPhase;
	
	if (! m_broadphase) { m_broadphase = new DefaultBroadPhase(); }
	if (! m_contactGenerator) { m_contactGenerator = new ContactGenerator<d>(); }
	if (! m_contactList) { m_contactList = new NaiveContactList<d>(); }
	if (! m_velocityResolver) { m_velocityResolver = new CattoVelocityResolver<d>(); }
	if (! m_positionResolver) { m_positionResolver = new CattoPositionResolver<d>(); }
	
	m_broadphase->init(this);
	m_contactGenerator->init(this);
	m_contactList->init(this);
	m_velocityResolver->init(this);
	m_positionResolver->init(this);
	
	frameRWABias = pow(Globals::sleepRWABias, m_timestep);
}

template <typename d>
World<d>::~World () {
	Profiler::profiler.end();
	
	delete m_broadphase;
	delete m_contactGenerator;
	delete m_contactList;
	delete m_velocityResolver;
	delete m_positionResolver;
	
	BOOST_FOREACH(Joint* j, joints) {
		delete j;
	}
	
	BOOST_FOREACH(ForceGenerator* f, forces) {
		delete f;
	}
	
	BOOST_FOREACH(Body* b, bodies) {
		delete b;
	}
}

template <typename d>
void World<d>::add (Body* body) {
	bodies.push_back(body);
	m_broadphase->bodyAdded(body);
}

template <typename d>
void World<d>::remove (Body* body) {
	this->remove(std::find(bodies.begin(), bodies.end(), body));
}

template <typename d>
typename World<d>::body_iterator World<d>::remove (body_iterator it) {
	if (it == bodies.end()) {
		return it;
	}
	
	Body* body = *it;
	
	m_broadphase->bodyRemoved(body);
	m_contactList->removeContacts(body);
	
	for (joint_iterator jit = joints.begin(); jit != joints.end(); /* Nothing */ ) {
		Joint* joint = *jit;
		
		if (joint->body1 == body || joint->body2 == body) {
			jit = joints.erase(jit);
			
			delete joint;
		} else {
			jit++;
		}
	}
	
	for (std::list<ForceGenerator*>::iterator fit = forces.begin(); fit != forces.end(); /* Nothing */ ) {
		AnchoredSpring<d>* spring1 = dynamic_cast<AnchoredSpring<d>*>(*fit);
		CleverAnchoredSpring<d>* spring2 = dynamic_cast<CleverAnchoredSpring<d>*>(*fit);
		
		if ((spring1 && spring1->getBody() == body) || (spring2 && spring2->getBody() == body)) {
			delete *fit;
			
			fit = forces.erase(fit);
		} else {
			fit++;
		}
	}
	
	return bodies.erase(it);
}

template <typename d>
void World<d>::add (Shape* shape) {
	add(Body::createStatic(shape));
}

template <typename d>
void World<d>::add (Joint* joint)
{
	joints.push_back(joint);
	
	joint->body1->joints.push_back(joint);
	joint->body2->joints.push_back(joint);
}

template <typename d>
void World<d>::remove (Joint* joint)
{
	joints.erase(std::remove(joints.begin(), joints.end(), joint), joints.end());
	
	joint->body1->joints.remove(joint);
	joint->body2->joints.remove(joint);
}

template <typename d>
void World<d>::add (ForceGenerator* g) {
	forces.push_back(g);
}

template <typename d>
void World<d>::remove (ForceGenerator* g) {
	forces.erase(std::remove(forces.begin(), forces.end(), g), forces.end());
}

template <typename d>
typename d::Body* World<d>::findBeneath (const Vector& p) const {
	return m_broadphase->findBeneath(p);
}

template <typename d>
std::list<typename d::Body*> World<d>::findBeneathAll (const AARect& r) const {
	return m_broadphase->findBeneathAll(r);
}

template <typename d>
typename d::Body* World<d>::intersectRay (const Vector& start, const Vector& dir, Vector& p) const {
	return m_broadphase->intersectRay(start, dir, p);
}

template <typename d>
typename d::Body* World<d>::intersectRay (const Vector& start, const Vector& dir) const {
	Vector p;
	
	return m_broadphase->intersectRay(start, dir, p);
}

template <typename d>
void World<d>::step () {
	Profiler::profiler.startStep();
	Profiler::profiler.setFrameVar("Body Count", bodies.size());
	
	if (Globals::whenToMove == Globals::MOVE_AFTER_DRAWING) {
		Profiler::profiler.startMethod("moveBodies");
		BOOST_FOREACH(Body* body, bodies) {
			body->move(m_timestep);
			body->moveBias(m_timestep);
		}
		Profiler::profiler.endMethod("moveBodies");
	}
	
	Profiler::profiler.startMethod("applyForces");
	applyForces();
	Profiler::profiler.endMethod("applyForces");
	
	if (Globals::whenToMove == Globals::MOVE_BEFORE_COLLISIONS) {
		Profiler::profiler.startMethod("moveBodies");
		BOOST_FOREACH(Body* body, bodies) {
			body->move(m_timestep);
			body->moveBias(m_timestep);
		}
		Profiler::profiler.endMethod("moveBodies");
	}
	
	Profiler::profiler.startMethod("findPairs");
	const body_pair_container& pairs = m_broadphase->findPairs();
	Profiler::profiler.setFrameVar("Pairs", pairs.size());
	Profiler::profiler.endMethod("findPairs");
	
	for (int i = 0; i < 1; i++) { //wth is this for o0? looks like a do...while{false} to me. 
		Profiler::profiler.startMethod("findContacts");
		
		const contact_container& contacts = m_contactGenerator->findContacts(pairs);
		Profiler::profiler.setFrameVar("New Contacts", contacts.size());
		Profiler::profiler.endMethod("findContacts");
		
		Profiler::profiler.startMethod("addContacts");
		m_contactList->addContacts(contacts);
		Profiler::profiler.endMethod("addContacts");
		
		Profiler::profiler.startMethod("notifyListeners");
		m_contactList->notifyListeners();
		Profiler::profiler.endMethod("notifyListeners");
		
		Profiler::profiler.setFrameVar("Total Contacts", m_contactList->size());
		Profiler::profiler.startMethod("resolveVelocities");
		m_velocityResolver->resolve(m_contactList, &joints);
		Profiler::profiler.endMethod("resolveVelocities");
		
		Profiler::profiler.startMethod("resolvePenetration");
		m_positionResolver->resolve(m_contactList, &joints);
		Profiler::profiler.endMethod("resolvePenetration");
	}
	
	if (Globals::whenToMove == Globals::MOVE_AFTER_COLLISIONS) {
		Profiler::profiler.startMethod("moveBodies");
		BOOST_FOREACH(Body* body, bodies) {
			body->move(m_timestep);
			body->moveBias(m_timestep);
		}
		Profiler::profiler.endMethod("moveBodies");
	}
	
	Profiler::profiler.endStep();
}

template <typename d>
void World<d>::applyForces () {
	BOOST_FOREACH(ForceGenerator* g, forces) {
		g->apply(m_timestep);
	}
	
	float frameRWABias = pow(Globals::sleepRWABias, m_timestep);
	
	BOOST_FOREACH(Body* body, bodies) {
		body->applyForces(m_timestep, m_gravity, frameRWABias);
	}
}

// Because we're implementing methods of a template class in a cpp file
// (rather than in the header), the compiler can have trouble linking the
// declaration and the implementation.
// To fix this, we use the following magic line of code.
// See: http://www.parashift.com/c++-faq-lite/templates.html#faq-35.15
// Note that this _must_ be at the bottom of the file

template class World<TwoDee>;
template class World<ThreeDee>;


