#ifndef _PHYSICS_WORLD_
#define _PHYSICS_WORLD_

#include "common.h"
#include "vector.h"

#include <list>
#include <vector>

#include <math.h>

#include "profiler.h"

namespace physics {

	class AARect;
	class ForceGenerator;
	
	template <typename d> class BroadPhase;
	template <typename d> class ContactGenerator;
	template <typename d> class ContactList;
	template <typename d> class ContactResolver;
	template <typename d> class Contact;
	
	template <typename d>
	/*! \class World
	*	This class defines the World within which the bodies are created and the physics simulation takes place.
	*/
	class World {
	
		public:
		
		typedef d Dimension;
		typedef typename d::Body Body;
		typedef typename d::Shape Shape;
		typedef typename d::Vector Vector;
		typedef typename d::Contact Contact;
		typedef typename d::Joint Joint;

		typedef std::vector<Body*> body_container;
		typedef typename body_container::iterator body_iterator;
		
		typedef std::vector<Joint*> joint_container;
		typedef typename joint_container::iterator joint_iterator;
		
		typedef std::vector< std::pair<Body*, Body*> > body_pair_container;
		
		typedef std::vector<Contact*> contact_container;
		
		/**
		 * You can pass in 0 to any of these parameters to use the default algorithm,
		 * or provide a specific implementation you want to use.
		 */
		World (
			float timestep = Globals::DEFAULT_TIMESTEP,
			BroadPhase<d>* broadphase = 0,
			ContactGenerator<d>* contactGenerator = 0,
			ContactList<d>* contactList = 0,
			ContactResolver<d>* velocityResolver = 0,
			ContactResolver<d>* positionResolver = 0
		);
		
		/**
		 * Deletes everything in this world. If there is something you don't want
		 * to be deleted, you should remove it from the world first.
		 */
		~World ();
		
		/**
		 * Add a body to the world.
		 * Note: this cannot be safely called from within a loop iterating over the bodies container.
		 */
		void add (Body* body);
		
		/**
		 * Remove this body from the world.
		 * Note: this cannot be safely called from within a loop iterating over the bodies container.
		 * You should use remove(World::body_iterator) instead.
		 */
		void remove (Body* body);
		
		/**
		 * Remove a body from the world given an iterator to that body.
		 * Returns an iterator to the next body in the container.
		 */
		body_iterator remove (body_iterator it);
		
		/**
		 * Add a static body to the world with this shape.
		 */
		void add (Shape* shape);
		
		/**
		 * Add a joint to the world.
		 */
		void add (Joint* joint);
		
		/**
		 * Remove a joint from the world.
		 */
		void remove (Joint* joint);
		
		/**
		 * Add a force generator to the world.
		 */
		void add (ForceGenerator* generator);
		
		/**
		 * Remove a force generator from the world.
		 */
		void remove (ForceGenerator* generator);
		
		/**
		 * Simulate one frame.
		 */
		void step ();
		
		float getTimeStep () const { return m_timestep; }
		void setTimeStep (float f) { m_timestep = f; frameRWABias = pow(Globals::sleepRWABias, m_timestep);}
		
		Vector getGravity () const { return m_gravity; }
		void setGravity (const Vector& g) { m_gravity = g; }
		
		/**
		 * Find all bodies located at this position (if any).
		 */
		std::list<Body*> findBeneathAll (const Vector& position);
		
		/**
		 * Find all bodies touching this rect.
		 */
		std::list<Body*> findBeneathAll (const AARect& rect) const;
		
		/**
		 * Find the first body located at this position (if any).
		 */
		Body* findBeneath (const Vector& position) const;
		
		/**
		 * Find the first body touching this rect
		 */
		Body* findBeneath (const AARect& rect);
		
		/**
		 * Intersect a ray with the world and return the first Body it hits
		 * 
		 * @param start Origin of the ray.
		 * @param dir Direction of the ray.
		 * @param p Set by reference to the intersection point found.
		 */
		Body* intersectRay (const Vector& start, const Vector& dir, Vector& p) const;
		
		/**
		 * Intersect a ray with the world and return the first Body it hits
		 * 
		 * @param start Origin of the ray.
		 * @param dir Direction of the ray.
		 */
		Body* intersectRay (const Vector& start, const Vector& dir) const;
		
		Profiler* getProfiler() { return &Profiler::profiler; }

		BroadPhase<d>* getBroadPhase () const { return m_broadphase; }
		ContactGenerator<d>* getContactGenerator () const { return m_contactGenerator; }
		ContactList<d>* getContactList () const { return m_contactList; }
		ContactResolver<d>* getVelocityResolver () const { return m_velocityResolver; }
		ContactResolver<d>* getPositionResolver () const { return m_positionResolver; }
		
		//TODO: sort out accessor methods for these containers
		body_container bodies;

		// temp container for joints
		joint_container joints;
		std::list<ForceGenerator*> forces;

		private:
		
		// Prevent World objects from being copied
		World (const World& w);
		World& operator = (const World& w);
		
		void init ();
		
		void applyForces ();
		
		float m_timestep;
		float frameRWABias;
		
		// Stages of the collision pipeline
		BroadPhase<d>* m_broadphase;
		ContactGenerator<d>* m_contactGenerator;
		ContactList<d>* m_contactList;
		ContactResolver<d>* m_velocityResolver;
		ContactResolver<d>* m_positionResolver;
		
		body_pair_container m_broadphasePairs;
		
		Vector m_gravity;
	};
};

#endif

