#include "camera3d.h"
#include "camera2d.h"
#include "input.h"

using namespace physics;

void Camera3D::copyState (const Camera2D* other) {
	centre = other->centre;
	
	float ratio = (float) Window::height() / (float) Window::width();
	
	distance = MAGIC_CONVERSION_CONSTANT * ratio / other->scale;
	
	targetDistance = distance;
	
	camera = centre;
	
	camera.z = distance;
}

physics::Vector3D Camera3D::getNear (int x, int y) const {
	GLdouble proj[16];
	GLdouble model[16];
	GLint view[4];

	glGetDoublev(GL_PROJECTION_MATRIX, proj);
	glGetDoublev(GL_MODELVIEW_MATRIX, model);
	glGetIntegerv(GL_VIEWPORT, view);
	
	// OpenGL has 0,0 at bottom-left not top-left
	y = view[3] - y;
	
	// Get mouse coordinates projected onto near clipping plane
	GLdouble nearx, neary, nearz;
	gluUnProject(x, y, 0, model, proj, view, &nearx, &neary, &nearz);
	
	return physics::Vector3D(nearx, neary, nearz);
}

physics::Vector3D Camera3D::getFar (int x, int y) const {
	GLdouble proj[16];
	GLdouble model[16];
	GLint view[4];

	glGetDoublev(GL_PROJECTION_MATRIX, proj);
	glGetDoublev(GL_MODELVIEW_MATRIX, model);
	glGetIntegerv(GL_VIEWPORT, view);
	
	// OpenGL has 0,0 at bottom-left not top-left
	y = view[3] - y;
	
	// Get mouse coordinates projected onto far clipping plane
	GLdouble farx, fary, farz;
	gluUnProject(x, y, 1, model, proj, view, &farx, &fary, &farz);
	
	return physics::Vector3D(farx, fary, farz);
}

physics::Vector3D Camera3D::screen2world (int x, int y) const {
	GLdouble proj[16];
	GLdouble model[16];
	GLint view[4];

	glGetDoublev(GL_PROJECTION_MATRIX, proj);
	glGetDoublev(GL_MODELVIEW_MATRIX, model);
	glGetIntegerv(GL_VIEWPORT, view);
	
	// OpenGL has 0,0 at bottom-left not top-left
	y = view[3] - y;
	
	float z = 1;
	
	glReadPixels( x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &z);
	
	// See if we can get world coordinates of whatever we're hovering over
	if (true3D || z < 1) {
		GLdouble objx, objy, objz;
		
		gluUnProject(x, y, z, model, proj, view, &objx, &objy, &objz);
		
		if (true3D || (objz >= -1 && objz <= 1)) {
			return Vector3D(objx, objy, objz);
		}
	}
	
	// Get mouse coordinates projected onto near clipping plane
	GLdouble nearx, neary, nearz;
	gluUnProject(x, y, 0, model, proj, view, &nearx, &neary, &nearz);
	
	// Get mouse coordinates projected onto far clipping plane
	GLdouble farx, fary, farz;
	gluUnProject(x, y, 1, model, proj, view, &farx, &fary, &farz);

	// Intersect the line between these two points with the z=0 plane
	float t = nearz / (nearz - farz);
	
	Vector3D v;
	
	v.x = nearx + t * (farx - nearx);
	v.y = neary + t * (fary - neary);
	
	return v;
}

void Camera3D::update () {
	int keyLR = 0;
	
	if (Input::isPressed(SDLK_LSHIFT) || Input::isPressed(SDLK_RSHIFT)) {
		keyLR = Input::isPressed(SDLK_RIGHT) - Input::isPressed(SDLK_LEFT);
	}
	
	//int keyUD = Input::isPressed(SDLK_UP) - Input::isPressed(SDLK_DOWN);
	
	/*centre += physics::Vector2D(keyLR / 5.0f, keyUD / 5.0f);
	camera += physics::Vector2D(keyLR / 5.0f, keyUD / 5.0f);*/
	
	Vector3D moveRelToCamera = -Vector3D(0, 1, 0).cross(Vector3D(centre) - camera) * keyLR / 30.0f;
	
	camera += moveRelToCamera;
	
	if (targetDistance > 200) { targetDistance = 200; }
	if (targetDistance < 4.0f) { targetDistance = 4.0f; }
	
	if (distance != targetDistance) {
		float distanceChange = (targetDistance - distance) / 10;
		distance += distanceChange;
		
		//physics::Vector2D offsetBefore = mouse - target;
		
		//centre = mouse - (centre - mouse) / (distance - distanceChange) * distance;
	}
	
	camera = Vector3D(centre) + (camera - Vector3D(centre)).normalise() * distance;
	
	if (true3D) return;
	
	const float offset = distance / 5;
	
	const float minY = 0 + offset;
	const float maxY = 100 - offset;
	const float minX = -100 + offset;
	const float maxX = 100 - offset;
	
	if (centre.y < minY) { camera.y = centre.y = minY; }
	if (centre.y > maxY) { camera.y = centre.y = maxY; }
	if (centre.x < minX) { camera.x = centre.x = minX; }
	if (centre.x > maxX) { camera.x = centre.x = maxX; }
	
	// Initialise frustum: for draw culling
	
	Vector3D forward = (centre - camera).normalise();
	Vector3D left = forward.cross(Vector3D(0, 1, 0));
	Vector3D upRel = forward.cross(left);
	
	float aspect = float(Window::width()) / float(Window::height());
	
	frustum.init(camera, upRel, forward, left, 45.0f, aspect);
}

void Camera3D::use () {
	int width = Window::width();
	int height = Window::height();
	
	// Make it scale when you resize window or go fullscreen
	height = (int) (height * 640.0f / width);
	width = 640;
	
	// Set up the projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 1, 400);
	
	gluLookAt(camera.x, camera.y, camera.z, centre.x, centre.y, 0, 0, 1, 0);
	
	/*glTranslatef(-centre.x, -centre.y, 0);
	
	glRotatef(angle, 0, 1, 0);
	
	glTranslatef(0, 0, -distance);*/
	
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
	
	// Lighting
	
	glShadeModel (GL_SMOOTH);
	
	glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE ) ;
	glEnable ( GL_COLOR_MATERIAL ) ;
	
	GLfloat global_ambient[] = {0.3f, 0.3f, 0.3f, 1.0f};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
	
	GLfloat ambient[] = {0.0f, 0.0f, 0.0f, 1.0f};
	GLfloat diffuse[] = {0.8f, 0.8f, 0.8f, 1.0f};
	GLfloat specular[] = {0.6f, 0.6f, 0.6f, 1.0f};
	GLfloat emission[] = {0.0f, 0.0f, 0.0f, 1.0f};
	
	GLfloat light_position[] = { 1.0, 1.0, 1.0, 0.0 };
	
	glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	glMaterialfv ( GL_FRONT_AND_BACK, GL_EMISSION, emission);

	glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 32);
	
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	
	glEnable(GL_NORMALIZE);
	
	// Make lines look better
	
	glEnable( GL_LINE_SMOOTH );
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glLineWidth (1.5);
}

