#include "director.h"
#include "random.h"
#include "physics.h"
#include <math.h>
#include <map>
#include <boost/foreach.hpp>

#define PI 3.14159265



using namespace physics;

Director::Director(World2D* newWorld) {
	world = newWorld;
	
	meteorShower = false;
	rain = false;
	fireworks = false;

	activeDirectorScene = 42;

	stepsElapsed = 0;
	getScript(0, &eventQueue);
}

Director::Director(World2D* newWorld, int script) {
	world = newWorld;
	
	meteorShower = false;
	rain = false;
	fireworks = false;
	
	activeDirectorScene = 0;

	stepsElapsed = 0;
	getScript(script, &eventQueue);
}

bool Director::step() {
	stepsElapsed++;
	
	// Add constant movement to preset 1
	if (false) {//if (getActiveDirectorScene() == 0 && (stepsElapsed % 750 == 0)) {
		
		// For each body add a force away from the origin
		BOOST_FOREACH(Body2D* b, world->bodies) {
			if (b->isStatic()) { continue; }
			if (b->isPinned()) { continue; }
			
			Vector2D origin(3, 10);
			Vector2D offset = b->getPosition() - origin;
			float strength = random_f(40,75);
			float distSq = offset.lengthSquared();
			if (distSq > strength * strength) { continue; }
			float dist = sqrt(distSq);
			if (dist < 1) {
				offset *= 1 / dist;
				dist = 1;
				distSq = 1;
			}
			
			Vector2D force = b->getMass() * (strength * offset / dist - offset);
			offset = Vector2D(random_f(-1, 1), random_f(-1, 1)) * 0.1f;
			b->applyImpulse(force, b->getPosition() + offset);
		}
	}
	
	if (meteorShower && (stepsElapsed % 20 == 0)) {
		Vector2D pos(random_f(-75, 75), random_f(20, 75));
		addMeteor(pos);
	}
	
	if (rain && (stepsElapsed % 10 == 0)) {
		Vector2D pos(stepsElapsed % 100 - 45, 90);
		world->add(new Body2D(new Circle(pos, 0.5)));
	}
	
	if (fireworks && (stepsElapsed % 100 == 0)) {
		Vector2D pos(random_f(-75, 75), random_f(20, 75));
		addExplosion(pos);
	}
	
	if (eventQueue.count(stepsElapsed) > 0) {
		
		for (std::map<long, Event>::iterator i = eventQueue.equal_range(stepsElapsed).first; i != eventQueue.equal_range(stepsElapsed).second; i++) {
			Body2D* body;
			Vector2D pentagon[5];
			switch (i->second.getType()) {
				case 0:
					return false;
					break;
				case 1:
					//Circle
					body = new Body2D(new Circle(i->second.getLocation(), i->second.getSize()));
					body->setVelocity(i->second.getVelocity());
					world->add(body);
					break;
				case 2:
					//Square
					body = new Body2D(new Rect(i->second.getLocation(), i->second.getSize(), i->second.getSize()));
					body->setVelocity(i->second.getVelocity());
					world->add(body);
					break;
				case 3:
					// Pentagon
					for (int n = 0; n < 5; n++) {
						pentagon[n] = i->second.getLocation();
						pentagon[n].x += 1.5f * cos(2 * 3.1415f * n / 5);
						pentagon[n].y += 1.5f * sin(2 * 3.1415f * n / 5);
					}
					body = Body2D::create(physics::Polygon::create(pentagon, 5));
					body->setVelocity(i->second.getVelocity());
					world->add(body);
					break;
				case 4:
					addExplosion(i->second.getLocation());
					break;
				case 5:
					toggleMeteorShower();
					break;
				case 6:
					toggleRain();
					break;
				case 7:
					toggleFireworks();
					break;
			}
		}
	}
	
	return true;
	
}
		
void Director::startMeteorShower() { meteorShower = true; }
void Director::stopMeteorShower() { meteorShower = false; }
void Director::toggleMeteorShower() {meteorShower = !meteorShower; }

void Director::addMeteor(Vector2D pos){
	Vector2D velocity(random_f(50, 75), -random_f(25, 50));
	if (random_f() < 0.5f) { velocity.x *= -1; }
	Body2D* body = Body2D::create(new Circle(pos - velocity / 5, 1.0f));
	body->setVelocity(velocity);
	world->add(body);
}

void Director::startRain() {rain = true;}
void Director::stopRain() {rain = false;}
void Director::toggleRain() {rain = !rain; }


void Director::startFireworks() { fireworks = true; }
void Director::stopFireworks() { fireworks = false; }
void Director::toggleFireworks() {fireworks = !fireworks; }

void Director::addExplosion(Vector2D pos){
	
	for (int i = 0; i < 36; i++){
		double radians = i*10*PI/180.0;
		float x = cos(radians);
		float y = sin(radians);
		
		Vector2D v = Vector2D(x, y);
		
		Body2D* body = Body2D::create(new Circle(pos + v, 0.1f));
		body->setVelocity(v * 10);
		world->add(body);
	}
}

void Director::getScript(int script, std::map<long, Event> * result){
	switch (script) {
		case 0: 
			break;
		case 1:
			(*result)[1] = Event(6); 
			(*result)[5000] = Event(4, physics::Vector2D(0, 10));
			(*result)[5001] = Event(6);
			(*result)[10000] = Event(0);
			break;
		case 2:
			(*result)[1] = Event(5);
			(*result)[30000] = Event(0);
			break;
	}
}

void Director::setActiveDirectorScene(int p){
	activeDirectorScene = p;
}

int Director::getActiveDirectorScene() {
	return activeDirectorScene;
}
