#include "drawing.h"
#include "window.h"

#include "box.h"

#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>

GLUquadric* quadric = 0;
GLuint sphereDisplayList;
GLuint circleDisplayList;
GLuint emptyCircleDisplayList;
GLuint cylinderDisplayList;
GLuint cubeDisplayList;

using namespace physics;

void Drawing::init () {
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	
	#ifndef WIN32
	// In Linux, we don't need to recalculate everything.
	if (quadric) {
		return;
	}
	#endif
	
	quadric = gluNewQuadric();
	
	// Sphere
	sphereDisplayList = glGenLists(1);
	glNewList(sphereDisplayList, GL_COMPILE);
	gluSphere(quadric, 1, 32, 32);
	glEndList();
	
	// Circle
	circleDisplayList = glGenLists(1);
	glNewList(circleDisplayList, GL_COMPILE);
	gluDisk(quadric, 0, 1, 32, 1); // inner radius, outer radius, slices, loops
	glEndList();
	
	// Cylinder
	cylinderDisplayList = glGenLists(1);
	glNewList(cylinderDisplayList, GL_COMPILE);
		glTranslatef(0, 0, 0.5f);
		gluDisk(quadric, 0, 1, 32, 1); // inner radius, outer radius, slices, loops
		glTranslatef(0, 0, -1);
		gluCylinder(quadric, 1, 1, 1, 32, 2); // base radius, top radius, height, slices, stacks
		glRotatef(180, 1, 0, 0);
		gluDisk(quadric, 0, 1, 32, 1); // inner radius, outer radius, slices, loops
	glEndList();
	
	gluQuadricDrawStyle(quadric, GLU_SILHOUETTE);
	
	// Empty circle
	emptyCircleDisplayList = glGenLists(1);
	glNewList(emptyCircleDisplayList, GL_COMPILE);
	gluDisk(quadric, 0, 1, 64, 1); // inner radius, outer radius, slices, loops
	glEndList();
	
	gluQuadricDrawStyle(quadric, GLU_FILL);
	
	// Cube
	cubeDisplayList = glGenLists(1);
	glNewList(cubeDisplayList, GL_COMPILE);
		glBegin(GL_QUADS);
			// front, xy plane
			glNormal3d(0.0, 0.0, 1.0);
			glVertex3f(-0.5, -0.5, 0.5);
			glVertex3f( 0.5, -0.5, 0.5);
			glVertex3f( 0.5,  0.5, 0.5);
			glVertex3f(-0.5,  0.5, 0.5);
			
			// right, yz plane
			glNormal3d(1.0, 0.0, 0.0);
			glVertex3f( 0.5, -0.5,  0.5);
			glVertex3f( 0.5, -0.5, -0.5);
			glVertex3f( 0.5,  0.5, -0.5);
			glVertex3f( 0.5,  0.5,  0.5);
			
			// back, xy plane
			glNormal3d(0.0, 0.0, -1.0);
			glVertex3f( 0.5, -0.5, -0.5);
			glVertex3f(-0.5, -0.5, -0.5);
			glVertex3f(-0.5,  0.5, -0.5);
			glVertex3f( 0.5,  0.5, -0.5);
			
			// left, yz plane
			glNormal3d(-1.0, 0.0, 0.0);
			glVertex3f(-0.5, 0.5,  0.5);
			glVertex3f(-0.5, 0.5, -0.5);
			glVertex3f(-0.5,  -0.5, -0.5);
			glVertex3f(-0.5,  -0.5,  0.5);
			
			// top xz plane
			glNormal3d(0.0, 1.0, 0.0);
			glVertex3f( 0.5, 0.5,  0.5);
			glVertex3f( 0.5, 0.5, -0.5);
			glVertex3f(-0.5, 0.5, -0.5);
			glVertex3f(-0.5, 0.5,  0.5);
			
			// bottom, xz plane
			glNormal3d(0.0, -1.0, 0.0);
			glVertex3f(-0.5, -0.5, -0.5);
			glVertex3f(0.5, -0.5,  -0.5);
			glVertex3f( 0.5, -0.5,  0.5);
			glVertex3f( -0.5, -0.5, 0.5);
		glEnd();
	glEndList();
}

void Drawing::setOutlineMode (bool outlineMode) {
	if (outlineMode) {
		glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
		glDisable(GL_CULL_FACE);
		gluQuadricDrawStyle(quadric, GLU_SILHOUETTE);
	} else {
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_CULL_FACE);
		gluQuadricDrawStyle(quadric, GLU_FILL);
	}
}

void Drawing::drawAABB (const Body2D* body) {
	const AARect& aabb = body->getAABB();
	
	glColor4f(0, 0, 0, 0.5f);
	glBegin(GL_LINE_LOOP);
		for (int i = 0; i < 4; i++) {
			const Vector2D& p = aabb.getVertex(i);
			glVertex3f(p.x, p.y, 0);
		}
	glEnd();
}

void Drawing::drawBoundingRadius (const Body2D* body) {
	float radius = body->getBoundingRadius();
	Vector2D p = body->getPosition();
	
	glPushMatrix();
	
	glTranslatef(p.x, p.y, 0);
	
	glColor4f(0, 0, 0, 0.5f);
	//gluQuadricDrawStyle(quadric, GLU_SILHOUETTE);
	//gluDisk(quadric, 0, radius, 64, 1);
	glScalef(radius, radius, 1);
	glCallList(emptyCircleDisplayList);
	
	glPopMatrix();
}

void Drawing::drawBoundingRadius (const Body3D* body) {
	float radius = body->getBoundingRadius();
	Vector3D p = body->getPosition();
	
	glPushMatrix();
	
	glTranslatef(p.x, p.y, p.z);
	
	glRotatef(90, 1, 0, 0);
	
	glColor4f(0, 0, 0, 0.5f);
	
	gluQuadricDrawStyle(quadric, GLU_SILHOUETTE);
	gluSphere(quadric, radius, 16, 16);
	
	glPopMatrix();
}

void Drawing::drawContact (Contact<TwoDee>* c) {
	float x = c->position.x;
	float y = c->position.y;
	float dx = c->normal.x;
	float dy = c->normal.y;
	
	glVertex3f(x + dx * 0.1f, y + dy * 0.1f, 0);
	glVertex3f(x - dx * 0.1f, y - dy * 0.1f, 0);
	glVertex3f(x - dy * 0.05f, y + dx * 0.05f, 0);
	glVertex3f(x + dy * 0.05f, y - dx * 0.05f, 0);
}

void Drawing::drawContact (Contact<ThreeDee>* c) {
	float x = c->position.x;
	float y = c->position.y;
	float z = c->position.z;
	float dx = c->normal.x;
	float dy = c->normal.y;
	float dz = c->normal.z;
	
	glVertex3f(x + dx * 0.1f, y + dy * 0.1f, z + dz * 0.1f);
	glVertex3f(x - dx * 0.1f, y - dy * 0.1f, z - dz * 0.1f);
}

void Drawing::drawShape (const Area* area, const Vector3D& colour, bool in3D, bool outlineMode) {

	switch (area->type()) {
		case Area::CIRCLE:
			drawCircle(static_cast<const Circle*>(area), colour, in3D, outlineMode);
		break;
		case Area::RECT:
		case Area::POLYGON:
			drawPolygon(area, colour, in3D, outlineMode);
		break;
		case Area::LINE:
			drawLine(static_cast<const Line*>(area), colour, in3D);
		break;
		case Area::CAPSULE:
			drawCapsule(static_cast<const Capsule2D*>(area), colour, in3D, outlineMode);
		break;
		default: {}
	}
}

void Drawing::drawShape (const Volume* volume, const Vector3D& colour, bool in3D, bool outlineMode) {

	switch (volume->type()) {
		case Volume::SPHERE:
			drawSphere(static_cast<const Sphere*>(volume), colour, outlineMode);
		break;
		case Volume::BOX:
			drawBox(static_cast<const Box*>(volume), colour, outlineMode);
		break;
		default: {}
	}
}

void Drawing::drawPolygon (const Area* shape, const Vector3D& colour, bool in3D, bool outlineMode) {
	if (! in3D) {
		drawPolygon2D(shape, colour, 0);
		return;
	}
	
	float z = sqrt(shape->getArea()) / 2.0f;
	
	drawPolygon2D(shape, colour, z);
	drawPolygon2D(shape, colour, -z);
	
	glColor3f(colour.x, colour.y, colour.z);
	glBegin(GL_QUADS);
		Vector2D p = shape->getVertex(0);
		
		Vector2D n = shape->getNormal(0);
		glNormal3f(n.x, n.y, 0);
		
		glVertex3f(p.x, p.y, z);
		glVertex3f(p.x, p.y, -z);
		
		for (int i = 1; i < shape->getVertexCount(); i++) {
			p = shape->getVertex(i);
			glVertex3f(p.x, p.y, -z);
			glVertex3f(p.x, p.y, z);
			
			n = shape->getNormal(i);
			glNormal3f(n.x, n.y, 0);
			
			glVertex3f(p.x, p.y, z);
			glVertex3f(p.x, p.y, -z);
		}
		
		p = shape->getVertex(0);
		glVertex3f(p.x, p.y, -z);
		glVertex3f(p.x, p.y, z);
	glEnd();
	
	if (z > 5 && outlineMode) {
		glColor3f(0, 0, 0);
		
		for (float z2 = -z; z2 <= z; z2 += 4.0f) {
			drawPolygon2D(shape, colour, z2);
		}
		
		glBegin(GL_LINES);
			for (int i = 0; i < shape->getVertexCount(); i++) {
				Vector2D p1 = shape->getVertex(i);
				Vector2D p2 = shape->getVertex(i+1);
				
				const float increment = 4.0f / p1.distance(p2);
				
				for (float t = 0; t <= 1; t += increment) {
					Vector2D p = Vector2D::lerp(p1, p2, t);
					glVertex3f(p.x, p.y, z);
					glVertex3f(p.x, p.y, -z);
				}
			}
		glEnd();
	}
}

void Drawing::drawPolygon2D (const Area* shape, const Vector3D& colour, float z) {
	if (z >= 0) {
		glNormal3f(0, 0, 1);
		glFrontFace(GL_CCW);
	} else {
		glNormal3f(0, 0, -1);
		glFrontFace(GL_CW);
	}
	
	glColor3f(colour.x, colour.y, colour.z);
	glBegin(GL_POLYGON);
		for (int i = 0; i < shape->getVertexCount(); i++) {
			const Vector2D& p = shape->getVertex(i);
			glVertex3f(p.x, p.y, z);
		}
	glEnd();
	
	glFrontFace(GL_CCW);
}

void Drawing::drawCircle (const Circle* c, const Vector3D& colour, bool in3D, bool outlineMode) {
	glColor3f(colour.x, colour.y, colour.z);
	
	Vector2D p = c->getPosition();
	float radius = c->getRadius();
	
	glPushMatrix();
	glTranslatef(p.x, p.y, 0);
	glScalef(radius, radius, 1);
	
	float rot = c->getOrientation().getAngle();
	glRotatef(rot * 180 / Globals::PI, 0, 0, 1);
	
	bool drawLines = (c->getBody() && (! c->getBody()->isStatic()) && c->getBody()->getShapes().size() <= 1);
	
	if (outlineMode) {
		gluQuadricDrawStyle(quadric, GLU_SILHOUETTE);
	} else {	
		gluQuadricDrawStyle(quadric, GLU_FILL);
	}
	
	if (in3D) {
		float z = (radius) / 2;
		
		if (z < 0.2f) {
			z = 0.2f;
		}
		
		if (outlineMode) {
			z += 0.01f;
			
			glTranslatef(0, 0, -z / 2);
			drawCircle(false, drawLines);
			glTranslatef(0, 0, z);
			drawCircle(false, drawLines);
		} else {
			glScalef(1, 1, z);
			glCallList(cylinderDisplayList);
		}
	} else {
		drawCircle(!outlineMode, drawLines);
	}
	
	glPopMatrix();
}

void Drawing::drawCapsule (const Capsule2D* c, const Vector3D& colour, bool in3D, bool outlineMode) {
	glColor3f(colour.x, colour.y, colour.z);
	
	Vector2D p = c->getPosition();
	float radius = c->getRadius();
	float length = c->getLength();
	
	glPushMatrix();
	glTranslatef(p.x, p.y, 0);
	
	Vector2D normal = c->getNormal();
	
	float rot = atan2(normal.y, normal.x);
	glRotatef(rot / Globals::PI * 180, 0, 0, 1);
	
	float z = in3D ? 2 * radius : 0;
	
	if (outlineMode) {
		z += 0.01f;
		
		glBegin(GL_LINES);
			glVertex3f(-radius, -length/2, -z/2);
			glVertex3f(-radius, length/2, -z/2);
			glVertex3f(radius, -length/2, -z/2);
			glVertex3f(radius, length/2, -z/2);
			glVertex3f(-radius, -length/2, z/2);
			glVertex3f(-radius, length/2, z/2);
			glVertex3f(radius, -length/2, z/2);
			glVertex3f(radius, length/2, z/2);
		glEnd();
		
		gluQuadricDrawStyle(quadric, GLU_SILHOUETTE);
		
		glTranslatef(0, length/2, z/2);
		gluPartialDisk(quadric, radius, radius, 64, 1, -90, 180);
		
		glTranslatef(0, 0, -z);
		gluPartialDisk(quadric, radius, radius, 64, 1, -90, 180);
		
		glTranslatef(0, -length, 0);
		gluPartialDisk(quadric, radius, radius, 64, 1, 90, 180);
		
		glTranslatef(0, 0, z);
		gluPartialDisk(quadric, radius, radius, 64, 1, 90, 180);
	} else {
		glPushMatrix();
		glScalef(2 * radius, length, z);
		glCallList(cubeDisplayList);
		glPopMatrix();
		
		glPushMatrix();
		glTranslatef(0, length / 2, 0);
		glScalef(radius, radius, z);
		glCallList(cylinderDisplayList);
		glPopMatrix();
		
		//glPushMatrix();
		glTranslatef(0, -length / 2, 0);
		glScalef(radius, radius, z);
		glCallList(cylinderDisplayList);
		//glPopMatrix();
	}
	
	glPopMatrix();
}

void Drawing::drawCircle(physics::Vector3D p, float radius, bool fill) {
	glPushMatrix();
	glTranslatef(p.x, p.y, p.z);
	
	glScalef(radius, radius, 1);
	
	drawCircle(fill);
	
	glPopMatrix();
}

void Drawing::drawCircle (bool fill, bool drawLines) {
	if (fill) {
		glCallList(circleDisplayList);
	} else {
		glCallList(emptyCircleDisplayList);
	}
	
	if (drawLines) {
		glBegin(GL_LINES);
			glVertex3f(1, 0, 0);
			glVertex3f(-1, 0, 0);
			glVertex3f(0, 1, 0);
			glVertex3f(0, -1, 0);
		glEnd();
	}
}

void Drawing::drawLine (const Line* line, const Vector3D& colour, bool in3D) {
	if (! in3D) {
		glColor3f(0, 0, 0);
		glBegin(GL_LINES);
			glVertex3f(line->x1(), line->y1(), 0);
			glVertex3f(line->x2(), line->y2(), 0);
		glEnd();
	} else {
		const float z = 0.4f * sqrt(line->getLength());
		
		Vector2D n = line->getNormal();
		
		glColor3f(colour.x, colour.y, colour.z);
		glBegin(GL_QUADS);
			glNormal3f(n.x, n.y, 0);
			glVertex3f(line->x1(), line->y1(), z);
			glVertex3f(line->x2(), line->y2(), z);
			glVertex3f(line->x2(), line->y2(), -z);
			glVertex3f(line->x1(), line->y1(), -z);
			
			glNormal3f(-n.x, -n.y, 0);
			glVertex3f(line->x1(), line->y1(), -z);
			glVertex3f(line->x2(), line->y2(), -z);
			glVertex3f(line->x2(), line->y2(), z);
			glVertex3f(line->x1(), line->y1(), z);
		glEnd();
	}
}

void Drawing::glRotate (const Orientation3D& orientation) {
	Matrix3 rotMatrix = orientation.getMatrix();
	
	float m[16];
	
	m[0] = rotMatrix.data[0];
	m[1] = rotMatrix.data[1];
	m[2] = rotMatrix.data[2];
	m[3] = 0;
	m[4] = rotMatrix.data[3];
	m[5] = rotMatrix.data[4];
	m[6] = rotMatrix.data[5];
	m[7] = 0;
	m[8] = rotMatrix.data[6];
	m[9] = rotMatrix.data[7];
	m[10] = rotMatrix.data[8];
	m[11] = 0;
	m[12] = 0;
	m[13] = 0;
	m[14] = 0;
	m[15] = 1;
	
	glMultMatrixf(m);
}

void Drawing::drawBox(const Box* box, Vector3D colour, bool outlineMode)
{
	Vector3D p = box->getPosition();

	Vector3D size = box->getSize();

	glColor3f(colour.x, colour.y, colour.z);

	glPushMatrix();

	glTranslatef(p.x, p.y, p.z);
	glRotate(box->getOrientation());
	glScalef(size.x, size.y, size.z);
	
	glCallList(cubeDisplayList);

	glPopMatrix();
}

void Drawing::drawSphere (const Sphere* sphere, const Vector3D& colour, bool outlineMode) {
	
	Vector3D p = sphere->getPosition();
	float radius = sphere->getRadius();
	
	glPushMatrix();
	
	glTranslatef(p.x, p.y, p.z);
	glRotate(sphere->getOrientation());
	
	glColor3f(colour.x, colour.y, colour.z);
	
	if (! outlineMode) {
		gluQuadricDrawStyle(quadric, GLU_FILL);
		//gluSphere(quadric, radius, 16, 16);
		glScalef(radius, radius, radius);
		glCallList(sphereDisplayList);
	} else if (! sphere->getBody()->isStatic()) {
		//gluQuadricDrawStyle(quadric, GLU_SILHOUETTE);
		
		radius += 0.001f;
		
		glScalef(radius, radius, radius);
		glCallList(emptyCircleDisplayList);
		
		//gluDisk(quadric, 0, radius, 64, 1);
		
		glRotatef(90, 1, 0, 0);
		
		//gluDisk(quadric, 0, radius, 64, 1);
		glCallList(emptyCircleDisplayList);
		
		glRotatef(90, 0, 1, 0);
		
		//gluDisk(quadric, 0, radius, 64, 1);
		glCallList(emptyCircleDisplayList);
	}
	
	glPopMatrix();
}

void Drawing::drawStatistics (int verbosity) {
	
	int height = Window::height();

	glTranslatef(15, height - 15, 0);	// draw at 15 pixels from left; 15 pixels from top
	glColor3f(0.0,0.0,0.0);
	
	// quick hack up to get some text in the sandbox - be nice to wrap these stream calls up or use boost (?)

	std::ostringstream buffer;
	buffer.setf(std::ios::fixed, std::ios::floatfield); // Display floating point numbers sensibly
	buffer.precision(2); // 2 digits after decimal point
	
	int frameTime = physics::Profiler::profiler.getLastFrameTime();
	if (frameTime == 0) frameTime = 1; //Less of that divide-by-zero business.

	if (verbosity > 0) {
		if (physics::Profiler::profiler.isEnabled()) {
			buffer << "Physics Time: " << std::setw(5) << frameTime / 1000.0 << " ms";
			std::string physicsTime = buffer.str();	
			Window::drawText(physicsTime.c_str());
			glTranslatef(0, -15, 0);	// define a lineheight or something
			
			int drawingTime = physics::Profiler::profiler.getLastFrameVar("Drawing");
			buffer.str("");
			buffer << "Drawing Time: " << std::setw(5) << drawingTime / 1000.0 << " ms";
			std::string drawingTimeStr = buffer.str();	
			Window::drawText(drawingTimeStr.c_str());
			glTranslatef(0, -15, 0);	// define a lineheight or something
		}
		
		int bodyCount = physics::Profiler::profiler.getLastFrameVar("Body Count");
		buffer.str("");		//clear the buffer
		buffer << "Bodies: " << bodyCount;
		std::string bodies = buffer.str();
		Window::drawText(bodies.c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something
		
		int contactCount = physics::Profiler::profiler.getLastFrameVar("Total Contacts");
		buffer.str("");		//clear the buffer
		buffer << "Contacts: " << contactCount;
		std::string contacts = buffer.str();
		Window::drawText(contacts.c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something
	}
	if (verbosity > 1) {
		int moveBodies = physics::Profiler::profiler.getLastFrameVar("moveBodies") ;
		double moveBodiespc = (moveBodies  * 100.0) / frameTime;
		buffer.str("");		//clear the buffer
		buffer << "moveBodies: " << (moveBodies / 1000.0) << "ms ( " << moveBodiespc << " % )";
		Window::drawText(buffer.str().c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something
		
		int applyForces = physics::Profiler::profiler.getLastFrameVar("applyForces");
		double applyForcespc = (applyForces  * 100.0) / frameTime;
		buffer.str("");		//clear the buffer
		buffer << "applyForces: " << (applyForces / 1000.0) << "ms ( " << applyForcespc << " % )";
		Window::drawText(buffer.str().c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something

		int findPairs = physics::Profiler::profiler.getLastFrameVar("findPairs");
		double findPairspc = (findPairs  * 100.0) / frameTime;
		buffer.str("");		//clear the buffer
		buffer << "findPairs: " << (findPairs / 1000.0) << "ms ( " << findPairspc << " % )";
		Window::drawText(buffer.str().c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something
		
		int findContacts = physics::Profiler::profiler.getLastFrameVar("findContacts");
		double findContactspc = (findContacts  * 100.0) / frameTime;
		buffer.str("");		//clear the buffer
		buffer << "findContacts: " << (findContacts / 1000.0) << "ms ( " << findContactspc << " % )";
		Window::drawText(buffer.str().c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something

		int addContacts = physics::Profiler::profiler.getLastFrameVar("addContacts") ;
		double addContactspc = (addContacts  * 100.0) / frameTime;
		buffer.str("");		//clear the buffer
		buffer << "addContacts: " << (addContacts / 1000.0) << "ms ( " << addContactspc << " % )";
		Window::drawText(buffer.str().c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something
		
		int notifyListeners = physics::Profiler::profiler.getLastFrameVar("notifyListeners");
		double notifyListenerspc = (notifyListeners  * 100.0) / frameTime;
		buffer.str("");		//clear the buffer
		buffer << "notifyListeners: " << (notifyListeners / 1000.0) << "ms ( " << notifyListenerspc << " % )";
		Window::drawText(buffer.str().c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something
		
		int resolveVelocities = physics::Profiler::profiler.getLastFrameVar("resolveVelocities");
		double resolveVelocitiespc = (resolveVelocities  * 100.0) / frameTime;
		buffer.str("");		//clear the buffer
		buffer << "resolveVelocities: " << (resolveVelocities / 1000.0) << "ms ( " << resolveVelocitiespc << " % )";
		Window::drawText(buffer.str().c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something

		int resolvePenetration = physics::Profiler::profiler.getLastFrameVar("resolvePenetration");
		double resolvePenetrationpc = (resolvePenetration  * 100.0) / frameTime;
		buffer.str("");		//clear the buffer
		buffer << "resolvePenetration: " << (resolvePenetration / 1000.0) << "ms ( " << resolvePenetrationpc << " % )";
		Window::drawText(buffer.str().c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something

		buffer.str("");		//clear the buffer
		buffer << "Verbosity: " << verbosity;
		Window::drawText(buffer.str().c_str());
		glTranslatef(0, -15, 0);	// define a lineheight or something
	}
}
