#ifndef _DRAWING_
#define _DRAWING_

#include "camera.h"
#include "physics.h"
#include "random.h"
#include "SDL_opengl.h"
#include <boost/foreach.hpp>
#include "display_object.h"

namespace Drawing {
	using namespace physics;
	
	void init ();
	
	void glRotate (const Orientation3D& o);
	
	void setOutlineMode (bool b);
	
	template <typename Container>
	void drawBodies (Container& bodies, Camera* camera);
	
	Vector3D getColour (Body2D* body);
	Vector3D getColour (Body3D* body);
	
	template <typename Body>
	void drawBody (const Body* body, const Vector3D& colour, bool in3D, bool outlineMode);
	
	template <typename Container>
	void drawJoints (Container& joints);
	
	template <typename ContactList>
	void drawContacts (ContactList* contacts);
	
	void drawContact (Contact<TwoDee>* contact);
	void drawContact (Contact<ThreeDee>* contact);
	
	void drawAABB (const Body2D* body);
	void drawBoundingRadius (const Body2D* body);
	
	void drawAABB (const Body3D* body);
	void drawBoundingRadius (const Body3D* body);
	
	void drawShape (const Area* a, const Vector3D& colour, bool in3D, bool outlineMode);
	void drawShape (const Volume* v, const Vector3D& colour, bool in3D, bool outlineMode);
	
	void drawPolygon (const Area* a, const Vector3D& colour, bool in3D, bool outlineMode);
	void drawCircle (const Circle* a, const Vector3D& colour, bool in3D, bool outlineMode);
	void drawCapsule (const Capsule2D* a, const Vector3D& colour, bool in3D, bool outlineMode);
	void drawLine (const Line* a, const Vector3D& colour, bool in3D);
	
	void drawPolygon2D (const Area* a, const Vector3D& colour, float z);
	
	void drawBox(const Box* a, Vector3D colour, bool outlineMode);
	void drawSphere (const Sphere* a, const Vector3D& colour, bool outlineMode);
	
	void drawCircle (physics::Vector3D p, float radius, bool fill = true);
	void drawCircle (bool fill = true, bool drawLines = false);
	
	void drawStatistics (int verbosity);
	
};

template <typename Container>
void Drawing::drawBodies (Container& bodies, Camera* camera) {
	typedef typename Container::value_type PBody; // Pointer to Body
	
	bool in3D = camera->is3D();
	
	setOutlineMode(false);
	
	BOOST_FOREACH(PBody body, bodies) {
		if (! camera->isVisible(body)) { continue; }
		
		Vector3D colour = getColour(body);
		
		drawBody(body, colour, in3D, false);
	}
	
	setOutlineMode(true);
	
	const Vector3D black(0, 0, 0);
	
	BOOST_FOREACH(PBody body, bodies) {
		if (! camera->isVisible(body)) { continue; }
		
		drawBody(body, black, in3D, true);
	}
	
	setOutlineMode(false);
	
	/*BOOST_FOREACH(PBody body, bodies) {
		drawAABB(body);
	}*/
	
	/*BOOST_FOREACH(PBody body, bodies) {
		drawBoundingRadius(body);
	}*/
}

template <typename Container>
void Drawing::drawJoints (Container& joints) {
	typedef typename Container::value_type PJoint; // Pointer to Joint

	glColor3f(0.5f, 0.5f, 0.5f);
	
	glBegin(GL_LINES);
		BOOST_FOREACH(PJoint j, joints)
		{
			Vector2D body1 = j->body1->getPosition();
			Vector2D body2 = j->body2->getPosition();
			
			Vector2D anchor1 = j->body1->localToWorld(j->localAnchor1);
			Vector2D anchor2 = j->body2->localToWorld(j->localAnchor2);
			
			//Drawing::drawCircle(anchor1, 0.1f);
			//Drawing::drawCircle(anchor2, 0.1f);

			if (j->body1->isStatic())
			{
				glVertex3f(body2.x, body2.y, 0);
				glVertex3f(anchor1.x, anchor1.y, 0);
			}
			else if (j->body2->isStatic())
			{
				glVertex3f(body1.x, body1.y, 0);
				glVertex3f(anchor2.x, anchor2.y, 0);
			}
			else
			{
				Vector2D anchor = 0.5f * (anchor1 + anchor2);
				
				glVertex3f(body1.x, body1.y, 0);
				glVertex3f(anchor.x, anchor.y, 0);
				
				glVertex3f(body2.x, body2.y, 0);
				glVertex3f(anchor.x, anchor.y, 0);
			}
		}
	glEnd();
}

inline physics::Vector3D Drawing::getColour (Body2D* body) {

	return DisplayObject::getDisplayObject(body)->getColour();
}

inline physics::Vector3D Drawing::getColour (Body3D* body) {
	static const Vector3D staticColour(0.65f, 0.8f, 0.5f);
	static const Vector3D sleepColour(0.5f, 0.5f, 0.5f);
	
	Vector3D colour;
	
	if (body->isStatic()) {
		colour = staticColour;
	} else if (!body->isAwake()) {
		colour = sleepColour;
	} else if (body->data) {
		colour = *static_cast<Vector3D*>(body->data);
	} else {
		colour = random_colour();
		body->data = new Vector3D(colour);
	}
	
	return colour;
}

template <typename Body>
void Drawing::drawBody (const Body* body, const physics::Vector3D& colour, bool in3D, bool outlineMode) {
	typedef typename Body::Shape Shape;
	
	BOOST_FOREACH(Shape* shape, body->getShapes()) {
		drawShape(shape, colour, in3D, outlineMode);
	}
}

template <typename ContactList>
void Drawing::drawContacts (ContactList* contacts) {
	glColor3f(1, 0, 0);
	
	glBegin(GL_LINES);
		contacts->foreach(drawContact);
	glEnd();
}

#endif

