#include "frustum.h"
#include "quaternion.h"
#include "body.h"

using namespace physics;

inline float degToRad (float x) {
	return x * physics::Globals::PI / 180;
}

void Frustum::init (Vector3D& position, Vector3D& up, Vector3D& forward, Vector3D& left, float fovY, float aspect) {

	float fov = fovY * aspect;
	
	//	Get camera perspective
	float a = degToRad(fovY / 2.0f);
	float b = degToRad(fov  / 2.0f);
	
	//	Get the plane normals
	Quaternion q = Quaternion::create(-a, left.x, left.y, left.z);
	plane[0].normal = q.transform(up) * -1;			//	TOP
	
	q = Quaternion::create(a, left.x, left.y, left.z);
	plane[1].normal = q.transform(up);			//	BOTTOM
	
	q = Quaternion::create(-b, up.x, up.y, up.z);
	plane[2].normal = q.transform(left);		//	RIGHT
	
	q = Quaternion::create(b, up.x, up.y, up.z);
	plane[3].normal = q.transform(left) * -1;		//	LEFT
	
	plane[4].normal = forward;				//	NEAR
	
	//plane[5].normal = forward * -1;				//	FAR
	
	//	Next calculate the plane offsets from the normals and the eye position.
	plane[0].d = -position.dot( plane[0].normal );
	plane[1].d = -position.dot( plane[1].normal );
	plane[2].d = -position.dot( plane[2].normal );
	plane[3].d = -position.dot( plane[3].normal );
	plane[4].d = -position.dot( plane[4].normal );
	
	/*
	//	for far, need the point on the far plane
	Vector3D farPt = position + (forward * farClip);
	plane[5].d = -farPt.dot( plane[5].normal );
	*/
}

//	Frustum Tests	//

bool Frustum::inside (const physics::Body2D* body) const {
	return inside(body->getPosition(), body->getBoundingRadius());
}

bool Frustum::inside (const physics::Body3D* body) const {
	return inside(body->getPosition(), body->getBoundingRadius());
}

bool Frustum::inside (const Vector3D& p, float r) const {
	//	Test each clip plane
	for (int i=0; i<5; i++) {
		if(! inside(plane[i], p, r)) {
			return false;
		}
	}
	
	return true;
}

//	Individual planes	//

//point (or circle) inside a plane
bool Frustum::inside (const Plane& p, const Vector3D& v, float r) const {
	float planeEqVal = p.d + p.normal.dot(v);
	return (planeEqVal >= -r);
}

