#include "sandbox2d.h"

#include "camera2d.h"
#include "camera3d.h"
#include "drawing.h"
#include "window.h"
#include "director.h"
#include "random.h"
#include "world_factory.h"
#include "world_io.h"
#include "display_object.h"

#include "physics.h"

#include <math.h>
#include <boost/foreach.hpp>

#include "contact_list_base.h"

using namespace physics;

Sandbox2D::Sandbox2D (int scene, int script, bool force2D, bool interactive)
	: world(0), director(0), drawContacts(1), dragObject(0), paused(false), panning(false), interfaceState(IDLE),
	activeScene(scene), activeScript(script), verbosity(1), interactiveMode(interactive)
{
	camera2D = new Camera2D(6, 6, 16); //(x,y,z) new camera centred at (x, y) with zoom of z
	camera3D = new Camera3D();
	if (force2D) camera = camera2D;
	else camera = camera3D;

	reset();
}

void Sandbox2D::keyDown (SDL_Event& event) {
	if (interfaceState == IDLE || interfaceState == CARRYING || interfaceState == LOADING) {
		//Mouse co-ords
		int screenX;
		int screenY;
		SDL_GetMouseState(&screenX, &screenY);
		mouse = camera->screen2world(screenX, screenY);

		//Key state (Used for modifier keys)
		Uint8 *keystate = SDL_GetKeyState(NULL);

		//Key events, alphabetical order.
		//Esc - Exit (Same as Q)
		//` (backtick /tilde) - Console
		//Space - Fire circle at mouse cursor
		//Ctrl+Space - Send existing bodies flying away from the cursor
		//Shift - Tilt camera UNIMPLEMENTED
		//Control - Hold down, mouse-click 3+ points, then release to create convex polygon with those verticies.
				//UNIMPLEMENTED -> Should support creation of conCAVE poly
				//Whilst dragging, control sets drag object's rotation to 0. 
				//Ctrl+C add selected bodies to clipboard
				//Ctrl+V paste bodies in clipboard
		//Del - Same as D
		//Pageup/Pagedown - Zoom in/out
		//Numbers 1-9, restart world with various test presets (also numberpad)
		//Tab - Toggle 3D perspective
		//F1  - Load world from file
		//F2  - Save world to file
		//F5  - Toggle meteor shower (continuous circle meteors)
		//F6  - Toggle Rain (falling stacks of circles)
		//F7  - Toggle Fireworks (randomly spawning circle rings)
		//Arrow keys - Move camera
		//B   - Add new rectangle at mouse co-ords
		//C   - Add new circle at mouse co-ords
		//D   - Delete first shape found at mouse co-ords, if no shapes selected 
			  //Delete selected shapes otherwise
		//F	  - Toggle full screen
		//I   - Cycle verbosity of info displayed
		//J   - Create joint. (Hold down. 3 clicks, body body point. If no bodies found nothing happens on release)
		//L   - Toggle stats logging
		//M   - Merge selected shapes
		//N   - Add random polygon at mouse co-ords
		//O   - Spawn fireworks (ring of circles) at mouse co-ords
		//P   - (Un)pause world
		//Q   - Quit (Same as Esc)
		//R   - Reset world to most recently loaded preset (default is initial scenario)
		//S   - Toggle the static-ness of object at mouse co-ords
		//V   - Add fixed circle at mouse co-ords
		//W   - Wake all bodies up
		//X   - Empty the world
		
		switch (event.key.keysym.sym) {
			case SDLK_ESCAPE:
			case SDLK_q:
				physics::Profiler::profiler.end(); //write profiler output on quit.
				exit(0);
			break;

			/*case SDLK_HASH:
			case SDLK_BACKQUOTE: {
				if (interfaceState == IDLE) {
					interfaceState = LOADING;
				}
				else {
					interfaceState = IDLE;
				}
			} break;*/

			case SDLK_TAB:
				toggleCamera();
			break;
			
			case SDLK_SPACE: {
				if (paused) {
					drawContacts = true;
					if (!(director->step())) {
						physics::Profiler::profiler.end(); //write profiler output on quit.
						exit(0);
					}
					world->step();
				}
				
				if (!paused && interactiveMode) {
					if (Input::isPressed(SDLK_RCTRL) || Input::isPressed(SDLK_LCTRL)) { // Throw things away
						
						/*Body2D* body = world->findBeneath(mouse);
						
						if (body && ! body->isStatic()) {
							BOOST_FOREACH(Shape2D* shape, body->getShapes()) {
								Body2D* newBody = Body2D::create(shape);
								
								newBody->setVelocity(body->getVelocityAtPoint(newBody->getPosition()));
								newBody->setRotation(body->getRotation());
								newBody->data = body->data;
								
								world->add(newBody);
							}
							
							world->remove(body);
						}*/
						
						BOOST_FOREACH(Body2D* b, world->bodies) {
							if (b->isStatic()) { continue; }
							if (b->isPinned()) { continue; }
							
							Vector2D offset = b->getPosition() - mouse;
							float strength = random_f(20,35);
							float distSq = offset.lengthSquared();
							if (distSq > strength * strength) { continue; }
							float dist = sqrt(distSq);
							if (dist < 1) {
								offset *= 1 / dist;
								dist = 1;
								distSq = 1;
							}
							
							Vector2D force = b->getMass() * (strength * offset / dist - offset);
							offset = Vector2D(random_f(-1, 1), random_f(-1, 1)) * 0.1f;
							b->applyImpulse(force, b->getPosition() + offset);
						}
					} else { // Shoot a projectile at cursor
						Vector2D velocity(random_f(50, 75), -random_f(15, 65));
						if (random_f() < 0.5f) { velocity.x *= -1; }
						Body2D* body = Body2D::create(new Circle(mouse - velocity / 5, 1.0f));
						
						body->setDensity(10);
						body->setVelocity(velocity);
						world->add(body);
					}
				}
			} break;
			
			case SDLK_PAGEUP: {
				camera->zoomIn(screenX, screenY);
			} break;
			
			case SDLK_PAGEDOWN: {
				camera->zoomOut(screenX, screenY);
			} break;
			
			case SDLK_LEFT: {
				if (interactiveMode && !Input::isPressed(SDLK_LSHIFT) && !Input::isPressed(SDLK_RSHIFT)) loadPreset(activeScene - 1);
			} break;
			
			case SDLK_RIGHT: {
				if (interactiveMode && !Input::isPressed(SDLK_LSHIFT) && !Input::isPressed(SDLK_RSHIFT)) loadPreset(activeScene + 1);
			} break;
			
			case SDLK_F1: { //Load, although F1 should be "Help" really
			//case SDLK_F7: { //Commonly used (half-life) save key. Used by rain already. 
				Sandbox2D::killWorld();
				world = WorldIO::load2d();
				director = new Director(world);
				setActiveDirectorScene(activeScene);
			} break;
			
			case SDLK_F2: {
			//case SDLK_F6: { //Commonly used (half-life) save key. Used by rain already. 
				WorldIO::save2d(world);
			} break;			
			
			case SDLK_F4: { //Is there some clever way to combine this with the Esc/Q code above?
				if (keystate[SDLK_LALT]) {
						physics::Profiler::profiler.end(); //write profiler output on quit.
						exit(0);
				}
			} break;
			case SDLK_F5: { 
				if (interactiveMode) director->toggleMeteorShower(); 
			} break;
			
			case SDLK_F6: { 
				if (interactiveMode) director->toggleRain(); 
			} break;

			case SDLK_F7: { 
				if (interactiveMode) director->toggleFireworks(); 
			} break;

			case SDLK_f: {
				Window::toggleFullscreen();	
			} break;
			
			case SDLK_r: {
				if (interactiveMode) reset();	
			} break;
			
			case SDLK_i: {
				verbosity = (verbosity + 1) % 3;	
			} break;
			
			case SDLK_j: {
				if (interactiveMode && interfaceState == IDLE) {
					vertexList.clear();
					interfaceState = JOINT;
				}
			} break;

			case SDLK_k: {
				if (interactiveMode && interfaceState == IDLE) {
					vertexList.clear();
					interfaceState = SPRING;
				}
			} break;

			case SDLK_l: {
				if (interactiveMode) {
					if (physics::Profiler::profiler.isEnabled()) physics::Profiler::profiler.disable();
					else physics::Profiler::profiler.enable();
				}
			} break;

			case SDLK_v: {
				if (Input::isPressed(SDLK_RCTRL) || Input::isPressed(SDLK_LCTRL)) {
					/* Paste shapes in clipboard */
					/* Handled on keyup */
				}
				else if (interactiveMode && interfaceState == IDLE) world->add(new Circle(mouse, random_f(0.5f, 1.0f)));
			} break;
			
			case SDLK_c: {
				if (Input::isPressed(SDLK_RCTRL) || Input::isPressed(SDLK_LCTRL)) {
					/* Shapes being copied */
					/* Handled on keyup */
				}
				else if (interactiveMode && interfaceState == IDLE) {
					vertexList.clear(); //Only need to store one point
					vertexList.push_back(mouse);
					interfaceState = CIRCLE;
				}
			} break;

			case SDLK_b: {
				if (interactiveMode && interfaceState == IDLE) {
					vertexList.clear(); //Only need to store one point
					vertexList.push_back(mouse);
					interfaceState = RECT;
				}
			} break;

			case SDLK_n: {
				if (interactiveMode) {
					physics::Polygon* polygon = 0;
						while (polygon == 0) {
							int vertex_count = random_i(3,6);
							Vector2D* vertexList = new Vector2D[vertex_count];
							
							for( int i=0; i<vertex_count ;i++) {
								vertexList[i].x = mouse.x+random_f(-2.0f,2.0f);
								vertexList[i].y = mouse.y+random_f(-2.0f,2.0f);
							}
							
							polygon = physics::Polygon::create(vertexList, vertex_count);		
							
							delete[] vertexList;
						
							// Move so centre of mass is at mouse position
							polygon->translate(mouse - polygon->getPosition());
						
							if (polygon->getArea() < 0.75f || polygon->getArea() > 9.0f) {
								// try again
								delete polygon;
								polygon = 0;
							}
						}
						world->add(new Body2D(polygon));
					}
			} break;

			case SDLK_m: { //Merge bodies
				if (interactiveMode) {
					Body2D* newBody = new Body2D;
					bool selected = false;
					for (World2D::body_iterator it = world->bodies.begin(); it!=world->bodies.end(); /*Nothing*/) {
						selected = false;
						Body2D* b = *it;
						if (b->data) {
							selected = DisplayObject::getDisplayObject(b)->isSelected();
						}
						if (selected) {
							BOOST_FOREACH(Shape2D* shape, b->getShapes()) {
								newBody->add(shape);
							}
							it = world->remove(it);
						} else {
							it++;
						}
					}
					world->add(newBody);
				}
			} break;

			case SDLK_u: { //unmerge: break a body into its component shapes
				if (interactiveMode) {
					bool selected = false;
					std::list<Shape2D*> shapeList;
					for (World2D::body_iterator it = world->bodies.begin(); it!=world->bodies.end(); /*Nothing*/) {
						selected = false;
						Body2D* b = *it;
						if (b->data) {
							selected = DisplayObject::getDisplayObject(b)->isSelected();
						}
						if (selected) {
							BOOST_FOREACH(Shape2D* shape, b->getShapes()) {
								//Cannot add the body here as it will update the iterator which
								//we are currently iterating over.
								//Not a problem normally, but sometimes will cause the vector to
								//resize, then iterator pointers are all invalid. 
								shapeList.push_front(shape);
							}
							it = world->remove(it);
						} else {
							it++;
						}
					}
					//It's now safe to add the shapes we stored earlier to the world
					BOOST_FOREACH(Shape2D* shape, shapeList) {
						world->add(new Body2D(shape));
					}
				}
			} break;
			
			case SDLK_DELETE: //Intentional overflow
			case SDLK_d: {
				if (interactiveMode) {
					bool someDeleted = false;
					bool selected = false;
					for (World2D::body_iterator it = world->bodies.begin(); it!=world->bodies.end(); /*Nothing*/) {
						selected = false;
						Body2D* b = *it;
						if (b->data) {
							selected = static_cast<DisplayObject*>(b->data)->isSelected();
						}
						if (selected) {
							someDeleted = true;
							it = world->remove(it);
						} else {
							it++;
						}
					}

					if (!someDeleted) { //if there were no bodies selected, delete what's under mouse
						// Delete body
						Body2D* body = world->findBeneath(mouse);
						if (body) {
							world->remove(body);
							interfaceState = IDLE;
							dragObject = 0;
						}						
					}
				}
			} break;

			case SDLK_s: { 	// Toggle static-ness
				if (interactiveMode) {
					bool someStatified = false;
					bool selected = false;
					BOOST_FOREACH(Body2D* body, world->bodies) {
						if (body->data) {
							selected = static_cast<DisplayObject*>(body->data)->isSelected();
							if (selected) {
								someStatified = true;
								body->setStatic(!(body->isStatic())); //Toggle
								body->setRotation(0);
								body->setVelocity(Vector2D());
							}
						}
					}

					//Nothing was selected, so toggle the body under the cursor
					if (!someStatified) { 
						Body2D* body = world->findBeneath(mouse);
						if (body) {
							body->setStatic(! body->isStatic());
							body->setRotation(0);
							body->setVelocity(Vector2D());
							interfaceState = IDLE;
							dragObject = 0;
						}
					}
				}
			} break;

			case SDLK_w: if (interactiveMode) {
				 BOOST_FOREACH(Body2D* b, world->bodies) {
			 		b->setAwake(true);
				 }
			} break; 

			case SDLK_x: {
				if (interactiveMode) clearWorld();			
			} break; 

			case SDLK_o: {
				if (interactiveMode) director->addExplosion(mouse); 
			} break;

			case SDLK_PAUSE: //Intentional overflow
			case SDLK_p: {
				togglePause(); 
			} break;

			case SDLK_KP1: 
			case SDLK_1: if (interactiveMode) {activeScene = 1; loadPreset(1);}	break;
			case SDLK_KP2: 
			case SDLK_2: if (interactiveMode) loadPreset(2); break;
			case SDLK_KP3: 
			case SDLK_3: if (interactiveMode) loadPreset(3); break;
			case SDLK_KP4: 
			case SDLK_4: if (interactiveMode) loadPreset(4); break;
			case SDLK_KP5: 
			case SDLK_5: if (interactiveMode) loadPreset(5); break;
			case SDLK_KP6: 
			case SDLK_6: if (interactiveMode) loadPreset(6); break;
			case SDLK_KP7: 
			case SDLK_7: if (interactiveMode) loadPreset(7); break;
			case SDLK_KP8: 
			case SDLK_8: if (interactiveMode) loadPreset(8); break;
			case SDLK_KP9: 
			case SDLK_9: if (interactiveMode) loadPreset(9); break;
			case SDLK_KP0: 
			case SDLK_0: if (interactiveMode) loadPreset(10); break;

			default: {}
		}
	}
}

//void DisplayKey(SDL_Event& event)
//{
//	SDL_KeyboardEvent key = event.key;
//	printf( "%s\n", SDL_GetKeyName(key));
//}

void Sandbox2D::keyUp (SDL_Event& event) {
	if (interactiveMode) {
		//Mouse co-ords
		int screenX;
		int screenY;
		SDL_GetMouseState(&screenX, &screenY);
		mouse = camera->screen2world(screenX, screenY);

		switch (event.key.keysym.sym) {
			case SDLK_RCTRL : //Intentional overflow
			case SDLK_LCTRL : {
				if (interfaceState == POLYGON) {
					//Add the custom polygon to world
					if (vertexList.size() > 2) {
						physics::Vector2D* vertexArray = new Vector2D[vertexList.size()];
						std::copy(vertexList.begin(), vertexList.end(), vertexArray);
						world->add(new Body2D(physics::Polygon::create(vertexArray, vertexList.size())));
						delete[] vertexArray;
						interfaceState = IDLE;
					}
					vertexList.clear();
					interfaceState = IDLE;
				}

				if (interfaceState == ORIENTATING) {
					if (SDL_GetMouseState(&screenX, &screenY)&SDL_BUTTON_LMASK) { //left mouse down
						//If the user is still holding down, restore drag object
						dragObject = orientateObject;
						orientateObject = 0; 
						interfaceState = CARRYING;
					} else {
						//Otherwise, let object fall
						orientateObject = 0; 
						interfaceState = IDLE;
					}
				}
			} break;

			case SDLK_v: {
				if (Input::isPressed(SDLK_RCTRL) || Input::isPressed(SDLK_LCTRL)) {
					BOOST_FOREACH(Body2D* body, clipboard) {
						//world->add(body); //I suspect this makes no sense; can't add an object that's already in the world?
						world->add(new Body2D(new Circle(Vector2D(0, 6),1))); //Testing
					}
				}
			}

			case SDLK_c: {
				if (Input::isPressed(SDLK_RCTRL) || Input::isPressed(SDLK_LCTRL)) {
					clipboard.clear();
					/*bool selected = false;
					BOOST_FOREACH(Body2D* body, world->bodies) {
						selected = false;

						if (body->data) {
							selected = DisplayObject::getDisplayObject(body)->isSelected();
							if (selected) {
								clipboard.push_back(body);
							}
						}
					}*/
				}
				else if (interfaceState == CIRCLE) {
					Vector2D centre = vertexList.front();
					float radius = centre.distance(mouse);
					
					//If mouse not moved, c must be pressed for random circle
					if (radius < 0.4f) {
						radius = random_f(0.5f, 1.0f);
					}
					
					world->add(new Body2D(new Circle(centre, radius)));

					vertexList.clear();
					interfaceState = IDLE;
				}
			} break;

			case SDLK_j: {
				if (interfaceState == JOINT) {

					//While array size >= 3
						//Are the first 3 points valid? (ie first 2 are bodies)
							//break
						//else remove invalid points and shift other points up
					//end

					//TEMP compromise for above:
					//Check vertexList has at least 3 points
					if (vertexList.size() < 3) {
						vertexList.clear();
						interfaceState = IDLE;
						return;
					}

					//Copy to vector
					physics::Vector2D* vertexArray = new Vector2D[vertexList.size()];
					std::copy(vertexList.begin(), vertexList.end(), vertexArray);

					//Find bodies
					Body2D* body1 = world->findBeneath(vertexArray[0]); //Will probably break if no bodies found
					Body2D* body2 = world->findBeneath(vertexArray[1]);

					if (body1 && body2) {
						Joint<TwoDee>* joint = new Joint<TwoDee>();
						joint->set(body1, body2, Vector2D(vertexArray[2].x, vertexArray[2].y));
						joint->ignoreContacts = false;
						world->add(joint);
					}

					vertexList.clear();
					interfaceState = IDLE;
				}

			} break;

			case SDLK_k: {
				//Check vertexList has at least 2 points
				if (vertexList.size() < 2) {
					vertexList.clear();
					interfaceState = IDLE;
					return;
				}

				//Copy to vector
				physics::Vector2D* vertexArray = new Vector2D[vertexList.size()];
				std::copy(vertexList.begin(), vertexList.end(), vertexArray);
				
				//Find body
				Body2D* body = world->findBeneath(vertexArray[0]); 

				if (body) {
					Vector2D SpringPos = Vector2D(vertexArray[1].x, vertexArray[1].y);
					Vector2D BodyPos = Vector2D(vertexArray[0].x, vertexArray[0].y); //in world space
					BodyPos = body->worldToLocal(BodyPos); //in local space
					//Magic number defaults TODO move to globals
					float springConstant = 600;//6;
					//float restLength = 3;
					float damping = 40;
					CleverAnchoredSpring2D* spring = new CleverAnchoredSpring2D(SpringPos, body, BodyPos, springConstant, damping);//restLength);
					world->add(spring);
				}
				vertexList.clear();
				interfaceState = IDLE;

			}break;

			case SDLK_b: {
				if (interfaceState == RECT) {
					Vector2D vertex1 = vertexList.front();
					Vector2D vertex2 = mouse; 
					
					float mouseLength = fabs(vertex1.x - vertex2.x);
					float mouseWidth = fabs(vertex1.y - vertex2.y);

					//If mouse not moved, b must be pressed for random rectangle
					if (mouseLength < 0.1 || mouseWidth < 0.1) {
						mouseLength = random_f(1.0f, 2.0f);
						mouseWidth = random_f(1.0f, 2.0f);
						Orientation2D ori = Orientation2D::random();
						world->add(new Body2D(new Rect(mouse, mouseLength, mouseWidth, ori)));
					}
					
					//Add custom rectangle
					else {
						world->add(new Body2D(new Rect(vertex1, vertex2)));
					}
					
					vertexList.clear();
					interfaceState = IDLE;
				}
			}
			default: {}
		}
	}
}


void Sandbox2D::mouseDown (SDL_Event& event) {
	//Mouse co-ords
	int screenX;
	int screenY;
	SDL_GetMouseState(&screenX, &screenY);
	mouse = camera->screen2world(screenX, screenY);

	switch (event.button.button) {
		case SDL_BUTTON_LEFT: {
			//Drawing polygon
			if (interactiveMode && (interfaceState == IDLE || interfaceState == POLYGON) && (Input::isPressed(SDLK_RCTRL) || Input::isPressed(SDLK_LCTRL))) { //zomg
					interfaceState = POLYGON; //Used to set the state after we make first point
					vertexList.push_back(mouse); // Add point to polygon					
					physics::Polygon::makeConvexHull(vertexList); //Make sure vertexList is still convex
					return;
			//Drawing joint
			} else if (interactiveMode && (interfaceState == IDLE || interfaceState == JOINT) && (Input::isPressed(SDLK_j))) {
					vertexList.push_back(mouse);
					return;
			//Drawing spring
			} else if (interactiveMode && (interfaceState == IDLE || interfaceState == SPRING) && (Input::isPressed(SDLK_k))) {
					vertexList.push_back(mouse);
					return;
			} else if (interfaceState==IDLE) {			// interactiveMode && ?		
				Body2D* body = world->findBeneath(mouse);
				//Dragging body
				if (body && !body->isStatic()) {
					if (interactiveMode) {
						interfaceState = CARRYING;
						dragObject = body;
						dragOffset = body->worldToLocal(mouse);
						return;
					}
				} else { //Selection rectangle    // interactiveMode && ?
					interfaceState = SELECTING;
					vertexList.clear(); 
					vertexList.push_back(mouse);
				}
			}
		} break; 
		case SDL_BUTTON_RIGHT:
			clickPos = camera->screen2world(event.button.x, event.button.y);
			panning = true;
		break;
		case SDL_BUTTON_WHEELUP:
			if (Input::isPressed(SDLK_LCTRL) || Input::isPressed(SDLK_RCTRL)) {
				if (paused) {
					if (dragObject) dragObject->setOrientation(dragObject->getOrientation()+0.2f);
				}
				else if ((interfaceState == CARRYING || interfaceState == ORIENTATING)) {
					if (dragObject) {
						orientateObject = dragObject;
						dragObject = 0;
					}
					interfaceState = ORIENTATING;
					orientateObject->setOrientation(orientateObject->getOrientation()+0.2f);
				}
			} else {
				camera->zoomIn(event.button.x, event.button.y);
			}
		break;
		case SDL_BUTTON_WHEELDOWN:
			if (Input::isPressed(SDLK_LCTRL) || Input::isPressed(SDLK_RCTRL)) {
				if (paused) {
					if (dragObject) dragObject->setOrientation(dragObject->getOrientation()-0.2f);
				}
				else if ((interfaceState == CARRYING || interfaceState == ORIENTATING)) {
					if (dragObject) {
						orientateObject = dragObject;
						dragObject = 0;
					}
					interfaceState = ORIENTATING;
					orientateObject->setOrientation(orientateObject->getOrientation()-0.2f);
				}
			} else {
				camera->zoomOut(event.button.x, event.button.y);
			}
		break;
	}
}

void Sandbox2D::mouseUp (SDL_Event& event) {
	//Mouse co-ords
	int screenX;
	int screenY;
	SDL_GetMouseState(&screenX, &screenY);
	mouse = camera->screen2world(screenX, screenY);

	switch (event.button.button) {
		case SDL_BUTTON_LEFT: {
			
			if (interfaceState == CARRYING) {
				dragObject = 0;
				interfaceState = IDLE;
			}
			
			if (interfaceState == SELECTING) {
				Vector2D corner = vertexList.front();
				physics::AARect rect = AARect(corner, mouse);
				
				BOOST_FOREACH(Body2D* b, world->bodies) {
					DisplayObject* dObj = DisplayObject::getDisplayObject(b);
					if (b->intersects(rect)) {
						dObj->setSelected(true);
					} else {
						dObj->setSelected(false);
					}
				}
				interfaceState = IDLE;
				vertexList.clear(); 
			}
			
		} break;
		
		case SDL_BUTTON_RIGHT: {
			//Can't say interfaceState = IDLE as left mouse could still be dragging
			panning = false; 
		} break;
	}
}

void Sandbox2D::update () {
	int screenX;
	int screenY;
	
	SDL_GetMouseState(&screenX, &screenY);
	mouse = camera->screen2world(screenX, screenY);
	
	if (panning) {
		camera->translate(clickPos - mouse);
	}

	if (orientateObject) {
		orientateObject->setVelocity(Vector2D());
	}

	if (dragObject) {
		Vector2D worldPos = dragObject->localToWorld(dragOffset);

		if (paused) {
			dragObject->translate(mouse - worldPos);
			dragObject->setVelocity(Vector2D());
			dragObject->setRotation(0);
			drawContacts = false;
		} else {
			Vector2D impulse = mouse - worldPos - dragObject->getVelocityAtPoint(worldPos) / 20;
			impulse *= dragObject->getMass();
			dragObject->applyImpulse(impulse, worldPos);
			
			if (Input::isPressed(SDLK_LCTRL) || Input::isPressed(SDLK_RCTRL)) {
				dragObject->setRotation(0);
			}
		}
	}
	
	camera->update();
	
	if (!paused) {
		drawContacts = true;
		if (!(director->step())) {
			physics::Profiler::profiler.end(); //write profiler output on quit.
			exit(0);
		}
		world->step();
	}
}

void Sandbox2D::draw () const {
	physics::Profiler::profiler.startMethod("Drawing");
	
	camera->use();
	
	// Set up the model view matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Start drawing
	Drawing::drawBodies(world->bodies, camera);
	
	Drawing::drawJoints(world->joints);

	// Draw force generators  (TODO: Generalise)
	BOOST_FOREACH(ForceGenerator* forcegenerator, world->forces) {
		// assuming here that all the force generators are Anchoredsprings - flesh out some force generator subclasses
		CleverAnchoredSpring<TwoDee>* mySpring = dynamic_cast<CleverAnchoredSpring<TwoDee>*>(forcegenerator);
		
		if (mySpring) {
			Vector2D bodyPos = mySpring->getBody()->localToWorld(mySpring->getBodyAnchor());
			Vector2D anchor = mySpring->getAnchor();
			glColor3f(1.0f, 0.0f, 0.0f);
			glBegin(GL_LINES);
				glVertex3f(bodyPos.x, bodyPos.y, 0);
				glVertex3f(anchor.x, anchor.y, 0);
			glEnd();
		}
	}

	//Draw line between mouse and dragged-object
	if (dragObject) {
		glColor3f(0.5f, 0.5f, 0.5f);
		Vector2D worldPos = dragObject->localToWorld(dragOffset);
		glBegin(GL_LINES);
			glVertex3f(mouse.x, mouse.y, 0);
			glVertex3f(worldPos.x, worldPos.y, 0);
		glEnd();
	}
	
	// Draw polygons
	glColor3f(0.5f, 0.5f, 0.5f);
	
	if (vertexList.size() == 1) {
		glBegin(GL_POINTS);
	} else {
		glBegin(GL_LINE_LOOP);
	}
	
	BOOST_FOREACH(Vector2D v, vertexList) {
		glVertex3f(v.x, v.y, 0);
	}
	glEnd();
	
	glColor4f(0.5f, 0.5f, 0.5f, 0.2f);
	glBegin(GL_POLYGON);
		BOOST_FOREACH(Vector2D v, vertexList) {
			glVertex3f(v.x, v.y, 0);
		}
	glEnd();
	
	//Draw circle preview
	if (interfaceState == CIRCLE) {
		//Co-ords of first point
		Vector2D centre = vertexList.front();
		
		//Find radius. Distance between point in the vertex list and mouse current co-ords.
		float radius = mouse.distance(centre);
		
		//Draw
		glColor4f(0.5f, 0.5f, 0.5f, 0.2f);
		Drawing::drawCircle(centre, radius, false); // outline
		
		if (radius < 0.4f) {//probably should be abstracted to a global var
			glColor4f(1.0f, 0.0f, 0.0f, 0.4f); //Circle too small. Show warning colour
		} else {
			glColor4f(0.5f, 0.5f, 0.5f, 0.2f);
		}
		Drawing::drawCircle(centre, radius, true); // fill
	}
	
	//Draw rectangle preview
	if (interfaceState == RECT) {
		//Co-ords of first point
		Vector2D corner = vertexList.front();
		
		//Draw border
		glColor3f(0.5f, 0.5f, 0.5f);
		glBegin(GL_LINE_LOOP);
			glVertex3f(corner.x, corner.y, 0);
			glVertex3f(mouse.x, corner.y, 0);
			glVertex3f(mouse.x, mouse.y, 0);
			glVertex3f(corner.x, mouse.y, 0);
		glEnd();
		
		//Size checks
		if ( (fabs(corner.x - mouse.x) < 0.1) || (fabs(corner.y - mouse.y) < 0.1) ) {
			glColor4f(1.0f, 0.0f, 0.0f, 0.4f); //Rectangle too small. Show warning colour. 
		} else {
			glColor4f(0.5f, 0.5f, 0.5f, 0.2f);
		}
		
		//Draw fill
		glDisable(GL_CULL_FACE);
		glBegin(GL_QUADS);
			glVertex3f(corner.x, corner.y, 0);
			glVertex3f(mouse.x, corner.y, 0);
			glVertex3f(mouse.x, mouse.y, 0);
			glVertex3f(corner.x, mouse.y, 0);
		glEnd();
	}
	
	//Draw marquee selection box
	if (interfaceState == SELECTING) {
		//Co-ords of first point
		Vector2D corner = vertexList.front();

		glColor3f(0.0f, 0.0f, 0.0f); //Black
		glBegin(GL_LINE_LOOP);
			glVertex3f(corner.x, corner.y, 0);
			glVertex3f(mouse.x, corner.y, 0);
			glVertex3f(mouse.x, mouse.y, 0);
			glVertex3f(corner.x, mouse.y, 0);
		glEnd();
	}

	// Draw collisions
	if (drawContacts) {
		Drawing::drawContacts(world->getContactList());
	}
	
	physics::Profiler::profiler.endMethod("Drawing");

	// the following should give us a screen based representation - co-ordinates will be in screen space (un-normalised)
	// this is so we can draw a user interface/text to the screen
	int width = Window::width();
	int height = Window::height();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, width, 0, height);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glDisable(GL_DEPTH_TEST);
	
	//Console
	//Draw alpha-blended quad on top of upper 30% of screen. 
	glLoadIdentity();
	if (interfaceState == LOADING) {
		glColor4f(0.8f, 0.8f, 0.8f, 0.5f);
		glBegin(GL_QUADS);
			glVertex3f(0,2*height/3, 0); 
			glVertex3f(width, 2*height/3, 0);
			glVertex3f(width, height, 0);
			glVertex3f(0, height, 0); 
		glEnd();
	
		//TODO Populate console with files
		//boost::filesystem::path p(".\\");
		//std::list<bfs::path> fileList = WorldIO::getSaves(p);
		//xxxxx

	} else {
		Drawing::drawStatistics(verbosity);
	}

	// now go back into the camera perspective so that Camera::screen2world will work
	camera->use();
	
	// Set up the model view matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void Sandbox2D::toggleCamera () {
	if (camera->is2D()) {
		camera = camera3D;
		camera3D->copyState(camera2D);
	} else {
		camera = camera2D;
		camera2D->copyState(camera3D);
	}
}

void Sandbox2D::togglePause() {
	paused = !paused;
	dragObject = 0;
}

void Sandbox2D::loadPreset(int p) {
	World2D* newWorld = WorldFactory::load2D(p);
	
	if (newWorld) {
		activeScene = p;
		killWorld();
		world = newWorld;
		director = new Director(world, activeScript);
		setActiveDirectorScene(activeScene);
	}
}

int Sandbox2D::getActiveScene() {
	return activeScene;
}


void Sandbox2D::reset() {
	World2D* newWorld = WorldFactory::load2D(activeScene);
	
	if (newWorld) {
		killWorld();
		world = newWorld;
		director = new Director(world, activeScript);
		setActiveDirectorScene(activeScene);
	}
}

void Sandbox2D::clearWorld() {
	killWorld();
	world = WorldFactory::init2D();
	director = new Director(world, activeScript);
	setActiveDirectorScene(activeScene);
}

void Sandbox2D::killWorld() {
	if (world) { delete world; }
	if (director) { delete director; }
	
	dragObject = 0;
	orientateObject = 0;
	vertexList.clear();
	clipboard.clear();

	interfaceState = IDLE;
	panning = false;
}

void Sandbox2D::setActiveDirectorScene(int activeScene) {
	if (activeScene == 0) {
		//Director adds constant energy to "hello world". Stop doing this if we clear the world. 
		director->setActiveDirectorScene(1);
	} else {
		director->setActiveDirectorScene(activeScene);
	}
}
