#include "sandbox3d.h"

#include "camera3d.h"
#include "drawing.h"
#include "random.h"
#include "world_factory.h"

#include "physics.h"

#include <sstream>
#include <iomanip>
#include <string>
#include <math.h>
#include <boost/foreach.hpp>

using namespace physics;

Sandbox3D::Sandbox3D () : world(0), dragging(false), paused(false), activeScene(0), dragObject(0), dragDistance(0), cameraBody(0) {
	Globals::cacheContacts = false;
	
	camera = new Camera3D(true);

	reset();
	
	camera->use();
}

void Sandbox3D::keyDown (SDL_Event& event) {

	//Mouse co-ords
	int screenX;
	int screenY;
	SDL_GetMouseState(&screenX, &screenY);
	mouse = camera->screen2world(screenX, screenY);
	
	//Key events
	//Esc - Exit
	//F	  - Toggle full screen
	//R   - Reset world to initial setup
	//C   - Add new circle at mouse co-ords
	//B   - Add new rectangle at mouse co-ords
	//V   - Add fixed circle at mouse co-ords
	//D   - Delete first shape found at mouse co-ords
	//P   - Pause world
	//Numbers 1-9, restart world with various test presets. 
	switch (event.key.keysym.sym) {
		case SDLK_ESCAPE:
		case SDLK_q:
			exit(0);
		break;
		case SDLK_f:
			Window::toggleFullscreen();
		break;
		case SDLK_r:
			reset();
		break;
		case SDLK_p: togglePause(); break;

		case SDLK_1: loadPreset(1);	break;
		case SDLK_2: loadPreset(2);	break;
		case SDLK_3: loadPreset(3); break;
		case SDLK_4: loadPreset(4); break;
		case SDLK_5: loadPreset(5); break;
		case SDLK_6: loadPreset(6); break;
		case SDLK_7: loadPreset(7); break;
		case SDLK_8: loadPreset(8); break;
		case SDLK_9: loadPreset(9);	break;
		case SDLK_0: loadPreset(0);	break;
		
		case SDLK_SPACE: {
			Vector3D spawn = camera->getNear(screenX, screenY);
			
			Vector3D target = camera->screen2world(screenX, screenY);
			
			Vector3D offset = target - spawn;
			
			Vector3D dir = offset.normalise();
			
			spawn += dir * 3;
			
			Sphere* sphere = new Sphere(spawn, 1);
			
			sphere->setDensity(5);
			
			Body3D* body = Body3D::create(sphere);
			
			world->add(body);
			
			/*float distXZ = sqrt(offset.x * offset.x + offset.z * offset.z);
			
			float t = distXZ / 100;
			
			float a = -9.8f;
			
			Vector3D v = offset;
			v.y = 0;
			v = v.normalise() * 100;
			
			v.y = offset.y / t - 0.5f * a * t;
			
			if (v.y > 50) v.y = 50;
			else if (v.y < -50) v.y = -50;*/
			
			Vector3D v = offset.normalise() * 75;
			
			body->setVelocity(v);
		}
		break;

		default: {}
	}
}

void Sandbox3D::keyUp (SDL_Event& event) {
	return;
	
	/*switch (event.key.keysym.sym) {
		default: {}
	}*/
}

void Sandbox3D::mouseDown (SDL_Event& event) {
	switch (event.button.button) {
		case SDL_BUTTON_LEFT:
			{
				int mx = event.button.x;
				int my = event.button.y;
				mouse = camera->screen2world(mx, my);
				
				Body3D* body = world->findBeneath(mouse + camera->getNormal(mx, my) / 100.0f);
				if (body && ! body->isStatic()) {
					dragObject = body;
					dragOffset = body->worldToLocal(mouse);
					dragDistance = mouse.distance(camera->getPosition());
					break;
				}
			}
		// no break here so that if we haven't clicked on a body we move the camera instead
		case SDL_BUTTON_RIGHT:
			clickPos = camera->screen2world(event.button.x, event.button.y);
			dragDistance = clickPos.distance(camera->getPosition());
			
			if (dragDistance > 25) {
				dragDistance = 25;
				clickPos = camera->getPosition() + camera->getNormal(event.button.x, event.button.y) * 25;
			}
			
			dragging = true;
		break;
		case SDL_BUTTON_WHEELUP:
			camera->zoomIn(event.button.x, event.button.y);
		break;
		case SDL_BUTTON_WHEELDOWN:
			camera->zoomOut(event.button.x, event.button.y);
		break;
	}
}

void Sandbox3D::mouseUp (SDL_Event& event) {
	switch (event.button.button) {
		case SDL_BUTTON_LEFT:
			dragObject = 0;
			dragging = false;
		break;
		case SDL_BUTTON_RIGHT:
			dragging = false;
		break;
	}
}

void Sandbox3D::update () {
	int screenX;
	int screenY;
	
	SDL_GetMouseState(&screenX, &screenY);
	mouse = camera->getPosition() + camera->getNormal(screenX, screenY) * dragDistance;//camera->screen2world(screenX, screenY);
	
	if (dragging) {
		camera->translate(clickPos - mouse);
	}
	
	if (dragObject) {
		Vector3D oldPos = dragObject->localToWorld(dragOffset);
		
		//dragDistance = camera->getPosition().distance(oldPos);
		
		Vector3D newPos = camera->getPosition() + camera->getNormal(screenX, screenY) * dragDistance;
		
		mouse = newPos;
		
		if (paused) {
			
			dragObject->translate(newPos - oldPos);
			dragObject->setVelocity(Vector3D());
			dragObject->setRotation(Rotation3D());
			//drawContacts = false;
		} else {
			Vector3D impulse = newPos - oldPos - dragObject->getVelocityAtPoint(oldPos) / 20;
			impulse *= dragObject->getMass();
			dragObject->applyImpulse(impulse, oldPos);
			
			if (Input::isPressed(SDLK_LCTRL) || Input::isPressed(SDLK_RCTRL)) {
				dragObject->setRotation(Rotation3D());
			}
		}
	}
	
	camera->update();
	
	/*cameraBody->setPosition(camera->getPosition());
	cameraBody->setVelocity(Vector3D());
	cameraBody->setRotation(Rotation3D());*/
	
	if (! paused) {
		world->step();
		
		BOOST_FOREACH(Body3D* b, world->bodies) {
			if (b->getPosition().y < -50) {
				world->remove(b);
				
				if (b == dragObject) { dragObject = 0; }
				
				// The iterator is now invalid, so we can't keep looping
				break;
			}
		}
	}
	
	//camera->translate(cameraBody->getPosition() - camera->getPosition());
}

void Sandbox3D::draw () const {
	physics::Profiler::profiler.startMethod("Drawing");
	
	camera->use();
	
	// Set up the model view matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Start drawing
	
	Drawing::drawBodies(world->bodies, camera);
	
	if (dragObject) {
		glColor3f(0.5f, 0.5f, 0.5f);
		
		Vector3D oldPos = dragObject->localToWorld(dragOffset);
		Vector3D newPos = mouse;
		
		glBegin(GL_LINES);
			glVertex3f(oldPos.x, oldPos.y, oldPos.z);
			glVertex3f(newPos.x, newPos.y, newPos.z);
		glEnd();
	}
	
	physics::Profiler::profiler.endMethod("Drawing");
	
	Drawing::drawStatistics(2);
	
	// now go back into the camera perspective so that Camera::screen2world will work
	camera->use();
	
	// Set up the model view matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void Sandbox3D::loadPreset(int p) {
	activeScene = p;
	reset();
}

void Sandbox3D::togglePause() {
	paused = ! paused;
}

void Sandbox3D::reset() {
	if (world) { delete world; }
	
	dragObject = 0;
	
	world = WorldFactory::load3D(activeScene);
	
	BOOST_FOREACH(Body3D* body, world->bodies) {
		if (! body->isStatic()) {
			//body->setOrientation(Quaternion(0.133286, 0.249324, 0.446291, 0.849056));
			body->setOrientation(Orientation3D::random());
		}
	}
	
	Sphere* sphere = new Sphere(camera->getPosition(), 2);
	
	sphere->setDensity(1000);
	
	cameraBody = Body3D::create(sphere);
	
	world->add(cameraBody);
}

