#include "world_factory.h"
#include "world_io.h"

#include "physics.h"

#include "box.h"

#include "broadphase.h"
#include "broadphase_quadtree.h"
#include "random.h"

#include <iostream>
#include <sstream>
#include <string>

using namespace physics;

void WorldFactory::createPyramid (World2D* world, const physics::Vector2D& p, float size, int layers, float hgap, float vgap) {
	int k = layers;
	int c = k;
	for (int j = 1; j <= c; ++j) {
		for (int i = 1; i <= k; ++i) {
			Vector2D position = p;
			
			position.x += (size + hgap) * (i + (j - layers) / 2.0f);
			position.y += (size + vgap) * j - size / 2;
			
			world->add(new Body2D(new Rect(position, size, size)));
		}
		k--;
	}
}

void WorldFactory::createChipmunkDemo (World2D* world, const physics::Vector2D& position, int layers) {
	float h = 2.0f;
	float w = 0.4f;
	float spacing = 3.0f;
	
	Body2D* body;
	
	for (int ix = 0; ix < layers; ix++) {
		Vector2D p = position;
		p.x -= (ix + 0.5f) * spacing - layers / 2.0f * spacing;
		
		body = Body2D::create(new Rect(p, w, h));
		body->setCanSleep(false);
		world->add(body);
		
		p.y += (w + h) / 2;
		
		body = Body2D::create(new Rect(p, h, w));
		body->setCanSleep(false);
		world->add(body);
		
		if (ix == 0) { continue; }
		
		p.x += spacing / 2;
		p.y += w;
		
		body = Body2D::create(new Rect(p, h, w));
		body->setCanSleep(false);
		world->add(body);
	}

	body = Body2D::create(new Rect(position + Vector2D((layers-1) * spacing / 2 + h / 2 - w / 2, h + w), w, h));
	body->setCanSleep(false);
	world->add(body);
	
	body = Body2D::create(new Rect(position + Vector2D((layers-1) * -spacing / 2 - h / 2 + w / 2, h + w), w, h));
	body->setCanSleep(false);
	world->add(body);
	
	if (layers > 1) {
		Vector2D p = position;
		p.y += h + w + w;
		
		createChipmunkDemo(world, p, layers-1);
	}
}

World2D* WorldFactory::load2D (int i) {
	switch (i) {
		case 0: return WorldIO::load2d("data/Hello_world.xml"); //load2D_0(); // hello world!
		case 1: return load2D_1(); //pile of stuff
		//case 2: return load2D_17(); // varying friction and restitution
		//case 3: return load2D_4_stupid(); // the old...
		//case 3: return load2D_4(); // ... and the new
		//case 4: return load2D_2(); // towering above
		case 2: return load2D_6(); // the vast pyramid
		//case 4: return load2D_5(); // the baby pyramid
		case 3: return load2D_3(); // the best demo ever (courtesy of Chipmunk) - Sssshhhhhhhhhhhhhh
		//case 9: return load2D_3_stupid(); // not so great

		case 4: return load2D_8(); // dominoes crap

		case 5: return WorldIO::load2d("data/pendulum.xml"); // pendulum
		case 6: return load2D_11(); // double pendulum
		case 7: return WorldIO::load2d("data/newton.xml"); // newton cradle
		case 8: return WorldIO::load2d("data/seesaw.xml");
		
		case 9: return load2D_13(); // rope
		case 10: return load2D_14(); // bridge and car
		
		//case 17: return load2D_15(); // naive broadphase
		//case 13: return load2D_16(); // quadtree

		// insert some demos of the contact resolution here... 

		default: {
			return 0;
			/*std::ostringstream buffer;
			buffer << i << ".xml";
			std::string presetNumber = buffer.str();	
			return WorldIO::load2d(presetNumber.c_str());*/
		}
	}
}

// Create new World3D
World3D* WorldFactory::init3D () {
	World3D* world = new World3D();
	
	/*for (int x = -10; x <= 10; x++) {
		for (int z = -10; z <= 10; z++) {
			world->add(Body3D::createStatic(new Sphere(Vector3D(x*7, -5, z*7), 5)));
		}
	}*/
	
	world->add(Body3D::createStatic(new Box(Vector3D(0, -10, 0), 100, 5, 100)));
	
	world->add(Body3D::create(new Box(Vector3D(25, -4, 0), 4, 8, 1.5)));
	world->add(Body3D::create(new Box(Vector3D(9, -4, 0), 4, 8, 1.5)));
	world->add(Body3D::create(new Box(Vector3D(15, -4, 0), 4, 8, 1.5)));
	world->add(Body3D::create(new Box(Vector3D(20, -4, 0), 4, 8, 1.5)));
	
	world->add(Body3D::createStatic(new Sphere(4)));
	
	Body3D* combo = Body3D::create();
	
	combo->add(new Sphere(Vector3D(0, 7, 0), 3));
	combo->add(new Sphere(Vector3D(0, 13, 0), 3));
	combo->add(new Sphere(Vector3D(0, 19, 0), 3));
	
	world->add(combo);
	world->add(Body3D::create(new Sphere(Vector3D(0, 40, 0), 0.5)));
	world->add(Body3D::create(new Sphere(Vector3D(0, 50, 0), 1)));
	
	return world;
}

World3D* WorldFactory::load3D (int i) {
	return init3D();
	/*switch (i) {
		case 0: return load3D_0();
		case 1: return load3D_1();
		case 2: return load3D_2();
		case 3: return load3D_3();
		case 4: return load3D_4();
		case 5: return load3D_5();
		case 6: return load3D_6();
		case 7: return load3D_7();
		case 8: return load3D_8();
		case 9: return load3D_9();
		default: return 0;
	}*/
}

// Create new World2D with borders
World2D* WorldFactory::init2D () {
	World2D* world = new World2D();
	
	Body2D* floor = Body2D::createStatic(new Rect(-120, 0, 120, -20));
	Body2D* left = Body2D::createStatic(new Rect(-100, 0, -120, 100));
	Body2D* right = Body2D::createStatic(new Rect(100, 0, 120, 100));
	Body2D* ceiling = Body2D::createStatic(new Rect(-120, 120, 120, 100));
	
	world->add(floor);
	world->add(left);
	world->add(right);
	world->add(ceiling);
	return world;
}

World2D* WorldFactory::load2D_0 () { // nice empty world preset, useful for benchmarks.
	World2D* world = init2D();
	return world;
}

// shapes demo
World2D* WorldFactory::load2D_1 () {
	World2D* world = init2D();
	
	// Two circles attached to one another
	Body2D* body = new Body2D();
	body->add(new Circle(Vector2D(0, 6), 0.6f));
	body->add(new Circle(Vector2D(0.8f, 6.8f), 0.6f));
	world->add(body);
	
	// Two circles attached by a rod
	body = new Body2D();
	body->add(new Circle(Vector2D(3, 6), 0.8f));
	body->add(new Circle(Vector2D(6, 6), 0.8f));
	body->add(new Line(Vector2D(3, 6), Vector2D(6, 6)));
	world->add(body);
	
	// Cross shape with a circle on each end
	body = new Body2D();
	body->add(new Circle(Vector2D(10, 6), 0.4f));
	body->add(new Circle(Vector2D(12, 6), 0.4f));
	body->add(new Circle(Vector2D(10, 8), 0.4f));
	body->add(new Circle(Vector2D(12, 8), 0.4f));
	body->add(new Line(Vector2D(10, 6), Vector2D(12, 8)));
	body->add(new Line(Vector2D(10, 8), Vector2D(12, 6)));
	world->add(body);
	
	// Cross shape with only lines
	body = new Body2D();
	body->add(new Line(-1, 1, 3, 5));
	body->add(new Line(-1, 5, 3, 1));
	world->add(body);
	
	// See-saw
	body = new Body2D();
	body->add(new Line(10, 5, 20, 1));
	body->add(new Line(15, 3, 14, 1));
	Circle* c = new Circle(Vector2D(14, 1), 0.2f);
	c->setDensity(1);
	body->add(c);
	world->add(body);

	//Dense circles
	body = new Body2D();
	c = new Circle(Vector2D(15, 13), 1.5f);
	c->setDensity(1.5);
	body->add(c);
	c = new Circle(Vector2D(14.5, 11), 1.5f);
	c->setDensity(1);
	body->add(c);
	world->add(body);

	//Add compound body with shapes of different densities
	body = new Body2D();
	c = new Circle(Vector2D(45, 5), 1.6f);
	c->setDensity(15);
	body->add(c);
	c = new Circle(Vector2D(55, 2.5), 1.5f);
	c->setDensity(1);
	body->add(c);
	world->add(body);
	
	// A couple of boring circles
	world->add(new Body2D(new Circle(Vector2D(18, 10), 0.8f)));
	world->add(new Body2D(new Circle(Vector2D(18, 15), 1)));

	// A boring rect
	world->add(new Body2D(new Rect(Vector2D(14, 7), 2, 1)));

	//Fixed circles to bounce around
	world->add(new Circle(Vector2D(12, 12), 1));
	world->add(new Circle(Vector2D(5, 8), 1));
	world->add(new Circle(Vector2D(2, 10), 1));

	// Simple stacking test
	int boxheight = 2;
	int offset = 0;	// ambitious with current collision resolution
	for (int i = 0; i < 5; ++i)	{
		world->add(new Body2D(new Rect(Vector3D(-6 + (offset*i),(boxheight/2)+(boxheight)*i), boxheight , boxheight)));
	}
	
	// Triangle
	Vector2D triangle[3];
	triangle[0] = Vector2D(22, 6);
	triangle[1] = Vector2D(30, 7);
	triangle[2] = Vector2D(28, 5);
	world->add(Body2D::create(physics::Polygon::create(triangle, 3)));
	
	// Pentagon
	Vector2D pentagon[5];
	for (int i = 0; i < 5; i++) {
		pentagon[i] = Vector2D(-10, 5);
		pentagon[i].x += 1.5f * cos(2 * 3.1415f * i / 5);
		pentagon[i].y += 1.5f * sin(2 * 3.1415f * i / 5);
	}
	world->add(Body2D::create(physics::Polygon::create(pentagon, 5)));
	world->add(Body2D::create(new Capsule2D(Vector2D(8, 5.5f), Vector2D(8, 8), 1)));
	
	return world;
}


// stacking test
World2D* WorldFactory::load2D_2 () {
	World2D* world = init2D();
	
	// Simple stacking test
	int boxheight = 2;
	int offset = 0;	
	int stackHeight = 20;

	for (int i = 0; i < stackHeight; ++i)	{
		world->add(new Body2D(new Rect(Vector3D(10 + (offset * i),(boxheight/2)+(boxheight)*i), boxheight , boxheight)));
	}
	
	for (int i = 0; i < stackHeight; ++i)	{
		world->add(new Body2D(new Circle(Vector3D(0,(boxheight/2)+(boxheight)*i), 1)));
	}
	
	return world;
}


// chipmunk stack - catto
World2D* WorldFactory::load2D_3 () {
	World2D* world = new World2D(1.0f / 60.0f, new NaiveBroadPhase2D());
	
	Body2D* floor = Body2D::createStatic(new Rect(-120, 0, 120, -20));
	Body2D* left = Body2D::createStatic(new Rect(-100, 0, -120, 100));
	Body2D* right = Body2D::createStatic(new Rect(100, 0, 120, 100));
	Body2D* ceiling = Body2D::createStatic(new Rect(-120, 120, 120, 100));
	
	world->add(floor);
	world->add(left);
	world->add(right);
	world->add(ceiling);
	
	world->setGravity(Vector2D(0, -4.9f));
	
	createChipmunkDemo(world, Vector2D(5, 1), 12);
	
	return world;
}

// chipmunk stack - stupid
World2D* WorldFactory::load2D_3_stupid () {
	World2D* world = new World2D(1.0f / 60.0f, 0, 0, 0, new StupidVelocityResolver<TwoDee>(), new StupidPositionResolver<TwoDee>());
	
	Body2D* floor = Body2D::createStatic(new Rect(-120, 0, 120, -20));
	Body2D* left = Body2D::createStatic(new Rect(-100, 0, -120, 100));
	Body2D* right = Body2D::createStatic(new Rect(100, 0, 120, 100));
	Body2D* ceiling = Body2D::createStatic(new Rect(-120, 120, 120, 100));
	
	world->add(floor);
	world->add(left);
	world->add(right);
	world->add(ceiling);
	
	world->setGravity(Vector2D(0, -4.9f));
	
	createChipmunkDemo(world, Vector2D(5, 1), 12);
	
	return world;
}

// single box - stupid
World2D* WorldFactory::load2D_4_stupid () {
	World2D* world = new World2D(1.0f / 60.0f, 0, 0, 0, new StupidVelocityResolver<TwoDee>(), new StupidPositionResolver<TwoDee>());
	
	Body2D* floor = Body2D::createStatic(new Rect(-120, 0, 120, -20));
	Body2D* left = Body2D::createStatic(new Rect(-100, 0, -120, 100));
	Body2D* right = Body2D::createStatic(new Rect(100, 0, 120, 100));
	Body2D* ceiling = Body2D::createStatic(new Rect(-120, 120, 120, 100));
	
	world->add(floor);
	world->add(left);
	world->add(right);
	world->add(ceiling);
	
	world->add(new Body2D(new Rect(Vector3D(3, 10), 3, 3)));
	
	return world;
}

// single box - catto
World2D* WorldFactory::load2D_4 () {
	World2D* world = init2D();
	
	world->add(new Body2D(new Rect(Vector3D(3, 10), 3, 3)));
	
	return world;
}

// stacking test - pyramid on box 
World2D* WorldFactory::load2D_5 () {
	World2D* world = init2D();
	createPyramid(world, Vector2D(5, 3), 1.0f, 15, 0.05f, 0.05f);
	
	world->add(Body2D::create(new Rect(Vector2D(2, 0.1f), Vector2D(11.5f, 2.9f))));
	
	return world;
}

// stacking test - big pyramid demo
World2D* WorldFactory::load2D_6 () {
	World2D* world = init2D();
	
	createPyramid(world, Vector2D(5, -0.5f), 1.0f, 20, 0.1f, 0.5f);
	
	return world;
}

// ball hopper demo thingee
World2D* WorldFactory::load2D_7 () { //can any ball survive the spinny things... OF DOOM!?!?
	World2D* world = init2D();
	
	Body2D* body;
	
	//hoppers
	for (int x = -50; x <= 50; x += 50) {
		Vector2D rightHopper[4];
		rightHopper[0] = Vector2D(x + 4, 8);
		rightHopper[1] = Vector2D(x + 4, 0);
		rightHopper[2] = Vector2D(x + 25, 10);
		rightHopper[3] = Vector2D(x + 25, 0);
		body = Body2D::create(physics::Polygon::create(rightHopper, 4));
		body->setStatic(true);
		world->add(body);
		
		Vector2D leftHopper[4];
		leftHopper[0] = Vector2D(x-4, 8);
		leftHopper[1] = Vector2D(x-4, 0);
		leftHopper[2] = Vector2D(x-25, 10);
		leftHopper[3] = Vector2D(x-25, 0);
		body = Body2D::create(physics::Polygon::create(leftHopper, 4));
		body->setStatic(true);
		world->add(body);
	}
	
	//boundaries...
	body = new Body2D();
	body->add(new Line(-75, 75, -75, 0));
	body->setStatic(true);
	world->add(body);
	
	body = new Body2D();
	body->add(new Line(75, 75, 75, 0));
	body->setStatic(true);
	world->add(body);
	
	
	
	for (int y = 20; y <= 90; y += 5) {
		for (int x = (y % 10 == 0 ? -70 : -67); x <= 70; x += 5) {
			
			if (y > 50) {
				world->add(new Body2D(new Circle(Vector2D(x, y), 0.5)));
			} else if (x % 10 == 0) {
				body = new Body2D();
				body->add(new Line(x-1, y, x+1, y));
				body->add(new Line(x, y-1, x, y+1));
				body->setPinned(true);
				body->setRotation(((x / 10) % 2) == 0 ? 10 : -10);
				world->add(body);
			} else {
				Vector2D pentagon[5];
				for (int i = 0; i < 5; i++) {
					pentagon[i] = Vector2D(x, y);
					pentagon[i].x += 1.5f * cos(2 * 3.1415f * i / 5);
					pentagon[i].y += 1.5f * sin(2 * 3.1415f * i / 5);
				}
				body = Body2D::create(physics::Polygon::create(pentagon, 5));
				body->setStatic(true);
				world->add(body);
			}
			
		}
	}
	
	return world;
}


// 
World2D* WorldFactory::load2D_8 () {
	World2D* world = init2D();

	for (int i = -80; i <= 80; i += 2){ 
		{
			Rect* r = new Rect(Vector2D(i, 3), 1, 6);
			r->setFriction(0.1f);
			world->add(new Body2D(r));
		}
	}
	
	world->add(new Body2D(new Circle(Vector2D(-81.5, 35), 2)));
	
	return world;
}


// Joint Demo - pendulum
World2D* WorldFactory::load2D_9 () {
	World2D* world = init2D();
	
	// Set gravity slightly stronger than normal
	world->setGravity(Vector2D(0, -15));

	// Initial Joints test

	Joint<TwoDee>* joint1 = new Joint<TwoDee>();
	
	Body2D* fixed = new Body2D(new Circle( Vector2D(0,20), 0.5) );
	Body2D* pend = new Body2D(new Rect(Vector2D(9,11), 1, 15) );
	
	fixed->setStatic(true);

	joint1->set(fixed, pend, Vector2D(0,20));
	
	world->add(fixed);
	world->add(pend);
	world->add(joint1);

	return world;
	
}


// Joint demo - newton's cradle
World2D* WorldFactory::load2D_10 () {
	World2D* world = init2D();
	
	// Set gravity slightly stronger than normal
	world->setGravity(Vector2D(0, -15));


	// Newton's cradle
	for (int i = 0; i < 16; i += 2){
		Joint<TwoDee>* joint = new Joint<TwoDee>();
		
		Body2D* fixed = new Body2D(new Circle( Vector2D( i,30), 0.5) );
		Body2D* pend = new Body2D(new Circle( Vector2D( i,20), 0.99f) );
		
		fixed->setStatic(true);
		
		pend->setDensity(5);
		pend->setRestitution(0.95f);
		pend->setFriction(0);
		
		joint->set(fixed, pend, Vector2D( i,30));
		
		world->add(fixed);
		world->add(pend);
		world->add(joint);
	}

	return world;
	
}


// Joint demo - double pendulum
World2D* WorldFactory::load2D_11 () {
	World2D* world = init2D();
	
	// Set gravity slightly stronger than normal
	world->setGravity(Vector2D(0, -15));


// Double Pendulum
	Joint<TwoDee>* jointDP1 = new Joint<TwoDee>();
	Joint<TwoDee>* jointDP2 = new Joint<TwoDee>();
	
	Body2D* fixedDP = new Body2D(new Circle( Vector2D(0,20), 0.5) );
	Body2D* pendDP1 = new Body2D(new Circle( Vector2D(1,28), 1) );
	Body2D* pendDP2 = new Body2D(new Circle( Vector2D(-0.5f, 33), 1) );
	
	fixedDP->setStatic(true);
	
	jointDP1->set(fixedDP, pendDP1, fixedDP->getPosition());
	jointDP2->set(pendDP1, pendDP2, pendDP1->getPosition());
	
	world->add(fixedDP);
	world->add(pendDP1);
	world->add(pendDP2);
	world->add(jointDP1);
	world->add(jointDP2);

return world;
}


// Joint demo - see saw
World2D* WorldFactory::load2D_12 () {
	World2D* world = init2D();
	
	// Set gravity slightly stronger than normal
	world->setGravity(Vector2D(0, -15));

	Joint<TwoDee>* joint2 = new Joint<TwoDee>();
	Body2D* seesaw = new Body2D(new Rect(Vector2D(0,5), 36, 1) );
	seesaw->setFriction(10);
	seesaw->setRestitution(0);

	Body2D* fixed = new Body2D(new Circle( Vector2D(0,3), 0.5) );
	fixed->setStatic(true);

	joint2->set(fixed, seesaw, Vector2D(0,3));

	world->add(seesaw);
	world->add(fixed);
	world->add(joint2);


	return world;
}


// Joint demo - rope
World2D* WorldFactory::load2D_13 () {
	World2D* world = init2D();
	
	// Set gravity slightly stronger than normal
	world->setGravity(Vector2D(0, -15));

	// Rope
	Body2D* fixedRope = Body2D::createStatic(new Circle( Vector2D(0,31), 1) );
	
	world->add(fixedRope);
	
	Body2D* lastRope = fixedRope;
	
	const float ropeSize = 0.6f;
	const float ropeRadius = 0.05f;
	
	for (int i = 0; i < 30; i++) {
		
		Vector2D p(0, 30 - i * ropeSize);
		
		Vector2D p1 = p;
		Vector2D p2 = p1 - Vector2D(0, ropeSize );
		
		Body2D* thisRope = Body2D::create(new Capsule2D(p1, p2, ropeRadius));
		thisRope->setRestitution(0);
		thisRope->setFriction(0);
		thisRope->setDensity(50);
		
		// Tuning
		float frequencyHz = 3.0f;
		float dampingRatio = 1.0f;

		// frequency in radians
		float omega = 2.0f * Globals::PI * frequencyHz;
		
		float mass = thisRope->getMass();

		// damping coefficient
		float d = 2.0f * mass * dampingRatio * omega;

		// spring stifness
		float k = mass * omega * omega;
		
		float timestep = world->getTimeStep();
		
		// magic formulas
		float softness = 1.0f / (d + timestep * k);
		float biasFactor = timestep * k / (d + timestep * k);
		
		Joint<TwoDee>* thisRopeJoint = new Joint<TwoDee>();
		thisRopeJoint->softness = softness;
		thisRopeJoint->biasFactor = biasFactor;
		
		thisRopeJoint->set(lastRope, thisRope, p);
		
		world->add(thisRope);
		world->add(thisRopeJoint);
		
		lastRope = thisRope;
	}
	
	Circle* ropeCircle = new Circle(lastRope->getPosition() - Vector2D(0, ropeSize/2), ropeSize);
	ropeCircle->setDensity(0);
	
	lastRope->add(ropeCircle);
	lastRope->setVelocity(Vector2D(-10, 0));
	

	

	// some other dynamics tests
	Body2D* springbox = new Body2D(new Rect(Vector2D(20,45), 2, 2) );
	CleverAnchoredSpring2D* spring = new CleverAnchoredSpring2D(Vector2D(20,50), springbox, Vector2D(),600,40);

	world->add(springbox);
	world->add(spring);



	return world;
}

// Joints - car on bridge
World2D* WorldFactory::load2D_14 () {
	World2D* world = init2D();
	
	// Set gravity slightly stronger than normal
	world->setGravity(Vector2D(0, -15));


// Bridge
	Body2D* fixedBridge1 = Body2D::createStatic(new Circle( Vector2D(-20, 50.0f), 1) );
	
	world->add(fixedBridge1);
	
	Body2D* lastPlank = fixedBridge1;
	
	const float plankWidth = 1.75f;
	const float plankHeight = 0.5f;
	const float plankSpacing = 2.0f;
	
	for (int i = 0; i < 20; i++) {
		
		Vector2D p(-19 + plankSpacing / 2 + i * plankSpacing, 50);
		
		Body2D* thisPlank = Body2D::create(new Rect(p, plankWidth, plankHeight));
		
		thisPlank->setDensity(50);
		
		// Tuning
		float frequencyHz = 1.0f;
		float dampingRatio = 0.7f;

		// frequency in radians
		float omega = 2.0f * Globals::PI * frequencyHz;
		
		float mass = thisPlank->getMass();

		// damping coefficient
		float d = 2.0f * mass * dampingRatio * omega;

		// spring stiffness
		float k = mass * omega * omega;
		
		float timestep = world->getTimeStep();
		
		// magic formulas
		float softness = 1.0f / (d + timestep * k);
		float biasFactor = timestep * k / (d + timestep * k);
		
		Joint<TwoDee>* thisJoint = new Joint<TwoDee>();
		thisJoint->softness = softness;
		thisJoint->biasFactor = biasFactor;
		
		thisJoint->set(lastPlank, thisPlank, p - Vector2D(plankSpacing / 2, 0));
		
		world->add(thisPlank);
		world->add(thisJoint);
		
		lastPlank = thisPlank;
	}
	
	Vector2D fixedBridge2P = lastPlank->getPosition() + Vector2D(1 - plankSpacing / 2, 0.0f);
	lastPlank->remove(lastPlank->getShapes().front());
	lastPlank->add(new Circle(fixedBridge2P, 1));
	lastPlank->setStatic(true);
	
	
	
	

	// quick "car"

	float height = 10;
	Vector2D triangle[3];
	triangle[0] = Vector2D(0, 0);
	triangle[1] = Vector2D(0, height); 
	triangle[2] = Vector2D(70, 0);
	world->add(Body2D::createStatic(physics::Polygon::create(triangle, 3)));

	Joint<TwoDee>* axel1 = new Joint<TwoDee>();
	Joint<TwoDee>* axel2 = new Joint<TwoDee>();

	
	Body2D* car = new Body2D(new Rect(Vector2D(0,60), 10, 2) );
	
	Body2D* wheel1 = new Body2D(new Circle( Vector2D(-4,57), 1.5) );
	Body2D* wheel2 = new Body2D(new Circle( Vector2D(4,57), 1.5) );

	axel1->set(car, wheel1, wheel1->getPosition());
	axel2->set(car, wheel2, wheel2->getPosition());

	world->add(car);
	world->add(wheel1);
	world->add(wheel2);
	world->add(axel1);
	world->add(axel2);

	return world;
}





// Naive Broadphase Test
World2D* WorldFactory::load2D_15 () {

	//physics::Globals::useSleeping = false;
	
	World2D* world = new World2D(1.0f/60.0f, new NaiveBroadPhase2D(), 0, 0, 0);
	
	Body2D* floor = Body2D::createStatic(new Rect(-120, 0, 120, -20));
	Body2D* left = Body2D::createStatic(new Rect(-100, 0, -120, 100));
	Body2D* right = Body2D::createStatic(new Rect(100, 0, 120, 100));
	Body2D* ceiling = Body2D::createStatic(new Rect(-120, 120, 120, 100));
	
	world->add(floor);
	world->add(left);
	world->add(right);
	world->add(ceiling);

	// no gravity 
	world->setGravity(Vector2D(0, 0));
	int i = 0;
	for (i = 0 ; i < 2500; ++i)
	{
		Body2D* tmp = new Body2D(new Circle(Vector2D(random_f(-100, 100),random_f(0,100)), 0.5) );
		tmp->setFriction(0.0f);
		tmp->setRestitution(1.0f);
		world->add(tmp);
	}

	return world;

}

// Quadtree Broadphase Test
World2D* WorldFactory::load2D_16 () {

	World2D* world = init2D();
	// no gravity 
	world->setGravity(Vector2D(0, 0));
	int i = 0;
	for (i = 0 ; i < 2500; ++i)
	{
		Body2D* tmp = new Body2D(new Circle(Vector2D(random_f(-100, 100),random_f(0,100)), 0.5) );
		tmp->setFriction(0.0f);
		tmp->setRestitution(1.0f);
		world->add(tmp);
	}

	return world;
}


/**
* Friction demo
*/
World2D* WorldFactory::load2D_17 () {
	World2D* world = init2D();

	// Triangle
	float height = 30;
	Vector2D triangle[3];
	triangle[0] = Vector2D(-50, 0);
	triangle[1] = Vector2D(-50, height); 
	triangle[2] = Vector2D(50, 0);
	world->add(Body2D::createStatic(physics::Polygon::create(triangle, 3)));


	//Place rectangles on the tri at different positions
	physics::Rect* r;

	//High friction rect
	r = new Rect(Vector2D(-40, height + 6), 2, 5);
	r->setFriction(1);
	r->setDensity(0.1f);
	world->add(new Body2D(r));

	//Medium fiction rect
	r = new Rect(Vector2D(-20, height), 2, 5);
	r->setFriction(0.5f);
	r->setDensity(0.1f);
	world->add(new Body2D(r));

	//Low fiction rect
	r = new Rect(Vector2D(0, height - 6), 2, 5);
	r->setFriction(0.2f);
	r->setDensity(0.1f);
	world->add(new Body2D(r));

	//Very Low fiction rect
	r = new Rect(Vector2D(20, height - 12), 2, 5);
	r->setFriction(0.01f);
	r->setDensity(0.1f);
	world->add(new Body2D(r));
	
	physics::Circle* c;
	
	c = new Circle(Vector2D(-90, 60), 1);
	c->setRestitution(1);
	world->add(new Body2D(c));

	c = new Circle(Vector2D(-85, 60), 1);
	c->setRestitution(0.7f);
	world->add(new Body2D(c));

	c = new Circle(Vector2D(-80, 60), 1);
	c->setRestitution(0.4f);
	world->add(new Body2D(c));

	c = new Circle(Vector2D(-75, 60), 1);
	c->setRestitution(0);
	world->add(new Body2D(c));

	return world;
}

/**
* domi-oh-noes!
*/
World2D* WorldFactory::load2D_18 () {
	World2D* world = init2D();
	Body2D* body;

	Vector2D ramp[4];
	Vector2D triangle[3];
	Vector2D pentagon[5];

	ramp[0] = Vector2D(-100, 11);
	ramp[1] = Vector2D(-100, 16); 
	ramp[2] = Vector2D(-81, 13.5);
	ramp[3] = Vector2D(-81, 11);
	body = Body2D::create(physics::Polygon::create(ramp, 4));
	body->setStatic(true);
	world->add(body);

	ramp[0] = Vector2D(100, 11);
	ramp[1] = Vector2D(100, 16); 
	ramp[2] = Vector2D(81, 13.5);
	ramp[3] = Vector2D(81, 11);
	body = Body2D::create(physics::Polygon::create(ramp, 4));
	body->setStatic(true);
	world->add(body);
	
	triangle[0] = Vector2D(100, 31);
	triangle[1] = Vector2D(100, 50); 
	triangle[2] = Vector2D(50, 31);
	body = Body2D::create(physics::Polygon::create(triangle, 3));
	body->setStatic(true);
	world->add(body);
	
	triangle[0] = Vector2D(-100, 31);
	triangle[1] = Vector2D(-100, 50); 
	triangle[2] = Vector2D(-50, 31);
	body = Body2D::create(physics::Polygon::create(triangle, 3));
	body->setStatic(true);
	world->add(body);
	
	triangle[0] = Vector2D(30, 21);
	triangle[1] = Vector2D(0, 26); 
	triangle[2] = Vector2D(-30, 21);
	body = Body2D::create(physics::Polygon::create(triangle, 3));
	body->setStatic(true);
	world->add(body);
	
	pentagon[0] = Vector2D(6, 3);
	pentagon[1] = Vector2D(0, 6); 
	pentagon[2] = Vector2D(-6, 3);
	pentagon[3] = Vector2D(-6, 0);
	pentagon[4] = Vector2D(6, 0);
	body = Body2D::create(physics::Polygon::create(pentagon, 5));
	body->setStatic(true);
	world->add(body);
	
	physics::Rect* r;
	
	r = new Rect(Vector2D(-55, 30), 90, 2);
	body = new Body2D(r);
	body->setStatic(true);
	world->add(body);
	
	r = new Rect(Vector2D(55, 30), 90, 2);
	body = new Body2D(r);
	body->setStatic(true);
	world->add(body);
	
	r = new Rect(Vector2D(-52.5, 10), 95, 2);
	body = new Body2D(r);
	body->setStatic(true);
	world->add(body);
	
	r = new Rect(Vector2D(52.5, 10), 95, 2);
	body = new Body2D(r);
	body->setStatic(true);
	world->add(body);
	
	r = new Rect(Vector2D(0, 20), 180, 2);
	body = new Body2D(r);
	body->setStatic(true);
	world->add(body);
	
	for (int i = -80; i <= 80; i += 2){
		if (i < -6 || i > 6){ 
			r = new Rect(Vector2D(i, 14), 1, 6);
			r->setDensity(1);
			r->setFriction(1);
			world->add(new Body2D(r));
		} else if (i == -6 || i == 6) {
			r = new Rect(Vector2D(i, 14), 1, 6);
			r->setDensity(1);
			r->setFriction(0.01f);
			world->add(new Body2D(r));
		}
	}
	
	for (int i = -80; i <= 80; i += 2){ 
		if (i < -6 || i > 6){ 
			r = new Rect(Vector2D(i, 3), 1, 6);
			r->setDensity(1);
			r->setFriction(0.1f);
			world->add(new Body2D(r));
		}
	}
	
	world->add(new Body2D(new Circle(Vector2D(-98, 52), 1)));
	world->add(new Body2D(new Circle(Vector2D(98, 52), 1)));
	
	return world;
}


