#include "world_io.h"
#include "world_factory.h"
#include "drawing.h"
#include "body.h"
#include "shape.h"
#include "joint.h"
#include "capsule.h"
#include "forces.h"
#include "display_object.h"

#include <boost/foreach.hpp>
//#include <boost/basic_directory_iterator.hpp>

//#include <filesystem>
#include <iostream>
#include <string>
#include <stdio.h>
#include <vector>

#include "../tinyxml/tinyxml.h"
#include "../tinyxml/tinystr.h"

//namespace bfs = boost::filesystem;
using namespace std;
using namespace physics;
using namespace Drawing;

void WorldIO::save2d(World2D* world) {
	save2d(world, "data/save2D.xml");
}

void WorldIO::save2d(World2D* world, std::string filename) {
	
	//Prep file for input
	TiXmlDocument doc;

	//XML declaration so it opens in proper xml viewers (browsers)
	//TiXmlDeclaration * decl = new TiXmlDeclaration( "1.0", "", "" );
	//doc.LinkEndChild( decl );
	
	//Create root node
	TiXmlElement * worldRoot = new TiXmlElement( "World" );

	//Create bodies container
	TiXmlElement * bodiesRoot = new TiXmlElement( "Bodies" );

	//Iterate over all bodies in the world
	int i = 0;
	BOOST_FOREACH(Body2D* body, world->bodies) {
		TiXmlElement * bodyNode = new TiXmlElement("Body");
		
		//Display object for body
		DisplayObject* dObj = DisplayObject::getDisplayObject(body);

		//Create and store a unique bodyID
		//Necessary for referencing which bodies have joints
		bodyNode->SetAttribute("ID", i);
		dObj->setBodyID(i);

		//Set body generic attributes
		bodyNode->SetDoubleAttribute("Rotation", body->getRotation().toFloat());
		bodyNode->SetAttribute("Staticness", body->isStatic());
		bodyNode->SetAttribute("Awake", body->isAwake());

		//Velocity
		bodyNode->SetDoubleAttribute("Velocity_x", body->getVelocity().x);
		bodyNode->SetDoubleAttribute("Velocity_y", body->getVelocity().y);
		
		//Colour
		physics::Vector3D colour = dObj->getNormalColour();
		bodyNode->SetDoubleAttribute("Color_x", colour.x);
		bodyNode->SetDoubleAttribute("Color_y", colour.y);
		bodyNode->SetDoubleAttribute("Color_z", colour.z);
		
		BOOST_FOREACH(Shape2D* shape, body->getShapes()) {
			TiXmlElement * shapeNode = new TiXmlElement("Shape");
			
			Area::Type shapeType = shape->type();
			shapeNode->SetAttribute("Type", shapeType);

			//Set verticies
			int vertex_count = shape->getVertexCount();
			shapeNode->SetAttribute("Vertex_count", vertex_count);

			for (int i = 0; i < vertex_count; i++) {
				TiXmlElement * vertexNode = new TiXmlElement("Vertex");
				vertexNode->SetDoubleAttribute("Vertex_x", shape->getVertex(i).x);
				vertexNode->SetDoubleAttribute("Vertex_y", shape->getVertex(i).y);
				shapeNode->LinkEndChild(vertexNode); //Associate vertex to shape
			}

			//Shape specific operations
			switch (shapeType) {
				case Area::CIRCLE : {
					shapeNode->SetDoubleAttribute("Radius", shape->getRadius());
					shapeNode->SetDoubleAttribute("Orientation", shape->getOrientation().getAngle());
				} 
				break;
				case Area::CAPSULE :{
					shapeNode->SetDoubleAttribute("Radius", shape->getRadius());
				} break;
				case Area::RECT : {
					Rect* rect = static_cast<Rect*>(shape);
					shapeNode->SetDoubleAttribute("Orientation", shape->getOrientation().getAngle());
					shapeNode->SetDoubleAttribute("Width", rect->width());
					shapeNode->SetDoubleAttribute("Height", rect->height()); 
				} break;
				case Area::LINE : {
				} break;
				case Area::POLYGON : {
				} break;
				default: break;
			}

			//Set shape generic attributes
			shapeNode->SetDoubleAttribute("Friction", shape->getFriction());
			shapeNode->SetDoubleAttribute("Restitution", shape->getRestitution());
			//shapeNode->SetDoubleAttribute("MOI", shape->getMOI());
			shapeNode->SetDoubleAttribute("Density", shape->getDensity());

			//Position
			shapeNode->SetDoubleAttribute("Position_x", shape->getPosition().x);
			shapeNode->SetDoubleAttribute("Position_y", shape->getPosition().y);
			
			bodyNode->LinkEndChild( shapeNode ); //Associate shape to body
		} //End shape loop		
		bodiesRoot->LinkEndChild( bodyNode ); //Associate body to bodiesRoot
		i++;
	} //End body loop
	worldRoot->LinkEndChild( bodiesRoot ); //Associate bodiesRoot to world

	//Create joints container
	TiXmlElement * jointsRoot = new TiXmlElement( "Joints" );

	//Iterate over all joints in the world
	BOOST_FOREACH(Joint<TwoDee>* joint, world->joints) {
		TiXmlElement * jointNode = new TiXmlElement("Joint");

		//get the bodies from the joint,
		//	 ...the displayobjects from the bodies 
		//    ....and the ids from the displayobjects	
		int bodyID1 = DisplayObject::getDisplayObject(joint->body1)->getBodyID();
		int bodyID2 = DisplayObject::getDisplayObject(joint->body2)->getBodyID();
		
		jointNode->SetAttribute("Body1_ID", bodyID1);
		jointNode->SetAttribute("Body2_ID", bodyID2);
		
		jointNode->SetDoubleAttribute("softness",joint->softness); //wtb encapsulation
		jointNode->SetDoubleAttribute("biasFactor",joint->biasFactor);
		
		Vector2D pos = ((joint->body1->localToWorld(joint->localAnchor1)+joint->body2->localToWorld(joint->localAnchor2))/2);
		jointNode->SetDoubleAttribute("Position_x", pos.x);
		jointNode->SetDoubleAttribute("Position_y", pos.y);
		
		jointsRoot->LinkEndChild( jointNode ); //Associate joint to jointsRoot
	}
	worldRoot->LinkEndChild( jointsRoot ); //Associate joint to world


	//Create force-generator container
	TiXmlElement * forceRoot = new TiXmlElement( "ForceGenerators" );

	//Iterate over all force-generators in the world
	//TODO TEMP: Assumes everything is an AnchoredSpring
	BOOST_FOREACH(ForceGenerator* forcegenerator, world->forces) {
		TiXmlElement * forceNode = new TiXmlElement("ForceGenerator");
		
		//AnchoredSpring is an in implementation of forcegenerator
		CleverAnchoredSpring<TwoDee>* mySpring = dynamic_cast<CleverAnchoredSpring<TwoDee>*>(forcegenerator);
		if (!mySpring) return;
		
		//BodyID
		int bodyID = DisplayObject::getDisplayObject(mySpring->getBody())->getBodyID();
		forceNode->SetAttribute("BodyID", bodyID);
		
		//Body anchor position
		Vector2D bodyAnchorPos = mySpring->getBodyAnchor();
		forceNode->SetDoubleAttribute("Body_position_x", bodyAnchorPos.x);
		forceNode->SetDoubleAttribute("Body_position_y", bodyAnchorPos.y);

		//Spring anchor position
		Vector2D anchorPos = mySpring->getAnchor();
		forceNode->SetDoubleAttribute("Spring_position_x", anchorPos.x);
		forceNode->SetDoubleAttribute("Spring_position_y", anchorPos.y);
		
		//Spring constant & damping 
		forceNode->SetDoubleAttribute("k", mySpring->getSpringConstant()); //Spring constant
		forceNode->SetDoubleAttribute("damping", mySpring->getDamping()); //Not the spring constant
		
		forceRoot->LinkEndChild( forceNode ); //Associate joint to jointsRoot
	}
	worldRoot->LinkEndChild( forceRoot ); //Associate force-gens to world


	doc.LinkEndChild( worldRoot ); //Associate world to doc
	doc.SaveFile( filename.c_str() ); //tada!
}




physics::World2D* WorldIO::load2d() {
	return load2d("data/save2D.xml");
}


physics::World2D* WorldIO::load2d(std::string filename) {
	TiXmlDocument document( filename.c_str() );
    bool loadOkay = document.LoadFile();
	World2D* world = new World2D;

	if (loadOkay) {
		//Type safety compromise before implementing BOOST_ENUM_VALUES
		const Area::Type CIRCLE = Area::CIRCLE;
		const Area::Type LINE = Area::LINE;
		const Area::Type CAPSULE = Area::CAPSULE; 
		const Area::Type RECT = Area::RECT; 
		const Area::Type POLYGON = Area::POLYGON; 

		//Create xml handle for iterating over 0s
		TiXmlHandle docHandle( &document );
		
		//Create array for looking up bodies with joints later
		vector<Body2D*> bodyVector;

		//Create elements for shape and body, updated in below loops
		TiXmlElement* elemBodies = docHandle.FirstChild("World").FirstChild("Bodies").ToElement();
		TiXmlNode* elemBodyNode = elemBodies->FirstChild("Body");
		if (!elemBodyNode) return world; //No bodies
		TiXmlElement* elemBody = elemBodyNode->ToElement();		
		TiXmlElement* elemShape;
		TiXmlElement* elemVertex;	
		
		//For every body
		for( ; elemBody; elemBody=elemBody->NextSiblingElement()) {
			Body2D* body = new Body2D;

			//Store body pointer and bodyID into map for later refference in joints
			//int BodyID = atof(elemBody->Attribute("ID"));
			bodyVector.push_back(body);
			
			//Construct body from shapes
			elemShape = elemBody->FirstChild("Shape")->ToElement();
			for( ; elemShape; elemShape=elemShape->NextSiblingElement()) {
				
				//Vertex lists
				int vertex_count = (int) atof(elemShape->Attribute("Vertex_count"));
				Vector2D* vertexList;
				vertexList = new Vector2D[vertex_count];
				int i = 0;
				elemVertex = elemShape->FirstChild("Vertex")->ToElement();
				for(; elemVertex; elemVertex=elemVertex->NextSiblingElement()) {
					vertexList[i].x = atof(elemVertex->Attribute("Vertex_x"));
					vertexList[i].y = atof(elemVertex->Attribute("Vertex_y"));
					i++;
				}				
				
				physics::Area::Type shapeType;
				const char * shapeTypeString = elemShape->Attribute("Type");
				//====================Grok===============================				
				if (strcmp(shapeTypeString, "0")==0) shapeType = CIRCLE;
				else if (strcmp(shapeTypeString, "1")==0) shapeType = LINE;
				else if (strcmp(shapeTypeString, "2")==0) shapeType = CAPSULE;
				else if (strcmp(shapeTypeString, "3")==0) shapeType = RECT;
				else if (strcmp(shapeTypeString, "4")==0) shapeType = POLYGON;
				else shapeType = CIRCLE;
				//strcmp(x,y) = 0 when x==y. 
				//====================Grok end=============================
				//Ungrocky solution I can't find out how to do. (Converting the string to the enum)
				//shapeType = static_cast<physics::Area::Type>(shapeTypeString);
				//shapeType = (Area::Type)Enum::Parse(typeof(Area::Type) shapeTypeString);
				//==================Ungrok end=============================

				//Shape to be built
				Shape2D* newShape = 0;

				//Add shape specific attributes
				//Position
				float pos_x = atof(elemShape->Attribute("Position_x"));
				float pos_y = atof(elemShape->Attribute("Position_y"));
				switch (shapeType) {
					case CIRCLE: {
						float ori = atof(elemShape->Attribute("Orientation"));
						float radius = atof(elemShape->Attribute("Radius"));
						newShape = new Circle(Vector2D(pos_x, pos_y), radius);	
						newShape->setOrientation(ori); //Only actually affects the little X shape on the circle
					} break;
					case CAPSULE: {
						float radius = atof(elemShape->Attribute("Radius"));
						newShape = new Capsule2D(Vector2D(vertexList[0].x, vertexList[0].y), Vector2D(vertexList[1].x, vertexList[1].y), radius);
					} break;
					case LINE: {
						newShape = new Line(Vector2D(vertexList[0].x, vertexList[0].y), Vector2D(vertexList[1].x, vertexList[1].y));
					} break;
					case RECT: {
						float ori = atof(elemShape->Attribute("Orientation"));
						float width = atof(elemShape->Attribute("Width")); 
						float height = atof(elemShape->Attribute("Height")); 
						newShape = new Rect(Vector2D(pos_x, pos_y),width,height,ori);
					} break;
					case POLYGON: {
						newShape = physics::Polygon::create(vertexList, vertex_count); 
					} break;
				    default: {
						//TODO: Handle invalid shape enum error
					}
				}//Switch

				//Load shape generic attributes
				//float MOI = atof(elemShape->Attribute("MOI"));
				newShape->setDensity(atof(elemShape->Attribute("Density")));
				newShape->setFriction(atof(elemShape->Attribute("Friction")));	
				newShape->setRestitution(atof(elemShape->Attribute("Restitution")));
				
				//Add shape to the body
				body->add(newShape);

				//Tidy up
				delete [] vertexList;

			}//Shape
			
			//Load body generic attributes

			//Staticness and Sleepiness. Poor tired bodies.
			body->setStatic((strcmp(elemBody->Attribute("Staticness"),"1")==0));
			body->setAwake((strcmp(elemBody->Attribute("Awake"),"1")==0));

			//Rotation
			body->setRotation(atof(elemBody->Attribute("Rotation")));
			
			//Velocity
			float velocity_x = atof(elemBody->Attribute("Velocity_x"));
			float velocity_y = atof(elemBody->Attribute("Velocity_y"));
			body->setVelocity(Vector2D(velocity_x, velocity_y));

			//Colour
			float colour_x = atof(elemBody->Attribute("Color_x"));
			float colour_y = atof(elemBody->Attribute("Color_y"));
			float colour_z = atof(elemBody->Attribute("Color_z"));
			Vector3D* colour = new Vector3D(colour_x, colour_y, colour_z);
			body->data = DisplayObject::getDisplayObject(body); //New random colour created
			DisplayObject::getDisplayObject(body)->setColour(colour); //...and overwritten

			world->add(body);
		}//Body

		//Open up joints group
		TiXmlElement* elemJoints = docHandle.FirstChild("World").FirstChild("Joints").ToElement();
		TiXmlNode* elemJointNode = elemJoints->FirstChild("Joint");
		if (elemJointNode) {
			TiXmlElement* elemJoint = elemJointNode->ToElement();
			//For every joint
			for( ; elemJoint; elemJoint=elemJoint->NextSiblingElement()) {
				int Body1_ID = atof(elemJoint->Attribute("Body1_ID"));
				int Body2_ID = atof(elemJoint->Attribute("Body2_ID"));
				float position_x = atof(elemJoint->Attribute("Position_x"));
				float position_y = atof(elemJoint->Attribute("Position_y"));
				float softness = atof(elemJoint->Attribute("softness"));
				float biasFactor = atof(elemJoint->Attribute("biasFactor"));
				
				//Look up BodyID in the map of bodyID,body
				Body2D* body1 = bodyVector.at(Body1_ID);
				Body2D* body2 = bodyVector.at(Body2_ID);

				//Build joint and add to world
				if (body1 && body2) {
					Joint<TwoDee>* joint = new Joint<TwoDee>();
					joint->set(body1,body2,Vector2D(position_x,position_y));
					joint->softness = softness; //LPC recommends you use Fairy non-bio for all joint softening
					joint->biasFactor = biasFactor;
					world->add(joint);
				}
			}//Joint
		}


		//Open up forcegenerator group
		TiXmlElement* elemForces = docHandle.FirstChild("World").FirstChild("ForceGenerators").ToElement();
		TiXmlNode* elemForceNode = elemForces->FirstChild("ForceGenerator");
		if (elemForceNode) {
			TiXmlElement* elemForce = elemForceNode->ToElement();
			//For every forcegenerator
			for( ; elemForce; elemForce=elemForce->NextSiblingElement()) {
				int BodyID = atof(elemForce->Attribute("BodyID"));
				float springConstant = atof(elemForce->Attribute("k"));
				float damping = atof(elemForce->Attribute("damping"));

				//Look up BodyID in the map of bodyID,body
				Body2D* body = bodyVector.at(BodyID);
				
				float Body_position_x = atof(elemForce->Attribute("Body_position_x"));
				float Body_position_y = atof(elemForce->Attribute("Body_position_y"));
				float Spring_position_x = atof(elemForce->Attribute("Spring_position_x"));
				float Spring_position_y = atof(elemForce->Attribute("Spring_position_y"));
				
				CleverAnchoredSpring2D* spring = new CleverAnchoredSpring2D(Vector2D(Spring_position_x,Spring_position_y), body, Vector2D(Body_position_x,Body_position_y), springConstant, damping);
				world->add(spring);
			}
		}
	
	} else {
		//File open error
		//Return empty world
		return world;
	}

	return world;
}


void WorldIO::save3d(World3D* world) {
	save3d(world, "data/save3D.xml");
}

void WorldIO::save3d(World3D* world, std::string filename) {
	//Prep file for input
	TiXmlDocument doc;
	
	//Create root node
	TiXmlElement * worldRoot = new TiXmlElement( "World" );

	//Iterate over all bodies in the world
	BOOST_FOREACH(Body3D* body, world->bodies) {
		TiXmlElement * bodyNode = new TiXmlElement("Body");
		
		BOOST_FOREACH(Shape3D* shape, body->getShapes()) {
			TiXmlElement * shapeNode = new TiXmlElement("Shape");
			bodyNode->LinkEndChild( shapeNode ); //Associate shape to body
		}
		worldRoot->LinkEndChild( bodyNode ); //Associate body to world
	}
	doc.LinkEndChild( worldRoot ); //Associate world to doc
	doc.SaveFile( filename.c_str() ); //tada!
}

physics::World3D* WorldIO::load3d() {
	return load3d("data/save3D.xml");
}

physics::World3D* WorldIO::load3d(std::string filename) {
    TiXmlDocument document( filename.c_str() );
    bool loadOkay = document.LoadFile();
	World3D* world = new World3D;

	if (loadOkay) {
		return world;
	}
	else {
		//File open error
		//Return empty world
		return world;
	}
}

/*
std::list<bfs::path> WorldIO::getSaves(const bfs::path & p) {
	
	std::list<bfs::path> fileList;


	//TEMP
	//p = bfs::path p(".\");
	//
	
	//1)Google 2)Copy 3)Paste 4)??? 5)Profit!
	unsigned long fc=0, dc=0;
	if( !bfs::exists(p) )
		std::cout<<"\nFile Not Found:"<<p.native_file_string()<<"\n";
	else if( !bfs::is_directory(p) ) {
		std::cout<<"\nFound: " <<p.native_file_string() <<"\n";
		std::cout<<"In directory:"<<p.native_file_string()<<"\n";
		bfs::directory_iterator iter(p), end_iter;

		for(; iter != end_iter; ++iter) {
			try {
				if(bfs::is_directory(*iter)) {
					++dc;
					//fileList.push_back(iter->leaf());
				}
				else {
					++fc;
					std::cout<<iter->leaf()<<"\n";
					fileList.push_back(iter->leaf());
				}
			} catch(const std::exception & ex) {
				std::cout<<iter->leaf() <<": " <<ex.what() <<std::endl;
			}
			std::cout<<fc<<" "<<dc<<std::endl;
		}  //for
	}

	return fileList;
} 
*/

